/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Bitset.h"

/******************************************************************************
 *
 * Function definitions for class Bitset.
 *
 *****************************************************************************/

/* ========================================================================= */
Bitset::Bitset(unsigned long int initsize = 0) :
  number_of_bits(initsize), number_of_ones(0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class Bitset. Creates a Bitset of a given
 *                size.
 *
 * Argument:      initsize  --  Size of the set.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  unsigned long int bsize = byteSize();

  bitarray = new unsigned char[bsize];
  memset((void *) bitarray, 0, bsize);
}

/* ========================================================================= */
Bitset::Bitset(const Bitset& bitset) :
  number_of_bits(bitset.number_of_bits), number_of_ones(bitset.number_of_ones)
/* ----------------------------------------------------------------------------
 *
 * Description:   Copy constructor for class Bitset. Creates a copy of another
 *                bitset.
 *
 * Argument:      bitset  --  Bitset to be copied.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  unsigned long int bsize = bitset.byteSize();

  bitarray = new unsigned char[bsize];
  memcpy((void *) bitarray, (void *) bitset.bitarray, bsize);
}

/* ========================================================================= */
Bitset::~Bitset()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class Bitset. Deallocates the memory used by
 *                the bitset.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  delete[] bitarray;
}

/* ========================================================================= */
Bitset& Bitset::operator=(const Bitset& bitset)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class Bitset. Copies the contents
 *                of a bitset to another one.
 *
 * Argument:      bitset  --  Bitset to be copied.
 *
 * Returns:       A reference to the Bitset that was the target of the
 *                assignment.
 *
 * ------------------------------------------------------------------------- */
{
  if (&bitset != this)
  {
    unsigned long int bsize = bitset.byteSize();

    delete[] bitarray;

    number_of_bits = bitset.number_of_bits;
    number_of_ones = bitset.number_of_ones;
    bitarray = new unsigned char[bsize];
    memcpy((void *) bitarray, (void *) bitset.bitarray, bsize);
  }

  return *this;
}

/* ========================================================================= */
bool Bitset::operator!=(const Bitset& bitset) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests two bitsets for inequality. Two bitsets are not equal
 *                if their size or contents differ.
 *
 * Argument:      bitset  --  Target of the comparison.
 *
 * Returns:       Truth value according to the result of the inequality test.
 *
 * ------------------------------------------------------------------------- */
{
  if (number_of_bits != bitset.number_of_bits)
    return true;

  unsigned long int bsize = byteSize();

  if (bsize > 0)
  {
    for (unsigned long int i = 0; i < bsize - 1; ++i)
    {
      if (bitarray[i] != bitset.bitarray[i])
      return true;
    }

    char mask = 1 << (number_of_bits & 7) - 1;
    if (mask == 0)
      mask = 0xFF;

    if (((bitarray[bsize - 1] ^ bitset.bitarray[bsize - 1]) & mask) != 0)
      return true;
  }

  return false;
}

/* ========================================================================= */
unsigned long int Bitset::operator-(const Bitset& bitset) const
/* ----------------------------------------------------------------------------
 *
 * Description:  Computes the Hamming distance (the number of bit positions in
 *               which two bit vectors differ) between two Bitsets. If the
 *               vectors have different lengths, the comparison is made only
 *               for the first bits of the bitsets.
 *
 * Argument:     bitset  --  A reference to a constant Bitset.
 *
 * Returns:      The Hamming distance between two Bitsets.
 *
 * ------------------------------------------------------------------------- */
{
  unsigned long int limit, result = 0;

  if (bitset.size() > size())
    limit = bitset.size();
  else
    limit = size();

  for (unsigned int i = 0; i < limit; i++)
  {
    if (operator[](i) != bitset[i])
      ++result;
  }

  return result;
}

/* ========================================================================= */
bool Bitset::test(unsigned long int index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Determines the state of a single bit in the bitset.
 *
 * Argument:      index  --  Index of the bit to be tested.
 *
 * Returns:       State of the bit.
 *
 * ------------------------------------------------------------------------- */
{
  if (index >= number_of_bits)
    throw BitIndexException();

  return ((bitarray[index >> 3] & (1 << (index & 7))) != 0);
}

/* ========================================================================= */
void Bitset::set()
/* ----------------------------------------------------------------------------
 *
 * Description:   Fills the bitset with 1-bits.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  memset((void *) bitarray, 0xFF, byteSize());
  number_of_ones = number_of_bits;
}

/* ========================================================================= */
bool Bitset::set(unsigned long int index)
/* ----------------------------------------------------------------------------
 *
 * Description:   Sets a single bit in the bitset.
 *
 * Argument:      index  --  Index of the bit to be set.
 *
 * Returns:       State of the bit before setting it to 1.
 *
 * ------------------------------------------------------------------------- */
{
  if (index >= number_of_bits)
    throw BitIndexException();

  bool oldbit = operator[](index);

  if (!oldbit)
  {
    bitarray[index >> 3] |= 1 << (index & 7);
    ++number_of_ones;
  }

  return oldbit;
}

/* ========================================================================= */
void Bitset::clear()
/* ----------------------------------------------------------------------------
 *
 * Description:   Fills the bitset with 0-bits.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  memset((void *) bitarray, 0, byteSize());
  number_of_ones = 0;
}

/* ========================================================================= */
bool Bitset::clear(unsigned long int index)
/* ----------------------------------------------------------------------------
 *
 * Description:   Clears a single bit in the bitset.
 *
 * Argument:      index  --  Index of the bit to be cleared.
 *
 * Returns:       State of the bit before setting it to 0.
 *
 * ------------------------------------------------------------------------- */
{
  if (index >= number_of_bits)
    throw BitIndexException();

  bool oldbit = operator[](index);

  if (oldbit)
  {
    bitarray[index >> 3] &= ~(1 << (index & 7));
    number_of_ones--;
  }

  return oldbit;
}

/* ========================================================================= */
void Bitset::flip()
/* ----------------------------------------------------------------------------
 *
 * Description:   Changes the state of all bits in the bitset.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  unsigned long int i, bsize = byteSize();

  for (i = 0; i < bsize; ++i)
    bitarray[i] ^= 0xFF;

  number_of_ones = number_of_bits - number_of_ones;
}

/* ========================================================================= */
bool Bitset::flip(unsigned long int index)
/* ----------------------------------------------------------------------------
 *
 * Description:   Switches the state of a single bit in the bitset.
 *
 * Arguments:     index  --  Index of the bit.
 *
 * Returns:       State of the bit before the operation.
 *
 * ------------------------------------------------------------------------- */
{
  if (index >= number_of_bits)
    throw BitIndexException();

  bool oldbit = operator[](index);

  bitarray[index >> 3] ^= (1 << (index & 7));
  number_of_ones += oldbit ? -1 : 1;

  return oldbit;
}

/* ========================================================================= */
unsigned long int Bitset::find() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Finds the first 1-bit in the bitset.
 *
 * Arguments:     None.
 *
 * Returns:       Index of the first 1-bit in the bitset or `number_of_bits'
 *                if there are no 1-bits in the set.
 *
 * ------------------------------------------------------------------------- */
{
  if (number_of_bits == 0)
    return 0;

  unsigned long int bsize = byteSize();
  unsigned long int i;
  for (i = 0; i < bsize && bitarray[i] == 0; ++i)
    ;

  if (i == bsize)
    return number_of_bits;

  unsigned char c = bitarray[i];
  i <<= 3;
  
  while ((c & 0x01) == 0 && i < number_of_bits)
  {
    c >>= 1;
    ++i;
  }

  return i;
}

/* ========================================================================= */
string Bitset::toString() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts the bitset to a string (a sequence of characters
 *                `0' and `1'). The leftmost bit in the string represents the
 *                first bit in the set.
 *
 * Arguments:     None.
 *
 * Returns:       String representation of the bitset.
 *
 * ------------------------------------------------------------------------- */
{
  ostrstream bitstring;
  print(bitstring);
  bitstring << ends;
  return string(bitstring.str());
}

/* ========================================================================= */
void Bitset::print(ostream& stream) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Prints the bitset as a sequence of characters `0' and `1'.
 *                The bits are printed in the same order as they are in the
 *                set; thus, the first printed bit represents the first bit
 *                in the set.
 *
 * Arguments:     stream  --  A reference to the output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  unsigned long int i, j, bsize = byteSize();
  unsigned short int maskbit;
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  j = 0;
  for (i = 0; i < bsize; ++i)
  {
    maskbit = 1;
    while (j < number_of_bits && maskbit <= 0x80)
    {
      estream << ((bitarray[i] & maskbit) ? '1' : '0');
      ++j;
      maskbit <<= 1;
    }
  }
}
