/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _Bitset_h_
#define _Bitset_h_

#include <cstring>
#include <string>
#include <iostream>
#include <strstream>
#include "Exception.h"

using namespace std;

/******************************************************************************
 *
 * A class for sets of bits.
 *
 *****************************************************************************/

class Bitset
{
public:
  Bitset(unsigned long int initsize = 0);           /* Creates a new bitset. */

  Bitset(const Bitset& bitset);                     /* Copy constructor. */

  ~Bitset();                                        /* Destructor. */

  Bitset& operator=(const Bitset& bitset);          /* Assignment operator. */

  bool operator!=(const Bitset& bitset) const;      /* Test for inequality of
                                                     * two bitsets.
						     */

  bool operator==(const Bitset& bitset) const;      /* Test for equality of
                                                     * two bitsets.
						     */

  bool operator[](unsigned long int index) const;   /* Tells the value of a bit
                                                     * in a given index.
						     */

  unsigned long int operator-                       /* Computes the Hamming */
    (const Bitset& bitset) const;                   /* distance between two
                                                     * bit sets.
						     */

  bool test(unsigned long int index) const;         /* Returns the value of a
						     * bit in the bitset.
						     */
                                        
  void set();                                       /* Sets all bits in the
                                                     * bitset.
						     */

  bool set(unsigned long int index);                /* Sets a single bit in the
                                                     * bitset.
						     */

  void clear();                                     /* Clears all bits in the
                                                     * bitset.
						     */

  bool clear(unsigned long int index);              /* Clears a single bit in
                                                     * the bitset.
						     */

  void flip();                                      /* Flips all bits in the
                                                     * bitset.
						     */

  bool flip(unsigned long int index);               /* Flips a single bit in
                                                     * the bitset.
						     */

  unsigned long int size() const;                   /* Gives the size of the
                                                     * set.
						     */

  unsigned long int count() const;                  /* Gives the number on 1-
                                                     * bits in the set.
						     */

  unsigned long int zeroCount() const;              /* Gives the number of 0-
                                                     * bits in the set.
						     */

  unsigned long int find() const;                   /* Returns the index of
                                                     * the first 1-bit in the
                                                     * set (or size() if no
                                                     * such bit is found).
						     */

  void print(ostream& stream = cout) const;         /* Prints the bitset. */

  string toString() const;                          /* Converts the bitset
                                                     * to a string.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class BitIndexException;                          /* Class for reporting
                                                     * indexing errors.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

private:
  unsigned long int byteSize() const;               /* Returns the size of the
                                                     * bitset in bytes. */

  unsigned long int number_of_bits;                 /* Size of the bitset. */

  unsigned long int number_of_ones;                 /* Current number of 1-bits
                                                     * in the bitset.
						     */

  unsigned char* bitarray;                          /* Storage for the bits. */
};



/******************************************************************************
 *
 * An exception class for reporting bitset indexing errors.
 *
 *****************************************************************************/

class Bitset::BitIndexException : public Exception
{
public:
  BitIndexException();                              /* Constructor. */

  /* default copy constructor */

  ~BitIndexException();                             /* Destructor. */

  BitIndexException&                                /* Assignment operator. */
    operator=(const BitIndexException& e);

  /* `what' inherited from class Exception */
};




/******************************************************************************
 *
 * Inline function definitions for class Bitset.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool Bitset::operator[](unsigned long int index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the state of a bit in a bitset. This function will not
 *                perform any range checks on the argument.
 *
 * Argument:      index  --  index of the bit to test.
 *
 * Returns:       Truth value according to the state of the bit.
 *
 * ------------------------------------------------------------------------- */
{
  return ((bitarray[index >> 3] & (1 << (index & 7))) != 0);
}

/* ========================================================================= */
inline bool Bitset::operator==(const Bitset& bitset) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests whether the Bitset object is equal to another bitset.
 *
 * Argument:      bitset  --  A reference to a constant Bitset.
 *
 * Returns:       Truth value according to the result of the equality test.
 *
 * ------------------------------------------------------------------------- */
{
  return !(*this != bitset);
}

/* ========================================================================= */
inline unsigned long int Bitset::size() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the size of the bitset.
 *
 * Arguments:     None.
 *
 * Returns:       Size of the bitset.
 *
 * ------------------------------------------------------------------------- */
{
  return number_of_bits;
}

/* ========================================================================= */
inline unsigned long int Bitset::count() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the number of 1-bits currently in the bitset.
 *
 * Arguments:     None.
 *
 * Returns:       Number of 1-bits currently in the bitset.
 *
 * ------------------------------------------------------------------------- */
{
  return number_of_ones;
}

/* ========================================================================= */
inline unsigned long int Bitset::zeroCount() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the number of 0-bits currently in the bitset.
 *
 * Arguments:     None.
 *
 * Returns:       Number of 0-bits currently in the bitset.
 *
 * ------------------------------------------------------------------------- */
{
  return number_of_bits - number_of_ones;
}

/* ========================================================================= */
inline unsigned long int Bitset::byteSize() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the number of bytes needed for storing the bitset.
 *
 * Arguments:     None.
 *
 * Returns:       Size of the bitset in bytes (rounded up to the nearest
 *                integer).
 *
 * ------------------------------------------------------------------------- */
{
  return ((number_of_bits + 7) >> 3);
}

/* ========================================================================= */
inline ostream& operator<<(ostream& stream, const Bitset& bitset)
/* ----------------------------------------------------------------------------
 *
 * Description:   Defines an alterative method for printing a Bitset object
 *                using the << operator.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                bitset  --  The bitset to be printed.
 *
 * Returns:       A reference to the output stream.
 *
 * ------------------------------------------------------------------------- */
{
  bitset.print(stream);
  return stream;
}



/******************************************************************************
 *
 * Inline function definitions for class Bitset::BitIndexException.
 *
 *****************************************************************************/

/* ========================================================================= */
inline Bitset::BitIndexException::BitIndexException() :
  Exception("Index out of range.")
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class Bitset::BitIndexException. Creates a
 *                new exception object.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline Bitset::BitIndexException::~BitIndexException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class Bitset::BitIndexException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline Bitset::BitIndexException& Bitset::BitIndexException::operator=
  (const Bitset::BitIndexException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class Bitset::BitIndexException.
 *
 * Arguments:     e  --  A reference to another Bitset::BitIndexException.
 *
 * Returns:       A reference to the assigned exception object.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}

#endif  /* _Bitset_h_ */
