/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "BuchiAutomaton.h"

namespace Graph
{

/******************************************************************************
 *
 * Function definitions for class BuchiAutomaton.
 *
 *****************************************************************************/

/* ========================================================================= */
BuchiAutomaton::BuchiAutomaton
  (size_type initial_number_of_states = 0, size_type initstate = 0) :
  initial_state(initstate)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class BuchiAutomaton. Initializes an
 *                automaton with a given number of states and a given
 *                initial state.
 *
 * Arguments:     initial_number_of_states  --  Initial size of the automaton
 *                                              (can be grown later).
 *                initstate                 --  Index of the automaton's
 *                                              initial state.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (initstate >= initial_number_of_states
      && !(initial_number_of_states == 0 && initstate == 0))
    throw NodeIndexException();

  nodes.reserve(initial_number_of_states);
  for (size_type i = 0; i < initial_number_of_states; ++i)
    nodes.push_back(new BuchiState());
}

/* ========================================================================= */
BuchiAutomaton& BuchiAutomaton::operator=(const BuchiAutomaton& automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class BuchiAutomaton. Assigns the
 *                contents of a BuchiAutomaton to another one.
 *
 * Argument:      automaton  --  A constant reference to a BuchiAutomaton whose
 *                               contents are to be copied.
 *
 * Returns:       A reference to the BuchiAutomaton assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  if (&automaton != this)
  {
    Graph<GraphEdgeContainer>::operator=(automaton);
    initial_state = automaton.initial_state;
  }

  return *this;
}

/* ========================================================================= */
BuchiAutomaton::size_type BuchiAutomaton::expand(size_type node_count = 1)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts a given number of states to a BuchiAutomaton.
 *
 * Argument:      node_count  --  Number of states to be inserted.
 *
 * Returns:       The index of the last inserted state.
 *
 * ------------------------------------------------------------------------- */
{
  nodes.reserve(nodes.size() + node_count);

  while (node_count > 0)
  {
    nodes.push_back(new BuchiState());
    node_count--;
  }

  return nodes.size() - 1;
}

/* ========================================================================= */
void BuchiAutomaton::compact()
/* ----------------------------------------------------------------------------
 *
 * Description:   Removes all unreachable states from the automaton.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  /*
   *  The redundant states of the automaton are removed by performing a depth-
   *  first search in the automaton starting from its initial state. During the
   *  search, all visited states are copied to a new automaton, which will
   *  consist of exactly all the reachable states of the original automaton at
   *  the end of the search.
   */

  Bitset visited(nodes.size());
  stack<size_type, deque<size_type, ALLOC(size_type) > > state_stack;
  size_type current_state, target_state;

  /*
   *  A vector of pointers to nodes representing the new automaton.
   */

  vector<Graph<GraphEdgeContainer>::Node*,
         ALLOC(Graph<GraphEdgeContainer>::Node*) > reordered_nodes;
  
  /*
   *  Mapping from old state identifiers to new state identifiers.
   */

  size_type state_number_mapping[nodes.size()];

  /*
   *  Initialize the depth-first search stack with the initial state of the
   *  automaton.
   */

  state_stack.push(initial_state);
  visited.set(initial_state);
  state_number_mapping[initial_state] = 0;
  reordered_nodes.push_back
    (new BuchiState(operator[](initial_state).accepting()));

  while (!state_stack.empty())
  {
    /*
     *  Pop a state from the stack.
     */

    current_state = state_stack.top();
    state_stack.pop();

    for (GraphEdgeContainer::const_iterator transition = 
           nodes[current_state]->edges().begin();
         transition != nodes[current_state]->edges().end();
         ++transition)
    {
      /*
       *  Process the transitions of the current state. If there exists a
       *  transition to a state not yet included in the new automaton, create
       *  a new state and insert it to the new automaton. Push the target state
       *  also on the stack so that its successors will be eventually
       *  processed.
       */

      target_state = (*transition)->targetNode();

      if (!visited[target_state])
      {
        visited.set(target_state);
        state_stack.push(target_state);

        state_number_mapping[target_state] = reordered_nodes.size();
        reordered_nodes.push_back
          (new BuchiState(operator[](target_state).accepting()));
      }

      /*
       *  Adjust the transitions between the states of the old and the new
       *  automata.
       */

      static_cast<BuchiState*>
        (reordered_nodes[state_number_mapping[current_state]])
          ->connectTo(state_number_mapping[target_state],
                      static_cast<const BuchiTransition*>(*transition)
                        ->guard());
    }

    /*
     *  After processing a state of the old automaton, release the memory
     *  allocated to the state.
     */

    delete nodes[current_state];
  }

  /*
   *  Release the memory allocated to unreachable states.
   */

  for (size_type state = 0; state < nodes.size(); ++state)
    if (!visited[state])
      delete nodes[state];

  nodes = reordered_nodes;
  initial_state = 0;
}

/* ========================================================================= */
void BuchiAutomaton::read(istream& input_stream)
/* ----------------------------------------------------------------------------
 *
 * Description:   Reads an automaton description (which may represent a
 *                generalized Bchi automaton) from a stream and stores it
 *                into the automaton object, converting it to a regular
 *                Bchi automaton if necessary.
 *
 * Argument:      input_stream  --  A reference to an input stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  size_type number_of_states, current_state, neighbor_state;
  unsigned long int number_of_acceptance_sets, acceptance_set;
  Exceptional_istream einput_stream(&input_stream, ios::failbit | ios::badbit);

  clear();

  try 
  {
    /*
     *  Read in the number of states in the generalized Bchi automaton.
     */

    einput_stream >> number_of_states;

    /*
     *  If the automaton is empty, do nothing.
     */

    if (number_of_states == 0)
      return;

    einput_stream >> number_of_acceptance_sets;

    /*
     *  Allocate space for the regular Bchi automaton that will be constructed
     *  from the generalized Bchi automaton.
     */

    size_type number_of_nodes;

    if (number_of_acceptance_sets == 0)
      number_of_nodes = number_of_states;
    else
      number_of_nodes = number_of_states * number_of_acceptance_sets;

    nodes.reserve(number_of_nodes);
    for (size_type i = 0; i < number_of_nodes; ++i)
      nodes.push_back(new BuchiState(false));

    /*
     *  The automaton state numbers will be mapped from input file identifiers
     *  to the interval [0...(number of states - 1)].
     */

    map<long int, size_type, less<long int>, ALLOC(size_type) >
      state_number_map;
    pair<long int, size_type> state_mapping(0, 0);

    pair<map<long int, size_type, less<long int>, ALLOC(size_type) >
             ::const_iterator,
         bool>
      state_finder;

    /*
     *  Also the acceptance set numbers will be mapped to the interval
     *  [0...(number of acceptance sets - 1)].
     */

    map<long int, unsigned long int, less<long int>, ALLOC(unsigned long int) >
      acceptance_set_map;
    pair<long int, unsigned long int> acceptance_set_mapping(0, 0);

    pair<map<long int, unsigned long int, less<long int>,
             ALLOC(unsigned long int) >::const_iterator,
         bool>
      acceptance_set_finder;

    /*
     *  A bitset is used to keep information about states which have been
     *  processed. (This is used to verify that each state is defined exactly
     *  once in the input.)
     */

    Bitset processed_states(number_of_states);

    /*
     *  Another bitset is used to keep information about the acceptance sets
     *  of which the `current' state is a member.
     */

    Bitset acceptance_sets(number_of_acceptance_sets);

    bool initial_state_fixed = false;       /* will be set to true after    */
                                            /* processing the initial state */
                                            /* of the automaton             */

    int is_initial;

    LtlFormula* guard = 0;

    for (size_type i = 0; i < number_of_states; ++i)
    {
      /*
       *  Begin processing a new state by reading a state id from the stream.
       */

      einput_stream >> state_mapping.first;

      /*
       *  Try to insert a new state mapping into the state identifier map.
       *  If an insertion actually occurs (there is no element in the map with
       *  key equal to `state_mapping.first'), increment the state counter. In
       *  any case, set the value of `current_state' to the map value.
       *  (Note: It is an error to redefine a state that has already been
       *  processed.)
       */

      state_finder = state_number_map.insert(state_mapping);

      if (!state_finder.second)  /* no insertion occurred */
      {
        current_state = (state_finder.first)->second;
        if (processed_states.test(current_state))
          throw AutomatonParseException("State redefinition encountered.");
      }
      else
      {
        if (state_mapping.second >= number_of_states)
          throw AutomatonParseException("Number of different state "
                                        "identifiers does not match the size"
                                        " of the Bchi automaton.");

        current_state = state_mapping.second;
        state_mapping.second++;
      }

      /* 
       *  Check whether the current state is an initial state. (There must be
       *  exactly one initial state.)
       */

      einput_stream >> is_initial;

      if (is_initial != 0)
      {
        if (!initial_state_fixed)
        {
          initial_state = current_state;
          initial_state_fixed = true;
        }
        else
          throw AutomatonParseException("Multiple initial state definitions.");
      }

      /*
       *  Determine which acceptance sets the current state belongs to.
       *  The numbers of the acceptance sets are mapped to the proper
       *  interval, again by introducing mappings for acceptance set numbers
       *  whenever necessary.
       *
       *  If the state is a member of acceptance set 0 (in the mapping), make
       *  it an accepting state.
       *
       *  NOTE: If there are no acceptance sets, it is assumed that the state
       *        is accepting!
       */

      acceptance_sets.clear();

      while (1)
      {
        einput_stream >> acceptance_set_mapping.first;

        if (acceptance_set_mapping.first == -1)
          break;

        acceptance_set_finder =
          acceptance_set_map.insert(acceptance_set_mapping);

        if (!acceptance_set_finder.second)
          acceptance_set = (acceptance_set_finder.first)->second;
        else
        {
          if (acceptance_set_mapping.second >= number_of_acceptance_sets)
            throw AutomatonParseException("Number of acceptance sets "
                                          "does not match automaton state "
                                          "definitions.");

          acceptance_set = acceptance_set_mapping.second;
          ++acceptance_set_mapping.second;
        }

        if (acceptance_set == 0)
          operator[](current_state).accepting() = true;

        acceptance_sets.set(acceptance_set);
      }

      if (number_of_acceptance_sets == 0)
        operator[](current_state).accepting() = true;

      /*
       *  Process the transitions from the state to other states. Read a
       *  target state id and add a mapping for it in the translation table if
       *  necessary. Then, read the propositional formula guarding the
       *  transition. Finally, connect all `copies' (the number of which is the
       *  same as the number of acceptance sets) of the current state to the
       *  proper copies of the target state (the special case occurs when the
       *  copy of the current state belongs to some acceptance set).
       */

      while (1)
      {
        einput_stream >> state_mapping.first;

        if (state_mapping.first == -1)
          break;

        state_finder = state_number_map.insert(state_mapping);
        if (!state_finder.second)
          neighbor_state = (state_finder.first)->second;
        else
        {
          if (state_mapping.second >= number_of_states)
            throw AutomatonParseException("Number of different state "
                                          "identifiers does not match the size"
                                          " of the Bchi automaton.");

          neighbor_state = state_mapping.second;
          state_mapping.second++;
        }

        try
        {
          FormulaString guard_formula;

          input_stream >> guard_formula;

          guard = formulaStringToLtlFormula(guard_formula);

          if (!guard->propositional())
          {
            delete guard;
            throw AutomatonParseException("Illegal operators in guard "
                                          "formula.");
          }
        }
        catch (FormulaString::ParseErrorException)
        {
          throw AutomatonParseException("Error parsing guard formula.");
        }

        if (number_of_acceptance_sets <= 1)
          operator[](current_state).connectTo(neighbor_state, *guard);
        else
        {
          unsigned long int k;
          for (unsigned long int j = 0; j < number_of_acceptance_sets; ++j)
          {
            k = (acceptance_sets[j] ? (j + 1) % number_of_acceptance_sets : j);
            operator[](current_state + j * number_of_states).
              connectTo(neighbor_state + k * number_of_states, *guard);
          }
        }

        delete guard;
      }

      processed_states.set(current_state);
    }

    if (!initial_state_fixed)
      throw AutomatonParseException("No initial state specified.");
    
    if (processed_states.count() != number_of_states)
      throw AutomatonParseException("Incomplete automaton definition.");
  }
  catch (IOException e)
  {
    clear();
    if (input_stream.eof())
      throw IOException("Unexpected end of input.");
    else
      throw;
  }
  catch (...)
  {
    clear();
    throw;
  }

  /*
   *  Remove all unreachable states from the final automaton.
   */

  compact();
}

/* ======================================================================== */
void BuchiAutomaton::print
  (ostream& stream = cout, int indent = 0, GraphOutputFormat fmt = NORMAL)
  const
/* ---------------------------------------------------------------------------
 *
 * Description:   Writes information about a BuchiAutomaton to a stream.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the format of the output.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------ */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (fmt == DOT)
    estream << string(indent, ' ') + "digraph G {\n";

  if (nodes.empty())
  {
    if (fmt == NORMAL)
      estream << string(indent, ' ') + "The Bchi automaton is empty.\n";
  }
  else
  {
    if (fmt == NORMAL)
    {
      pair<size_type, unsigned long int> statistics = stats();
      pair<size_type, unsigned long int> reachable_part_statistics
        = subgraphStats(initial_state);

      estream << string(indent, ' ') + "The Bchi automaton consists of\n"
                 + string(indent + 4, ' ')
              << statistics.first
              << " states and\n" + string(indent + 4, ' ')
              << statistics.second
              << " transitions.\n" + string(indent, ' ')
                 + "The reachable part of the automaton contains\n"
                 + string(indent + 4, ' ')
              << reachable_part_statistics.first
              << " states and\n" + string(indent + 4, ' ')
              << reachable_part_statistics.second
              << " transitions.\n" + string(indent, ' ') + "Initial state: "
              << initial_state
              << '\n';
    }

    size_type s = nodes.size();
    for (size_type state = 0; state < s; ++state)
    {
      estream << string(indent, ' ');
      if (fmt == NORMAL)
      {
        estream << "State " << state << ":\n";
        operator[](state).print(stream, indent + 4);
      }
      else if (fmt == DOT)
      {
        GraphEdgeContainer::const_iterator transition;

        estream << "  n" << state << " [";
        if (state == 0)
          estream << "style=filled,";
        estream << string("shape=")
                   + (operator[](state).accepting() ? "double" : "")
                   + "circle,label=\""
                << state
                << "\",fontsize=12];\n";

        for (transition = nodes[state]->edges().begin();
             transition != nodes[state]->edges().end();
             ++transition)
        {
          estream << string(indent + 2, ' ') + 'n' << state;
          (*transition)->print(stream, indent, fmt);
          estream << ";\n";
        }
      }
    }
  }

  if (fmt == DOT)
    estream << string(indent, ' ') + "}\n";

  estream.flush();
}



/******************************************************************************
 *
 * Function definitions for class BuchiAutomaton::BuchiTransition.
 *
 *****************************************************************************/

/* ========================================================================= */
BuchiAutomaton::BuchiTransition& BuchiAutomaton::BuchiTransition::operator=
  (const BuchiTransition& transition)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class
 *                BuchiAutomaton::BuchiTransition. Assigns the value of a
 *                BuchiTransition to another one.
 *
 * Arguments:     transition  --  BuchiTransition to be assigned to "this"
 *                                transition.
 *
 * Returns:       A reference to the BuchiTransition assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  if (&transition != this)
  {
    delete guard_formula;
    guard_formula = transition.guard_formula->clone();
    Edge::operator=(transition);
  }

  return *this;
}

/* ========================================================================= */
void BuchiAutomaton::BuchiTransition::print
  (ostream& stream = cout, int indent = 0, GraphOutputFormat fmt = NORMAL)
  const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a transition between two states of
 *                a Bchi automaton.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the format of output.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (fmt == NORMAL)
  {
    estream << string(indent, ' ') + "Transition to state "
            << targetNode()
            << " [ guard: " << *guard_formula << " ]\n";
  }
  else if (fmt == DOT)
  {
    ostrstream formula_stream;
    formula_stream << *guard_formula << ends;
    formula_stream.freeze();
    char *formula = formula_stream.str();

    estream << " -> n" << targetNode() << " [label=\"";
    for (unsigned long int i = 0; formula[i] != '\0'; ++i)
    {
      if (formula[i] == '/')
      {
        estream << "&&";
        i++;
      }
      else if (formula[i] == '\\')
      {
        estream << "||";
        i++;
      }
      else
        estream << formula[i];
    }
    estream << "\",fontsize=10,fontname=\"Courier-Bold\"]";
  }

  estream.flush();
}



/******************************************************************************
 *
 * Function definitions for class BuchiAutomaton::BuchiState.
 *
 *****************************************************************************/

/* ========================================================================= */
void BuchiAutomaton::BuchiState::print
  (ostream& stream = cout, int indent = 0, GraphOutputFormat fmt = NORMAL)
  const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a state of a Bchi automaton.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the output format of the state.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------ */
{
  if (fmt == DOT)
    return;

  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << string(indent, ' ') + "The state is " + (accept ? "" : "not ")
             + "accepting.\n";

  if (!edges().empty())
  {
    GraphEdgeContainer::const_iterator edge;

    for (edge = edges().begin(); edge != edges().end(); ++edge)
      (*edge)->print(stream, indent);

  } else
    estream << string(indent, ' ') + "No transitions to other states.\n";

  estream.flush();
}

}
