/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _Configuration_h_
#define _Configuration_h_

#include <cctype>
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <map>
#include <set>
#include <strstream>
#include <string>
#include <utility>
#include <vector>
#include "Exception.h"
#include "RandomFormulaParameters.h"
#include "StlConfig.h"

using namespace std;

/******************************************************************************
 *
 * A class for storing program configuration information.
 *
 *****************************************************************************/

class Configuration
{
public:

  Configuration();                                  /* Constructor. */

  ~Configuration();                                 /* Destructor. */

  void read();                                      /* Reads the program
                                                     * configuration from the
                                                     * configuration file.
						     */

  void print                                        /* Writes the current    */
    (ostream& stream = cout, int indent = 0) const; /* configuration (in a
                                                     * textual form) to a
                                                     * stream.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  enum InteractionMode {NEVER, ALWAYS, ONERROR};    /* Enumeration constants
                                                     * affecting the behaviour
                                                     * of the program as
                                                     * regards user control.
						     */

  enum FormulaMode {NORMAL, NNF};                   /* Enumeration constants
                                                     * affecting the generation
                                                     * and output of random
                                                     * formulae.
						     */

  enum StateSpaceMode {RANDOMGRAPH, RANDOMPATH,     /* Enumeration constants */
                       ENUMERATEDPATH};             /* affecting the
						     * generation of random
                                                     * state spaces.
                                                     */

  enum ProductMode {LOCAL, GLOBAL};                 /* Enumeration constants
                                                     * for controlling the
                                                     * scope of synchronous
                                                     * products.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  struct AlgorithmInformation                       /* A structure for storing
                                                     * information about a
                                                     * particular algorithm
                                                     * (name, path to
                                                     * executable, command-line
                                                     * parameters).
						     */
  {
    string* name;                                   /* Name of the algorithm.
                                                     */

    string* path_to_program;                        /* Path to the executable
                                                     * required for running
                                                     * the algorithm.
						     */

    string* extra_parameters;                       /* Additional command-line
                                                     * parameters required for
                                                     * running the executable.
                                                     */

    bool enabled;                                   /* Determines whether the
						     * algorithm is enabled
						     * (whether it will be used
						     * in the tests or not).
						     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  struct GlobalConfiguration                        /* A structure for storing
                                                     * all the information
                                                     * affecting the general
                                                     * behaviour of the
                                                     * program.
						     */
  {
    int verbosity;                                  /* Determines the verbosity
                                                     * of program output (0-5,
                                                     * the bigger the value,
                                                     * the more information
                                                     * will be shown).
						     */

    InteractionMode interactive;                    /* Controls the behaviour
                                                     * of the program as
                                                     * regards the ability of
                                                     * the user to enter
                                                     * commands between test
                                                     * rounds. Possible values
                                                     * and their meanings are:
                                                     *
                                                     * NEVER:
                                                     *   Run all tests without
                                                     *   interruption.
                                                     * ALWAYS:
                                                     *   Pause after each test
                                                     *   round to wait for
                                                     *   user commands.
                                                     * ONERROR:
                                                     *   Try to run the tests
                                                     *   without interruption.
                                                     *   However, in case of
                                                     *   an error, pause and
                                                     *   wait for user
                                                     *   commands.
                                                     */

    unsigned long int number_of_rounds;             /* Number of test rounds.
						     */

    unsigned long int statespace_change_interval;   /* Determines the frequency
                                                     * (in rounds) of how often
                                                     * a new reachability graph
                                                     * is generated.
						     */

    unsigned long int formula_change_interval;      /* Determines the frequency
                                                     * (in rounds) of how often
                                                     * a new formula is
                                                     * generated.
						     */

    ProductMode product_mode;                       /* Determines the scope of
                                                     * the synchronous products
                                                     * computed by the program.
                                                     * Possible values and
                                                     * their meanings are:
                                                     *
                                                     * LOCAL:
                                                     *   The synchronous
                                                     *   products are computed
                                                     *   only with respect to
                                                     *   the initial state of
                                                     *   the system. This will
                                                     *   save memory but makes
                                                     *   the algorithm cross-
                                                     *   comparisons less
                                                     *   powerful, possibly
                                                     *   at the cost of
                                                     *   chances for finding
                                                     *   inconsistencies in the
                                                     *   results.
                                                     * GLOBAL:
                                                     *   The synchronous
                                                     *   products are computed
                                                     *   with respect to each
                                                     *   system state (i.e.
                                                     *   the formula is model
                                                     *   checked in each system
                                                     *   state separately. This
                                                     *   will usually require
                                                     *   much more memory than
                                                     *   the other alternative.
                                                     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  struct StatespaceConfiguration                    /* A structure for storing
                                                     * specific information
                                                     * affecting the
                                                     * construction of
                                                     * reachability graphs.
						     */
  {
    unsigned long int size;                         /* Size of the graph to be
                                                     * generated.
						     */

    unsigned long int max_size;                     /* Maximum size of the
                                                     * graph to be generated.
                                                     */

    unsigned long int atoms_per_state;              /* Number of propositional
                                                     * variables associated
                                                     * with each state.
						     */

    double edge_probability;                        /* Probability of adding
                                                     * a random transition
                                                     * between two different 
                                                     * states.
						     */

    double truth_probability;                       /* Probability of making
                                                     * the value of any
                                                     * propositional variable
                                                     * true in a state.
						     */

    Configuration::StateSpaceMode generate_mode;    /* Determines the type of
                                                     * the generated state
                                                     * spaces. Possible values
                                                     * are:
                                                     *
                                                     * RANDOMGRAPH:    
                                                     *   Generate random
                                                     *   connected graphs as
                                                     *   state spaces.
                                                     * RANDOMPATH:
                                                     *   Generate paths as
                                                     *   state spaces, choose
						     *   the loop and the
						     *   truth assignments for
						     *   atomic propositions
						     *   randomly.
						     * ENUMERATEDPATH:
						     *   Generate paths as
						     *   state spaces by
						     *   enumerating all
						     *   possible paths of a
						     *   given length.
                                                     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  struct FormulaConfiguration                       /* A structure for storing
                                                     * specific information
                                                     * affecting the generation
                                                     * of random formulae.
						     */
  {
    bool allow_abbreviated_operators;               /* Determines whether the
                                                     * operators ->, <->, <>
                                                     * and [] are allowed when
                                                     * generating random
                                                     * formulae (these are
                                                     * `abbreviated' operators
                                                     * since they could be
                                                     * written in an equivalent
                                                     * form by using another
                                                     * operators).
						     */

    Configuration::FormulaMode output_mode;         /* Determines whether the
                                                     * generated formulae are
                                                     * to be converted to
                                                     * negation normal form
                                                     * before passing them to
                                                     * the different
                                                     * algorithms. Possible
                                                     * values are:
                                                     *
                                                     * NORMAL:
                                                     *   No conversion is
                                                     *   required or desired.
                                                     * NNF:
                                                     *   Do the conversion
                                                     *   (this may affect the
                                                     *   length of the
                                                     *   formulae!)
                                                     */

    Configuration::FormulaMode generate_mode;       /* Determines whether the
                                                     * formulae are to be
                                                     * generated in negation
                                                     * normal form (strict
                                                     * length requirement for
                                                     * formulae). Possible
                                                     * values are:
                                                     *
                                                     * NORMAL:
                                                     *   Allow more flexibility
                                                     *   in the generation of
                                                     *   formulae.
                                                     * NNF:
                                                     *   Force generation into
                                                     *   negation normal form.
                                                     */

    RandomFormulaParameters generation_parameters;  /* Stores the parameters
                                                     * (e.g. probabilities of
                                                     * different operators) for
                                                     * generating the formulae.
                                                     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  vector<AlgorithmInformation,                      /* A vector containing   */
         ALLOC(AlgorithmInformation) > algorithms;  /* information about the
                                                     * algorithms used in
                                                     * the tests.
						     */

  GlobalConfiguration global_options;               /* General configuration
                                                     * information.
						     */

  StatespaceConfiguration statespace_options;       /* Configuration
                                                     * information for
                                                     * reachability graph
                                                     * construction.
						     */

  FormulaConfiguration formula_options;             /* Configuration
                                                     * information for
                                                     * generating random
                                                     * formulae.
						     */

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class ConfigurationException : public Exception   /* A class for reporting
                                                     * errors when reading
                                                     * the configuration file.
                                                     */
  {                                                 
  public:
    ConfigurationException                          /* Constructor. */
      (int line = -1, const string& msg = "");

    ~ConfigurationException();                      /* Destructor. */

    ConfigurationException&                         /* Copy constructor. */
      operator=(const ConfigurationException& e);

    /* `what' inherited from class Exception */

    /* `changeMessage' inherited from class Exception */

    int line_number;                                /* Number of the line with
                                                     * an error.
						     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

private:
  Configuration(const Configuration& cfg);          /* Copy constructor (hidden
						     * from external usage).
						     */

  Configuration& operator=                          /* Assignment operator   */
    (const Configuration& cfg);                     /* (hidden from external
						     * usage).
						     */

  void reset();                                     /* Initializes the
                                                     * configuration data
                                                     * to default values.
						     */
};



/******************************************************************************
 *
 * Inline function definitions for class Configuration::ConfigurationException.
 *
 *****************************************************************************/

/* ========================================================================= */
inline Configuration::ConfigurationException::ConfigurationException
  (int line = -1, const string& msg = "") :
  Exception(msg), line_number(line)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class Configuration::ConfigurationException.
 *                Creates a new exception object, initializing it with an
 *                error message and configuration file line number.
 *
 * Arguments:     line  --  Line number (-1 if no line number is to be shown
 *                          when displaying the message).
 *                msg   --  A reference to a constant string containing the
 *                          error message.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline Configuration::ConfigurationException::~ConfigurationException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class Configuration::ConfigurationException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline Configuration::ConfigurationException&
Configuration::ConfigurationException::operator=
  (const ConfigurationException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assigmment operator for class
 *                Configuration::ConfigurationException.
 *
 * Argument:      e  --  A reference to a constant exception object of the
 *                       same type.
 *
 * Returns:       A reference to the exception object assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  line_number = e.line_number;

  return *this;
}



#endif  /* _Configuration_h_ */
