/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "LtlFormula.h"

namespace LtlFormula
{

/******************************************************************************
 *
 * Function definitions for class LtlFormula.
 *
 *****************************************************************************/

/* ========================================================================= */
LtlFormula& LtlFormula::operator=(const LtlFormula& f)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class LtlFormula.
 *
 * Argument:      f  --  A reference to a constant LtlFormula that is assigned
 *                       to this LtlFormula.
 *
 * Returns:       A reference to the LtlFormula whose value was changed.
 *
 * ------------------------------------------------------------------------- */
{
  if (&f != this)
  {
    formula_string = f.formula_string;
    is_propositional = f.is_propositional;
    is_constant = f.is_constant;
  }
 
  return *this;
}

/* ========================================================================= */
bool LtlFormula::evaluatePath
  (const StateSpace& reachability_graph,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >&
     path_states,
   StateSpace::size_type loop_state,
   Bitset* result_set = 0) const
/* ----------------------------------------------------------------------------
 *
 * Description:
 *
 * Arguments:
 *
 * Returns:
 *
 * ------------------------------------------------------------------------- */
{
  if (path_states.empty())
    return false;

  map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) > result_cache;

  evalPath(reachability_graph, path_states, loop_state, result_cache);

  bool result = (*result_cache[this])[0];

  if (result_set != 0)
    *result_set = Bitset(*result_cache[this]);

  for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
         cache_iterator = result_cache.begin();
       cache_iterator != result_cache.end();
       ++cache_iterator)
    delete cache_iterator->second;

  return result;
}

/* ========================================================================= */
bool LtlFormula::displayPathProof
  (ostream& stream, int indent, const StateSpace& reachability_graph,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >&
     path_states,
   StateSpace::size_type loop_state) const
/* ----------------------------------------------------------------------------
 *
 * Description:
 *
 * Arguments:
 *
 * Returns:
 *
 * ------------------------------------------------------------------------- */
{
  if (path_states.empty())
    return false;

  map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) > result_cache;

  evalPath(reachability_graph, path_states, loop_state, result_cache);

  bool result = (*result_cache[this])[0];
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  try
  {
    prettyPrintPathProof(estream, indent, result, 0, result_cache,
			 path_states, loop_state);
  }
  catch (...)
  {
    for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
	   cache_iterator = result_cache.begin();
	 cache_iterator != result_cache.end();
	 ++cache_iterator)
      delete cache_iterator->second;

    throw;
  }

  for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
	 cache_iterator = result_cache.begin();
       cache_iterator != result_cache.end();
       ++cache_iterator)
    delete cache_iterator->second;

  return result;
}

/* ========================================================================= */
bool LtlFormula::evaluatePath
  (const StateSpace& path, Bitset* result_set = 0) const
/* ----------------------------------------------------------------------------
 *
 * ------------------------------------------------------------------------- */
{
  if (path.empty())
    return false;

  vector<StateSpace::size_type, ALLOC(StateSpace::size_type) > path_states;
  map<StateSpace::size_type, StateSpace::size_type,
      less<StateSpace::size_type>, ALLOC(StateSpace::size_type) > ordering;

  StateSpace::size_type state = path.initialState();
  StateSpace::size_type state_count = 0;

  while (1)
  {
    path_states.push_back(state);
    ordering[state] = state_count;
    state_count++;

    state = (*(path[state].edges().begin()))->targetNode();

    if (ordering.find(state) != ordering.end())
      break;
  }

  return evaluatePath(path, path_states, ordering[state], result_set);
}

/* ========================================================================= */
bool LtlFormula::displayPathProof
  (ostream& stream, int indent, const StateSpace& path) const
/* ----------------------------------------------------------------------------
 *
 * Description:
 *
 * Arguments:
 *
 * Returns:
 *
 * ------------------------------------------------------------------------- */
{
  if (path.empty())
    return false;

  vector<StateSpace::size_type, ALLOC(StateSpace::size_type) > path_states;
  map<StateSpace::size_type, StateSpace::size_type,
      less<StateSpace::size_type>, ALLOC(StateSpace::size_type) > ordering;

  StateSpace::size_type state = path.initialState();
  StateSpace::size_type state_count = 0;

  while (1)
  {
    path_states.push_back(state);
    ordering[state] = state_count;
    state_count++;

    state = (*(path[state].edges().begin()))->targetNode();

    if (ordering.find(state) != ordering.end())
      break;
  }

  return displayPathProof(stream, indent, path, path_states, ordering[state]);
}

/* ========================================================================= */
void LtlFormula::evalPath
  (const StateSpace& reachability_graph,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >&
     path_states,
   StateSpace::size_type loop_state,
   map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >& result_cache)
  const
/* ----------------------------------------------------------------------------
 *
 * ------------------------------------------------------------------------- */
{
  stack<const LtlFormula*, deque<const LtlFormula*,
                                 ALLOC(const LtlFormula*) > >
    subformula_stack;
  const LtlFormula* f;
  Bitset* val;

  recCollectSubformulae(subformula_stack);

  while (!subformula_stack.empty())
  {
    if (::user_break)
    {
      for (map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >::iterator
	     cache_iterator = result_cache.begin();
	   cache_iterator != result_cache.end();
	   ++cache_iterator)
	delete cache_iterator->second;

      throw UserBreakException();
    }

    f = subformula_stack.top();
    subformula_stack.pop();

    if (result_cache.find(f) == result_cache.end())
    {
      val = result_cache[f] = new Bitset(path_states.size());

      if (f->what() == LTL_UNTIL || f->what() == LTL_V
	  || f->what() == LTL_FINALLY || f->what() == LTL_GLOBALLY)
      {
        StateSpace::size_type marker_state;
        LtlFormula* g, *h;
        bool check_value;
        bool marker_valid = false;

	switch (f->what())
	{
  	  case LTL_UNTIL :
	    g = static_cast<const Until*>(f)->subformula1;
	    h = static_cast<const Until*>(f)->subformula2;
	    check_value = true;
	    break;

  	  case LTL_V :
	    val->set();
	    g = static_cast<const V*>(f)->subformula1;
	    h = static_cast<const V*>(f)->subformula2;
	    check_value = false;
	    break;

	  case LTL_FINALLY :
	    g = 0;
	    h = static_cast<const Finally*>(f)->subformula;
	    check_value = true;
	    break;

	  case LTL_GLOBALLY :
	    val->set();
	    g = 0;
	    h = static_cast<const Globally*>(f)->subformula;
	    check_value = false;
	    break;

	  default:
	    break;
        }

        for (StateSpace::size_type state = 0; state < path_states.size();
	     state++)
        {
          if (result_cache[h]->test(state) == check_value)
          {
            val->flip(state);

            if (marker_valid)
            {
              while (marker_state < state)
              {
                val->flip(marker_state);
                marker_state++;
              }
              marker_valid = false;
            }
          }
          else if (g == 0 || result_cache[g]->test(state) == check_value)
          {
            if (!marker_valid)
            {
              marker_valid = true;
              marker_state = state;
            }
          }
          else
            marker_valid = false;
        }

        if (result_cache[f]->test(loop_state) == check_value)
        {
          if (marker_valid)
          {
            while (marker_state < path_states.size())
            {
              val->flip(marker_state);
              marker_state++;
            }
          }
        }
      }
      else
      {
        if (f->propositional())
        {
          for (StateSpace::size_type state = 0; state < path_states.size();
	       state++)
          {
            if (f->eval(reachability_graph[path_states[state]].
			  positiveAtoms()))
              val->set(state);
          }
        }
        else
        {
          switch (f->what())
          {
            case LTL_NEGATION :
              {
                const LtlFormula* g = static_cast<const Not*>(f)->subformula;

                for (StateSpace::size_type s = 0; s < path_states.size(); s++)
                {
                  if (!result_cache[g]->test(s))
                    val->set(s);
                }

                break;
              }

            case LTL_CONJUNCTION :
              {
                const LtlFormula* g = static_cast<const And*>(f)->subformula1;
                const LtlFormula* h = static_cast<const And*>(f)->subformula2;

                for (StateSpace::size_type s = 0; s < path_states.size(); s++)
                {
                  if (result_cache[g]->test(s) && result_cache[h]->test(s))
                    val->set(s);
                }

                break;
              }

            case LTL_DISJUNCTION :
              {
                const LtlFormula* g = static_cast<const Or*>(f)->subformula1;
                const LtlFormula* h = static_cast<const Or*>(f)->subformula2;

                for (StateSpace::size_type s = 0; s < path_states.size(); s++)
                {
                  if (result_cache[g]->test(s) || result_cache[h]->test(s))
                    val->set(s);
                }

                break;
              }

	    case LTL_IMPLICATION :
	      {
		const LtlFormula* g
		  = static_cast<const Imply*>(f)->subformula1;
	        const LtlFormula* h
		  = static_cast<const Imply*>(f)->subformula2;

		for (StateSpace::size_type s = 0; s < path_states.size(); s++)
		{
		  if (!result_cache[g]->test(s) || result_cache[h]->test(s))
		    val->set(s);
		}
		break;
	      }

	    case LTL_EQUIVALENCE :
	      {
		const LtlFormula* g
		  = static_cast<const Equiv*>(f)->subformula1;
		const LtlFormula* h
		  = static_cast<const Equiv*>(f)->subformula2;

		for (StateSpace::size_type s = 0; s < path_states.size(); s++)
		{
		  if (result_cache[g]->test(s) == result_cache[h]->test(s))
		    val->set(s);
		}
		break;
	      }

            case LTL_NEXT :
              {
                const LtlFormula* g = static_cast<const Next*>(f)->subformula;
		StateSpace::size_type s;

                for (s = 0; (s + 1) < path_states.size(); s++)
                {
                  if (result_cache[g]->test(s + 1))
                    val->set(s);
                }

		if (result_cache[g]->test(loop_state))
		  val->set(s);

                break;
              }

            default :
              break;
          }
        }
      }
    }
  }
}

/* ========================================================================= */
void LtlFormula::prettyPrintPathProof
  (Exceptional_ostream& estream, int indent, bool prove_true,
   StateSpace::size_type state,
   map<const LtlFormula*, Bitset*, ptr_less, ALLOC(Bitset*) >& result_cache,
   const vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >& path,
   StateSpace::size_type loop_state)
  const
/* ----------------------------------------------------------------------------
 *
 * Description:
 *
 * Arguments:
 *
 * Returns:
 *
 * ------------------------------------------------------------------------- */
{
  static string line_prefix = "";

  estream << string(indent, ' ')
             + (line_prefix.size() == 0
		? ""
		: line_prefix.substr(0, line_prefix.size() - 4) + "+-> ")
             + "M,<s"
	  << path[state]
          << string(", ...> |") + (prove_true ? '=' : '/') + "= "
             + formula_string;

  switch (what())
  {
    case LTL_ATOM :
    case LTL_TRUE :
    case LTL_FALSE :
      estream << '\n';
      break;

    case LTL_NEGATION :
      /*
       *  To display a proof (refutation) for a negated formula, display a 
       *  refutation (proof) for the non-negated formula.
       */

      line_prefix += "    ";

      estream << " :\n";
      static_cast<const Not*>(this)->subformula->
        prettyPrintPathProof(estream, indent, !prove_true, state,
			     result_cache, path, loop_state);

      line_prefix.resize(line_prefix.size() - 4);
      break;

    case LTL_NEXT :
      {
	/*
	 *  To display a proof or refutation for a Next formula, display a
	 *  proof or refutation for the subformula of the Next operator in
	 *  the next state of the path.
	 */

	estream << " :\n" + string(indent, ' ') + line_prefix + "+-> s"
	        << path[state]
	        << " --> " << 's';

	state++;
	if (state == path.size())
	  state = loop_state;

	estream << path[state] << '\n';

	line_prefix += "    ";

	static_cast<const Next*>(this)->subformula->
	  prettyPrintPathProof(estream, indent, prove_true, state,
			       result_cache, path, loop_state);

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_CONJUNCTION :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1 = static_cast<const And*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const And*>(this)->subformula2;

	if (prove_true)
        {
	  /*
	   *  To display a proof for a conjunctive formula, display the proofs
	   *  for its subformulae.
	   */

	  f1->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);
	}
	else
        {
	  line_prefix[line_prefix.size() - 4] = ' ';

	  /*
	   *  To display a refutation for a conjunctive formula, display a
	   *  refutation for one of its false subformulae.
	   */

	  if (result_cache[f2]->test(state)
	      || (!result_cache[f1]->test(state)
		  && (f1->propositional() && !f2->propositional()
		      || (f1->constant() && !f2->constant())
		      || (f1->formula_string.length()
			  <= f2->formula_string.length()))))
	    f1->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	  else
	    f2->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_DISJUNCTION :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1 = static_cast<const Or*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const Or*>(this)->subformula2;

	if (prove_true)
	{
	  line_prefix[line_prefix.size() - 4] = ' ';

	  /*
	   *  To display a proof for a disjunctive formula, display a proof
	   *  for one of its true subformulae.
	   */

	  if (!result_cache[f2]->test(state)
	      || (result_cache[f1]->test(state)
		  && ((f1->propositional() && !f2->propositional())
		      || (f1->constant() && !f2->constant())
		      || (f1->formula_string.length()
			  <= f2->formula_string.length()))))
	    f1->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	  else
	    f2->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	}
	else
	{
	  /*
	   *  To display a refutation for a disjunctive formula, display
	   *  refutations for both of its subformulae.
	   */

	  f1->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_IMPLICATION :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1 = static_cast<const Imply*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const Imply*>(this)->subformula2;

	if (prove_true)
	{
	  line_prefix[line_prefix.size() - 4] = ' ';

	  /*
	   *  To display a proof for an implication, display either a
	   *  refutation for `f1' or a proof for `f2'.
	   */

	  if (!result_cache[f2]->test(state)
	      || (!result_cache[f1]->test(state)
		  && ((f1->propositional() && !f2->propositional())
		      || (f1->constant() && !f2->constant())
		      || (f1->formula_string.length()
			  <= f2->formula_string.length()))))
	    f1->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	  else
	    f2->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	}
	else
	{
	  /*
	   *  To display a refutation for an implication, display a proof for
	   *  `f1' and a refutation for `f2'.
	   */

	  f1->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_EQUIVALENCE :
      {
	estream << " :\n";

	line_prefix += "|   ";

	/*
	 *  To display a proof or a refutation for an equivalence, display
	 *  either a proof or a refutation for both subformulae (a proof for
	 *  the equivalence) or a proof for the other subformula and a
	 *  refutation for the other (a refutation for the equivalence).
	 */

	LtlFormula* f1 = static_cast<const Equiv*>(this)->subformula1;
	LtlFormula* f2 = static_cast<const Equiv*>(this)->subformula2;

	bool f1_proof_style, f2_proof_style;

	if (prove_true)
	  f1_proof_style = f2_proof_style = result_cache[f1]->test(state);
	else
	{
	  f1_proof_style = result_cache[f1]->test(state);
	  f2_proof_style = !f1_proof_style;
	}
	
	f1->prettyPrintPathProof(estream, indent, f1_proof_style, state,
				 result_cache, path, loop_state);

	line_prefix[line_prefix.size() - 4] = ' ';

	f2->prettyPrintPathProof(estream, indent, f2_proof_style, state,
				 result_cache, path, loop_state);

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_UNTIL :
    case LTL_FINALLY :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1;
	LtlFormula* f2;

	if (what() == LTL_UNTIL)
	{
	  f1 = static_cast<const Until*>(this)->subformula1;
	  f2 = static_cast<const Until*>(this)->subformula2;
	}
	else
	  f2 = static_cast<const Finally*>(this)->subformula;

	if (prove_true)
	{
	  /*
	   *  To display a proof for an Until or a Finally formula, scan
	   *  through the states of the path until a state where `f2' is true
	   *  is found and then display its proof. In case of an Until formula,
	   *  display also the proof of `f1' in every visited state where `f2'
	   *  does not hold.
	   */

	  while (!result_cache[f2]->test(state))
	  {
	    if (what() == LTL_UNTIL)
	      f1->prettyPrintPathProof(estream, indent, true, state,
				       result_cache, path, loop_state);

	    estream << string(indent, ' ')
	               + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	      state = loop_state;

	    estream << " --> " << 's' << path[state] << '\n';
	  }
	  
	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, true, state,
				   result_cache, path, loop_state);
	}
	else
	{
	  /*
	   *  To display a refutation for an Until formula, scan through the
	   *  states of the path until a state where `f1' and `f2' are both
	   *  false is found or until an already visited state is found.
	   *  Unless in an already visited state at the end, display
	   *  refutation proofs for `f1' and `f2'. In each visited state where
	   *  `f1' holds but `f2' doesn't, display a refutation for `f2'.
	   *
	   *  Displaying a refutation for a Finally formula is equivalent to
	   *  the Until case where the search ends in an already visited state.
	   */

	  StateSpace::size_type start_state = state;

	  while (!result_cache[f2]->test(state)
		 && (what() == LTL_FINALLY || result_cache[f1]->test(state)))
	  {
	    f2->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);

	    estream << string(indent, ' ')
		       + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	    {
	      state = loop_state;
	      if (start_state < loop_state)
	      {
		estream << " --> " << 's' << path[state] << '\n';
		break;
	      }
	    }

	    estream << " --> " << 's' << path[state] << '\n';

	    if (state == start_state)
	      break;
	  }

	  if (what() == LTL_UNTIL && !result_cache[f1]->test(state))
	  {
	    f1->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);

	    line_prefix[line_prefix.size() - 4] = ' ';

	    f2->prettyPrintPathProof(estream, indent, false, state,
				     result_cache, path, loop_state);
	  }
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;

    case LTL_V :
    case LTL_GLOBALLY :
      {
	estream << " :\n";

	line_prefix += "|   ";

	LtlFormula* f1;
	LtlFormula* f2;

	if (what() == LTL_V)
	{
	  f1 = static_cast<const V*>(this)->subformula1;
	  f2 = static_cast<const V*>(this)->subformula2;
	}
	else
	  f2 = static_cast<const Globally*>(this)->subformula;

	if (prove_true)
	{
	  /*
	   *  To display a proof for a Release formula, scan through the states
	   *  of the path until a state where `f1' and `f2' are both true is
	   *  found or until an already visited state is found. Unless in an
	   *  already visited state at the end, display the proofs for `f1' and
	   *  `f2'. In each visited state where `f2' holds but `f1' doesn't,
	   *  display the proof for `f2'.
	   *
	   *  Showing a proof for a Globally formula is similar to the Release
	   *  case where the search ends in an already visited state.
	   */

	  StateSpace::size_type start_state = state;

	  while (result_cache[f2]->test(state)
		 && (what() == LTL_GLOBALLY || !result_cache[f1]->test(state)))
	  {
	    f2->prettyPrintPathProof(estream, indent, true,
				     state, result_cache, path, loop_state);

	    estream << string(indent, ' ')
                       + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	    {
	      state = loop_state;
	      if (start_state < loop_state)
	      {
		estream << " --> " << 's' << path[state] << '\n';
		break;
	      }
	    }

	    estream << " --> " << 's' << path[state] << '\n';

	    if (state == start_state)
	      break;
	  }

	  if (what() == LTL_V && result_cache[f1]->test(state))
	  {
	    f1->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);

	    line_prefix[line_prefix.size() - 4] = ' ';
	
	    f2->prettyPrintPathProof(estream, indent, true, state,
				     result_cache, path, loop_state);
	  }
	}
	else
	{
	  /*
	   *  To show a refutation for a Release or a Globally formula, scan
	   *  through the states of the path until a state where `f2' is false 
	   *  is found and then display a refutation for it. In case of a
	   *  Release formula, display also a refutation for `f1' in every
	   *  visited state where `f2' holds.
	   */

	  while (result_cache[f2]->test(state))
	  {
	    if (what() == LTL_V)
	      f1->prettyPrintPathProof(estream, indent, false, state,
				       result_cache, path, loop_state);

	    estream << string(indent, ' ')
		       + line_prefix.substr(0, line_prefix.size() - 4)
                       + "+-> s"
		    << path[state];

	    state++;
	    if (state == path.size())
	      state = loop_state;

	    estream << " --> " << 's' << path[state] << '\n';
	  }

	  line_prefix[line_prefix.size() - 4] = ' ';

	  f2->prettyPrintPathProof(estream, indent, false, state,
				   result_cache, path, loop_state);
	}

	line_prefix.resize(line_prefix.size() - 4);
      }
      break;
  }
}



/******************************************************************************
 *
 * Function definitions for class AtomToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void AtomToLtl::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info,
   const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a new LtlAtom object from the
 *                printable representation of the atom. Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to a LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to LTL formulae.
 *                          The new propositional atom will be pushed on this
 *                          stack.
 *                atom  --  A propositional atom (a string of the form
 *                          "p<n>", where <n> is an unsigned long integer).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info.data.push(new Atom(strtoul(atom.c_str() + 1, 0, 10)));
}



/******************************************************************************
 *
 * Function definitions for class ConstantToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void ConstantToLtl<T>::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a new propositional constant. Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to an LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to LTL formulae.
 *                          The propositional constant object will be pushed on
 *                          this stack.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info.data.push(new T);
}



/******************************************************************************
 *
 * Function definitions for template class UnaryOperatorToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void UnaryOperatorToLtl<T>::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating an LTL formula with a unary operator.
 *                Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to a LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to LTL formulae,
 *                          storing the partial results of the traversal.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  LtlFormula *f = info.data.top();
  info.data.top() = new T(f);
}



/******************************************************************************
 *
 * Function definitions for template class BinaryOperatorToLtl.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void BinaryOperatorToLtl<T>::eval
  (FormulaString::TraversalInfo<stack<LtlFormula*,
                                      deque<LtlFormula*,
                                            ALLOC(LtlFormula*) > > >&
     info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating an LTL formula with a binary operator.
 *                Used by
 *                LtlFormula::formulaStringToLtlFormula(FormulaString& f) when
 *                translating a FormulaString to a LtlFormula object (during
 *                the traversal of the FormulaString object).
 *
 * Argument:      info  --  A reference to a stack of pointers to propositional
 *                          formulae holding the partial results of the
 *                          traversal.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  LtlFormula* f2 = info.data.top();
  info.data.pop();
  LtlFormula* f1 = info.data.top();
  info.data.top() = new T(f1, f2);
}



/* ========================================================================= */
LtlFormula* formulaStringToLtlFormula(FormulaString& f)
/* ----------------------------------------------------------------------------
 *
 * Description:    Creates an LtlFormula from a FormulaString.
 *
 * Argument:       f  --  A reference to a FormulaString object.
 *
 * Returns:        A pointer to a new LtlFormula object constructed from the
 *                 FormulaString.
 *
 * --------------------------------------------------------------------------*/
{
  FormulaString::TraversalInfo<stack<LtlFormula*,
                                     deque<LtlFormula*,
                                           ALLOC(LtlFormula*) > > >
    info;

  /*
   * Construct the LTL formula by `traversing' the FormulaString, using a stack
   * of pointers to LTL formulae to maintain the state of the construction
   * process. If the construction succeeds, the stack will contain a pointer to
   * the desired LtlFormula object. In the case an exception is thrown during
   * the traversal, the stack is emptied and the exception is rethrown.
   */

  try
  {
    f.template traverse<AtomToLtl,
                        ConstantToLtl<True>,
                        ConstantToLtl<False>,
                        UnaryOperatorToLtl<Not>,
                        UnaryOperatorToLtl<Next>,
                        UnaryOperatorToLtl<Finally>,
                        UnaryOperatorToLtl<Globally>,
                        BinaryOperatorToLtl<And>,
                        BinaryOperatorToLtl<Or>,
                        BinaryOperatorToLtl<Imply>,
                        BinaryOperatorToLtl<Equiv>,
                        BinaryOperatorToLtl<Until>,
                        BinaryOperatorToLtl<V> >
      (FormulaString::POSTORDER, info);
  }
  catch (...)
  {
    while (!info.data.empty())
    {
      delete info.data.top();
      info.data.pop();
    }
    throw;
  }

  return info.data.top();
}

}
