/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ProductAutomaton.h"

namespace Graph
{

/******************************************************************************
 *
 * Function definitions for class ProductAutomaton.
 *
 *****************************************************************************/

/* ========================================================================= */
ProductAutomaton::ProductAutomaton() :
  statespace_size(0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class ProductAutomaton. Initializes a
 *                new object for storing the product of a Bchi automaton and a
 *                reachability graph. The product must then be explicitly
 *                initialized by calling the function computeProduct on the
 *                object.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
ProductAutomaton::size_type ProductAutomaton::expand(size_type node_count = 1)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts states to a ProductAutomaton.
 *
 * Argument:      node_count  --  Number of states to be inserted.
 *
 * Returns:       The index of the last state inserted.
 *
 * ------------------------------------------------------------------------- */
{
  nodes.reserve(nodes.size() + node_count);

  while (node_count > 0)
  {
    nodes.push_back(new ProductState());
    node_count--;
  }

  return size() - 1;
}

/* ========================================================================= */
void ProductAutomaton::print(ostream& stream = cout, int indent = 0) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a product automaton to a stream.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (nodes.empty())
    estream << string(indent, ' ') + "The product automaton is empty.\n";
  else
  {
    pair<size_type, unsigned long int> statistics = stats();

    estream << string(indent, ' ') + "The product automaton consists of\n"
               + string(indent + 4, ' ')
            << statistics.first
            << " states and\n" + string(indent + 4, ' ')
            << statistics.second
            << " transitions.\n";

    size_type s = nodes.size();
    for (size_type state = 0; state < s; ++state)
    {
      estream << string(indent, ' ') + "State " << state << ":\n";
      operator[](state).print(stream, indent + 4);
    }
  }

  estream.flush();
}

/* ========================================================================= */
void ProductAutomaton::computeProduct
  (const BuchiAutomaton& automaton, const StateSpace& statespace,
   bool global_product)
/* ----------------------------------------------------------------------------
 *
 * Description:   Initializes the synchronous product of a Bchi automaton and
 *                a reachability graph.
 *
 * Arguments:     automaton       --  A reference to a constant BuchiAutomaton.
 *                statespace      --  A reference to a constant StateSpace.
 *                global_product  --  Controls whether the synchronous product
 *                                    of the automaton and the statespace is
 *                                    computed `globally' (i.e. with respect to
 *                                    all states in the reachability graph) or
 *                                    `locally' (i.e. with respect only to the
 *                                    initial state of the reachability graph).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  clear();

  statespace_size = statespace.size();

  /*
   *  If either the Bchi automaton or the state space is empty, their product
   *  is also empty.
   */

  if (automaton.empty() || statespace.empty())
    return;

  /*
   *  Pairs of Bchi automaton and system states are encoded by computing
   *  the product of the state identifiers from the following equation:
   *    <pair id> = <Bchi state id> * statespace.size() + <system state id>
   *  The pair identifiers are then mapped to the states of the product
   *  automaton.
   */

  map<size_type, size_type, less<size_type>, ALLOC(size_type) >
    product_state_mapping;

  /*
   *  Initialize the product automaton. If the product is to be computed
   *  globally, initalize the result with state pairs (q_0, s) where q_0 is the
   *  initial state of the Bchi automaton and s goes through all states in the
   *  state space. (These states will have the lowest indices in the product
   *  automaton, i.e. covering the index interval
   *  0 ... [statespace.size() - 1] .) In the case of a local product,
   *  initialize the result with the single state (q_0, s_0), a pair consisting
   *  of the initial states of the two other structures.
   *
   *  The final product automaton is obtained by computing the closure of this
   *  set of states under the product transition relation. For this purpose,
   *  a depth-first search will be used (with `mapping_stack' as the search
   *  stack; initially, it contains the same states as described above).
   */

  expand(global_product ? statespace.size() : 1);

  stack<size_type, deque<size_type, ALLOC(size_type) > > mapping_stack;

  pair<size_type, size_type> state_map_entry
    = make_pair(automaton.initialState() * statespace_size, 0);

  for (state_map_entry.second = 0;
       state_map_entry.second < (global_product ? statespace_size : 1);
       ++state_map_entry.first, ++state_map_entry.second)
  {
    product_state_mapping.insert(state_map_entry);
    mapping_stack.push(state_map_entry.first);
  }

  /*
   *  Compute the product automaton by using a depth-first search.
   */

  const GraphEdgeContainer* automaton_transitions;

  BuchiAutomaton::size_type current_automaton_state;
  StateSpace::size_type current_system_state;
  ProductAutomaton::size_type current_product_state;

  while (!mapping_stack.empty())
  {
    if (::user_break)
      throw UserBreakException();

    /*
     *  Pop a state mapping off the stack.
     */

    current_automaton_state = mapping_stack.top() / statespace_size;
    current_system_state = mapping_stack.top() % statespace_size;
    current_product_state = product_state_mapping[mapping_stack.top()];
    mapping_stack.pop();

    /*
     *  Store information about the product state:
     *     - The number of the associated system state.
     *
     *     - Information about whether the associated automaton state is an
     *       accepting state.
     */

    operator[](current_product_state).systemState()
      = current_system_state;
    operator[](current_product_state).accepting() =
      automaton[current_automaton_state].accepting();

    /*
     *  Go through the transitions of the original Bchi automaton. For all
     *  transitions enabled in the current state of the state space:
     *
     *     - Compute all product states that can be reached by firing the
     *       transition. These are the states (q', s') where q' is the
     *       state of the automaton after firing the transition and s'
     *       is a successor of the current state of the system.
     *
     *     - If there is no corresponding product state for the state pair
     *       (q', s'), create a new product state and insert it into the
     *       product automaton. Push the mapping also on the stack so that
     *       the newly created product state will be eventually processed.
     *
     *     - Connect the current state to the product state. Also, update
     *       the predecesessor information of the product state by adding
     *       the current system state to the predecessor list of the
     *       product state (this information is required later for performing
     *       a backward search in the product automaton).
     */

    automaton_transitions = &automaton[current_automaton_state].edges();

    for (GraphEdgeContainer::const_iterator
           automaton_transition = automaton_transitions->begin();
         automaton_transition != automaton_transitions->end();
         ++automaton_transition)
    {
      if ((static_cast<const BuchiAutomaton::BuchiTransition*>
           (*automaton_transition))->
            enabled(statespace[current_system_state].positiveAtoms()))
      {
        pair<map<size_type, size_type, less<size_type>, ALLOC(size_type) >
               ::iterator,
	     bool>
          check_visited;
        
        const GraphEdgeContainer& system_transitions =
          statespace[current_system_state].edges();

        for (GraphEdgeContainer::const_iterator
               system_transition = system_transitions.begin();
             system_transition != system_transitions.end();
             ++system_transition)
        {
          state_map_entry.first =
            (*automaton_transition)->targetNode() * statespace_size
            + (*system_transition)->targetNode();

          check_visited = product_state_mapping.insert(state_map_entry);

          if (check_visited.second)   /* insertion occurred */
          {
            size_type new_product_state = expand();

            (check_visited.first)->second = new_product_state;
            mapping_stack.push(state_map_entry.first);

            operator[](current_product_state).connectTo(new_product_state);
            operator[](new_product_state).predecessors().
              push_back(current_product_state);
          }
          else
          {
            size_type existing_state = (check_visited.first)->second;

            operator[](current_product_state).connectTo(existing_state);
            operator[](existing_state).predecessors().
              push_back(current_product_state);
          }
        }
      }
    }
  }
}

/* ========================================================================= */
void ProductAutomaton::emptinessCheck(Bitset& result) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Performs an emptiness check on the product automaton, i.e.
 *                finds the set of system states starting an execution
 *                sequence accepted by the Bchi automaton.
 *
 * Argument:      result  --  A reference to the Bitset in which the result
 *                            is stored. The Bitset must have room for as many
 *                            states as there are in the reachability graph.
 *                            A `0' bit in some position n in the result means
 *                            that no accepting executions begin from system
 *                            state n; a `1' bit means the opposite.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Bitset visited(nodes.size());

  result.clear();

  for (SccIterator<GraphEdgeContainer, ProductScc>
	 strongly_connected_component(*this);
       !strongly_connected_component.atEnd();
       ++strongly_connected_component)
  {
    if (::user_break)
      throw UserBreakException();

    if (!strongly_connected_component->isTrivialIn(*this))
    {
      /*
       *  Scan the strongly connected component for a product state whose
       *  associated Bchi automaton state is accepting.
       */

      ProductScc::const_iterator st;
      for (st = strongly_connected_component->begin();
           st != strongly_connected_component->end()
             && !operator[](*st).accepting();
           ++st)
        ;

      /*
       *  Starting from an `accepting' product state, perform a backward depth-
       *  first search to find all states (q_0, s) such that q_0 is the initial
       *  state of the automaton (the product states (q_0, s) are recognized
       *  by the product state numbering scheme, in which these states have the
       *  lowest indices). Add the corresponding system states to the result.
       */

      if (st != strongly_connected_component->end() && !visited[*st])
      {
        size_type state = *st;
        stack<size_type, deque<size_type, ALLOC(size_type) > >
	  backward_search_stack;
        vector<size_type, ALLOC(size_type) >* predecessors;

        visited.set(state);
        backward_search_stack.push(state);
          
        while (!backward_search_stack.empty())
        {
          state = backward_search_stack.top();
          backward_search_stack.pop();

          if (state < result.size())
            result.set(operator[](state).systemState());

          predecessors = &(operator[](state).predecessors());
          
          for (vector<size_type, ALLOC(size_type) >::const_iterator
		 predecessor = predecessors->begin();
               predecessor != predecessors->end();
               ++predecessor)
          {
            if (!visited[*predecessor])
            {
              backward_search_stack.push(*predecessor);
              visited.set(*predecessor);
            } 
          }
        }
      }
    }
  }
}

/* ========================================================================= */
void ProductAutomaton::findAcceptingExecution
  (StateSpace::size_type initial_state,
   pair<vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >,
        vector<StateSpace::size_type, ALLOC(StateSpace::size_type) > >&
     execution) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Extracts a system execution (beginning from a given system
 *                state) accepted by the Bchi automaton from the product
 *                space. The function behaves basically like
 *                ProductAutomaton::emptinessCheck, but it keeps additional
 *                information about the path of processed states during the
 *                backward search. This information is then used to extract
 *                the desired system execution from the graph.
 *
 * Arguments:     initial_state  --  Identifier of the system state.
 *                execution      --  A reference to a pair of vectors for
 *                                   storing the result. The `first' component
 *                                   of the pair represents an execution
 *                                   `prefix' (a sequence of system state
 *                                   identifiers) leading from `initial_state'
 *                                   to an accepting cycle. The `second'
 *                                   component of the pair contains the cycle
 *                                   itself.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Bitset visited(nodes.size());
  execution.first.clear();
  
  for (SccIterator<GraphEdgeContainer, ProductScc>
	 strongly_connected_component(*this);
       !strongly_connected_component.atEnd();
       ++strongly_connected_component)
  {
    if (!strongly_connected_component->isTrivialIn(*this))
    {
      /*
       *  Scan the strongly connected component for a product state whose
       *  associated Bchi automaton state is accepting.
       */

      ProductScc::const_iterator st;
      for (st = strongly_connected_component->begin();
           st != strongly_connected_component->end()
	     && !operator[](*st).accepting();
           ++st)
        ;

      if (st != strongly_connected_component->end() && !visited[*st])
      {
        size_type state = *st;
        stack<size_type, deque<size_type, ALLOC(size_type) > >
	  backward_search_stack;
        vector<size_type, ALLOC(size_type) >* predecessors;

        /*
         *  Variables to store information about the state path in the product
         *  space.
         */

	typedef set<StateSpace::size_type, ALLOC(StateSpace::size_type) >
	  ::size_type
	  PathStackElement;

        vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >
	  current_path;
        stack<PathStackElement, deque<PathStackElement,
                                      ALLOC(PathStackElement) > >
	  path_stack;
        size_type search_start_state = state;
        
        visited.set(state);
        backward_search_stack.push(state);

        execution.second.clear();

        while (!backward_search_stack.empty())
        {
          state = backward_search_stack.top();
          backward_search_stack.pop();

          if (!path_stack.empty())
            path_stack.top()--;

          current_path.push_back(operator[](state).systemState());

          /*
           *  Test whether `state' corresponds to the given initial state. If
           *  it does, extract the system execution from that state to the
           *  state from which the backward search was started.
           */

          if (execution.first.empty()
	      && state < statespace_size
              && operator[](state).systemState() == initial_state)
          {
            execution.first = 
              vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >
                (current_path.rbegin(), current_path.rend());
          }
    
          predecessors = &(operator[](state).predecessors());

          path_stack.push(0);

          for (vector<size_type, ALLOC(size_type) >::const_iterator
		 predecessor = predecessors->begin();
               predecessor != predecessors->end();
               ++predecessor)
          {
            if (!visited[*predecessor])
            {
              backward_search_stack.push(*predecessor);
              visited.set(*predecessor);
              path_stack.top()++;
            }

            /*
             *  If the state from where the backward search was started is a
             *  predecessor of the current state, we have found a cycle in the
             *  strongly connected component. Extract the shortest cycle from
             *  the maximal strongly connected component.
             */

            else if (*predecessor == search_start_state
                     && (execution.second.empty()
                         || current_path.size() < execution.second.size()))
            {
              execution.second =
                vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >
		  (current_path.rbegin(), current_path.rend());
            }
          }

          while (!path_stack.empty() && path_stack.top() == 0)
          {
            path_stack.pop();
            current_path.pop_back();
          }
        }
        if (!execution.first.empty())
          return;
      }
    }
  }

  /*
   *  If no accepting execution beginning from the given state could be found,
   *  clear the result.
   */

  execution.first.clear();
  execution.second.clear();
}



/******************************************************************************
 *
 * Function definitions for class ProductAutomaton::ProductState.
 *
 *****************************************************************************/

/* ========================================================================= */
void ProductAutomaton::ProductState::print
  (ostream& stream = cout, int indent = 0) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about the ProductState to a stream.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << string(indent, ' ') + "Associated system state: "
          << system_state 
          << '\n' + string(indent, ' ') + "The associated automaton state is "
             + (accepting() ? "" : "not ") + "an accepting state.\n";

  if (predecessor_states.empty())
    estream << string(indent,' ') + "The product state has no predecessors.\n";
  else
  {
    bool first_printed = false; 
    estream << string(indent, ' ') + "Predecessor states: {";

    for (vector<size_type, ALLOC(size_type) >::const_iterator predecessor
           = predecessor_states.begin();
         predecessor != predecessor_states.end();
         ++predecessor)
    {
      if (first_printed)
        estream << ", ";
      else
        first_printed = true;
      estream << *predecessor;
    }
    estream << "}\n";
  }

  if (edges().empty())
    estream << string(indent, ' ') + "The product state has no successors.\n";
  else
  {
    bool first_printed = false;
    estream << string(indent, ' ') + "Successor states: {";

    GraphEdgeContainer::const_iterator edge;

    for (edge = edges().begin(); edge != edges().end(); ++edge)
    {
      if (first_printed)
        estream << ", ";
      else
        first_printed = true;
      estream << (*edge)->targetNode();
    }
    estream << "}\n";
  }
  estream.flush();
}



/******************************************************************************
 *
 * Function definitions for class ProductAutomaton::ProductScc.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool ProductAutomaton::ProductScc::isTrivialIn
  (const ProductAutomaton& product_automaton) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests whether a strongly connected component is trivial in
 *                a product automaton. A strongly connected component is
 *                trivial if and only if it is empty or it contains a single
 *                state with no transition to itself.
 *
 * Arguments:     product_automaton  --  A constant reference to a
 *                                       ProductAutomaton.
 *
 * Returns:       A truth value telling whether the strongly connected
 *                component is trivial.
 *
 * ------------------------------------------------------------------------- */
{
  if (empty())
    return true;

  if (size() > 1)
    return false;

  size_type s = *begin();

  return (!product_automaton[s].connected(s));
}

}
