/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _ProductAutomaton_h_
#define _ProductAutomaton_h_

#include <deque>
#include <iostream>
#include <stack>
#include <vector>
#include "BuchiAutomaton.h"
#include "EdgeContainer.h"
#include "Exception.h"
#include "Graph.h"
#include "SccIterator.h"
#include "StateSpace.h"

using namespace std;

extern bool user_break;

namespace Graph 
{

/******************************************************************************
 *
 * A class for representing the synchronous product of a Bchi automaton and
 * a reachability graph.
 *
 *****************************************************************************/

class ProductAutomaton : public Graph<GraphEdgeContainer>
{
private:
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class ProductState :                              /* A class for           */
    public Graph<GraphEdgeContainer>::Node          /* representing the      */
  {                                                 /* states of the product
                                                     * automaton.
                                                     */
  public:
    explicit ProductState                           /* Constructor. */
      (size_type state_number = 0,
       bool accepting = false);

    /* default copy constructor */

    ~ProductState();                                /* Destructor. */
  
    /* default assignment operator */
  
    ProductState* clone() const;                    /* Creates an identical
                                                     * copy of the state.
                                                     */
    
    /* `edges' inherited from Graph<EdgeContainer>::Node */
    
    /* `connectTo' inherited from Graph<EdgeContainer>::Node */
    
    /* `disconnectFrom' inherited from Graph<EdgeContainer>::Node */
    
    /* `connected' inherited from Graph<EdgeContainer>::Node */

    size_type systemState() const;                  /* Get or set the system */
    size_type& systemState();                       /* state with which the
                                                     * product state is
                                                     * associated.
                                                     */

    bool accepting() const;                         /* Get or set the       */
    bool& accepting();                              /* `acceptance status'
                                                     * of the product state
                                                     * (i.e. whether the
                                                     * associated Bchi
                                                     * state is accepting).
                                                     */

    const vector<size_type, ALLOC(size_type) >&     /* Get or set the        */
      predecessors() const;                         /* predecessor states of */
    vector<size_type, ALLOC(size_type) >&           /* the product state.    */
      predecessors();    
    
    void print                                       /* Writes information */
      (ostream& stream = cout,                       /* about the product  */
       int indent = 0) const;                        /* state to a stream. */

  private:
    size_type system_state;                         /* Identifier of the
                                                     * system state with which
                                                     * the product state is
                                                     * associated.
                                                     */

    bool accept;                                    /* Acceptance status of
                                                     * the associated Bchi
                                                     * automaton state.
                                                     */

    vector<size_type, ALLOC(size_type) >            /* List of the product   */
      predecessor_states;                           /* state's predecessors.
                                                     * A vector is used
						     * instead of a set in
						     * order to reduce
                                                     * memory usage.
                                                     */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  class ProductScc :                                /* A class for storing  */
    public vector<ProductAutomaton::size_type,      /* maximal strongly     */
               ALLOC(ProductAutomaton::size_type) > /* connected components */
  {                                                 /* of the product.
						     */
  public:                        
    ProductScc();                                   /* Constructor. */

    /* default copy constructor */

    ~ProductScc();                                  /* Destructor. */

    /* default assignment operator */

    bool isTrivialIn                                /* Tests whether the */
      (const ProductAutomaton& product_automaton)   /* component is      */
      const;                                        /* trivial.          */
    
    void insert                                     /* Inserts a state into */
      (ProductAutomaton::size_type product_state);  /* the container.       */
  };

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

public:
  ProductAutomaton();                               /* Constructor. */

  /* default copy constructor */

  ~ProductAutomaton();                              /* Destructor. */

  /* default assignment operator */

  ProductState& operator[](size_type index) const;  /* Indexing operator. No
						     * range check is performed
						     * on the argument.
						     */

  ProductState& node(size_type index) const;        /* Synonym for the indexing
                                                     * operator. This function
						     * also checks the range of
						     * the argument.
                                                     */

  /* `size' inherited from Graph<EdgeContainer> */

  /* `empty' inherited from Graph<EdgeContainer> */

  /* `clear' inherited from Graph<EdgeContainer> */

  size_type expand(size_type node_count = 1);       /* Inserts states to the
						     * product automaton.
						     */

  /* `connect' inherited from Graph<EdgeContainer> */

  /* `disconnect' inherited from Graph<EdgeContainer> */

  /* `connected' inherited from Graph<EdgeContainer> */

  /* `stats' inherited from Graph<EdgeContainer> */

  /* `subgraphStats' inherited from Graph */

  void computeProduct                               /* Function for       */
    (const BuchiAutomaton& automaton,               /* initializing the   */
     const StateSpace& statespace,                  /* product automaton. */
     bool global_product);

  void emptinessCheck(Bitset& result) const;        /* Performs an emptiness
                                                     * check on the product.
                                                     */

  void findAcceptingExecution                       /* Extracts an accepting */
    (StateSpace::size_type initial_state,           /* system execution from */
     pair<vector<StateSpace::size_type,             /* the product           */
                 ALLOC(StateSpace::size_type) >,    /* automaton.            */
          vector<StateSpace::size_type,
                 ALLOC(StateSpace::size_type) > >&          
       execution) const;

  void print                                        /* Writes information   */
    (ostream& stream = cout,                        /* about the product    */
     int indent = 0) const;                         /* automaton to a
						     * stream.
                                                     */

private:
  StateSpace::size_type statespace_size;            /* Size of the state space
                                                     * used for constructing
                                                     * the product automaton.
                                                     */
};



/******************************************************************************
 *
 * Inline function definitions for class ProductAutomaton.
 *
 *****************************************************************************/

/* ========================================================================= */
inline ProductAutomaton::~ProductAutomaton()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class ProductAutomaton.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline ProductAutomaton::ProductState&
ProductAutomaton::operator[](size_type index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Indexing operator for class ProductAutomaton. Can be used to
 *                refer to the individual states of the product automaton. No
 *                range check will be performed on the argument.
 *
 * Argument:      index  --  Index of a state of the product automaton.
 *
 * Returns:       A reference to the product state corresponding to the index.
 *
 * ------------------------------------------------------------------------- */
{
  return static_cast<ProductState&>(*nodes[index]);
}

/* ========================================================================= */
inline ProductAutomaton::ProductState&
ProductAutomaton::node(size_type index) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for referring to a single state of a
 *                ProductAutomaton. This function will also check the range
 *                argument.
 *
 * Argument:      index  --  Index of a state of the product automaton.
 *
 * Returns:       A reference to the corresponding product state.
 *
 * ------------------------------------------------------------------------- */
{
  return static_cast<ProductState&>(Graph<GraphEdgeContainer>::node(index));
}



/******************************************************************************
 *
 * Inline function definitions for class ProductAutomaton::ProductState.
 *
 *****************************************************************************/

/* ========================================================================= */
inline ProductAutomaton::ProductState::
ProductState(size_type system_state_number = 0, bool accepting = false) :
  Graph<GraphEdgeContainer>::Node(), system_state(system_state_number),
  accept(accepting)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class ProductAutomaton::ProductState. Creates
 *                a new object representing a synchronous product of a state of
 *                a Bchi automaton with a state of a reachability graph.
 *
 * Arguments:     system_state_number  --  Identifier for the state of the
 *                                         reachability graph.
 *                accepting            --  True if the state of the Bchi
 *                                         automaton is an accepting state.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline ProductAutomaton::ProductState::~ProductState()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class ProductAutomaton::ProductState.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  predecessor_states.clear();
  predecessor_states.reserve(0);
}

/* ========================================================================= */
inline ProductAutomaton::ProductState*
ProductAutomaton::ProductState::clone() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Creates an identical copy of a ProductState.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to the copy.
 *
 * ------------------------------------------------------------------------- */
{
  return new ProductState(*this);
}

/* ========================================================================= */
inline ProductAutomaton::size_type
ProductAutomaton::ProductState::systemState() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the identifier of the system state with which the
 *                product state is associated. This function returns the
 *                identifier by value.
 *
 * Arguments:     None.
 *
 * Returns:       An identifier of a state in the reachability graph.
 *
 * ------------------------------------------------------------------------- */
{
  return system_state;
}

/* ========================================================================= */
inline ProductAutomaton::size_type&
ProductAutomaton::ProductState::systemState()
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the identifier of the system state with which the
 *                product state is associated. This function returns the
 *                identifier by reference and can therefore be used to change
 *                the association.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to an identifier of a state in the reachability
 *                graph.
 *
 * ------------------------------------------------------------------------- */
{
  return system_state;
}
    
/* ========================================================================= */
inline bool ProductAutomaton::ProductState::accepting() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests whether the Bchi automaton state with which the
 *                product state is associated is accepting.
 *
 * Arguments:     None.
 *
 * Returns:       A truth value telling the `acceptance status' of the product
 *                state.
 *
 * ------------------------------------------------------------------------- */
{
  return accept;
}

/* ========================================================================= */
inline bool& ProductAutomaton::ProductState::accepting()
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests whether the Bchi automaton state with which the
 *                product state is associated is accpeting. This function
 *                returns a reference to the information and can therefore be
 *                used to change the acceptance status.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to a truth value telling the `acceptance status'
 *                of the product state.
 *
 * ------------------------------------------------------------------------- */
{
  return accept;
}

/* ========================================================================= */
inline const vector<ProductAutomaton::size_type,
                    ALLOC(ProductAutomaton::size_type) >&
ProductAutomaton::ProductState::predecessors() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the container of identifiers of the product state's
 *                predecessors in the product automaton.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to a constant vector of product state
 *                identifiers.
 *
 * ------------------------------------------------------------------------- */
{
  return predecessor_states;
}

/* ========================================================================= */
inline vector<ProductAutomaton::size_type,
              ALLOC(ProductAutomaton::size_type) >&
ProductAutomaton::ProductState::predecessors()
/* ----------------------------------------------------------------------------
 *
 * Description:   Returns the container of identifiers of the product state's
 *                predecessors in the product automaton. This function can
 *                be also used to change the information.
 *
 * Arguments:     None.
 *
 * Returns:       A reference to a vector of product state identifiers.
 *
 * ------------------------------------------------------------------------- */
{
  return predecessor_states;
}



/******************************************************************************
 *
 * Inline function definitions for class ProductAutomaton::ProductScc.
 *
 *****************************************************************************/

/* ========================================================================= */
inline ProductAutomaton::ProductScc::ProductScc()
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class ProductAutomaton::ProductScc. Creates a
 *                new container for storing a maximal strongly connected
 *                component of a ProductAutomaton.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline ProductAutomaton::ProductScc::~ProductScc()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class ProductAutomaton::ProductScc.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline void ProductAutomaton::ProductScc::insert
  (ProductAutomaton::size_type product_state)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts a new product state identifier to the container.
 *
 * Argument:      product_state  --  Identifier of a product state.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  push_back(product_state);
}

}

#endif /* _ProductAutomaton_h_ */
