/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "StateSpace.h"

namespace Graph 
{

/******************************************************************************
 *
 * Function definitions for class StateSpace.
 *
 *****************************************************************************/

/* ========================================================================= */
StateSpace::StateSpace
  (unsigned long int propositions_per_state = 0,
   size_type initial_number_of_states = 0) :
  atoms_per_state(propositions_per_state), initial_state(0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class StateSpace. Initializes a reachability
 *                graph with a given number of states and with a given number
 *                of atomic propositions per state.
 *
 * Arguments:     propositions_per_state    --  Atomic propositions per state.
 *                initial_number_of_states  --  Initial size of the
 *                                              reachability graph (can be
 *                                              grown later).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  nodes.reserve(initial_number_of_states);
  for (size_type i = 0; i < initial_number_of_states; ++i)
    nodes.push_back(new State(propositions_per_state));
}

/* ======================================================================== */
void StateSpace::randomize
  (size_type number_of_states, unsigned long int number_of_atoms,
   double edge_probability, double truth_probability)
/* ---------------------------------------------------------------------------
 *
 * Description:   Initializes a random reachability graph.
 *
 * Arguments:     number_of_states   --  Number of states in the reachability
 *                                       graph.
 *                number_of_atoms    --  Number of propositional variables per
 *                                       state in the graph.
 *                edge_probability   --  Probability that two states will be
 *                                       connected to each other (does not
 *                                       affect the transitions which are
 *                                       forced between two states to ensure
 *                                       the connectedness of the result).
 *                truth_probability  --  Probability of a proposition being
 *                                       true in a state.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  clear();

  atoms_per_state = number_of_atoms;
  initial_state = 0;

  if (number_of_states == 0)
    return;

  nodes.reserve(number_of_states);
  for (size_type i = 0; i < number_of_states; i++)
    nodes.push_back(new State(atoms_per_state));

  /* Random graph generation algorithm:
     ----------------------------------

        allocate number_of_states nodes;

        insert node 0 to the set of `reachable but unprocessed' nodes (nodes
        that are reachable from the root node but whose children have not yet
        been selected)

        while the set of `reachable but unprocessed' nodes in nonempty 
        {

          select a (random) node X from the set;

          if there exists a node Y that is not yet reachable from root node 
          {
            connect X to Y by making X the parent node of Y;
            insert Y to the set of `reachable but unprocessed' nodes;
          }

          for each node Z (excluding node Y) in the graph 
          {

            randomly connect X to Z (making X the parent of Z);

            if Z was not previously reachable from the root node
              insert Z to the set of `reachable but unprocessed' nodes;
          }

          remove X from the set of `reachable but unprocessed' nodes;
        }

     In the following implementation, the random truth value allocation for
     propositions that hold in a state is interleaved with the previous
     algorithm.
  */

  size_type first_unreachable_state = 1, random_node;
  Bitset* atoms;

  multimap<long int, size_type, less<long int>, ALLOC(size_type) >
    reachable_but_unprocessed;

  reachable_but_unprocessed.insert(make_pair(0, 0));

  while (!reachable_but_unprocessed.empty()) 
  {
    random_node = (*(reachable_but_unprocessed.begin())).second;
    reachable_but_unprocessed.erase(reachable_but_unprocessed.begin());

    for (size_type n = 0; n < first_unreachable_state; ++n)
    {
      if (DRAND < edge_probability)
        nodes[random_node]->connectTo(n);
    }

    if (first_unreachable_state < number_of_states) 
    {
      nodes[random_node]->connectTo(first_unreachable_state);
      reachable_but_unprocessed.insert(make_pair(LRAND,
                                                 first_unreachable_state));
      ++first_unreachable_state;

      for (size_type i = first_unreachable_state; i < number_of_states; ++i) 
      {
        if (DRAND < edge_probability)
        {
          nodes[random_node]->connectTo(first_unreachable_state);
          reachable_but_unprocessed.insert(make_pair(LRAND,
                                                     first_unreachable_state));
          ++first_unreachable_state;
        }
      }
    }

    if (nodes[random_node]->edges().empty())
      nodes[random_node]->connectTo(random_node);

    atoms = &(operator[](random_node).positiveAtoms());
    for (unsigned long int i = 0; i < atoms_per_state; ++i) 
    {
      if (DRAND < truth_probability)
        atoms->set(i);
    }
  }
}

/* ========================================================================= */
void StateSpace::randomizePath
  (size_type number_of_states, unsigned long int number_of_atoms,
   double truth_probability)
/* ----------------------------------------------------------------------------
 *
 * Description:   Generates a random Kripke structure that consists of a single
 *                infinite path (a finite prefix and a loop) of a given size.
 *
 * Arguments:     number_of_states   --  Number of states in the structure.
 *                number_of_atoms    --  Number of propositional variables per
 *                                       state.
 *                truth_probability  --  Probability of a proposition being
 *                                       true in a state.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  clear();

  atoms_per_state = number_of_atoms;
  initial_state = 0;

  if (number_of_states == 0)
    return;

  nodes.reserve(number_of_states);
  for (size_type i = 0; i < number_of_states; i++)
    nodes.push_back(new State(atoms_per_state));

  for (size_type i = 0; i + 1 < number_of_states; i++)
  {
    nodes[i]->connectTo(i + 1);
    for (unsigned long int j = 0; j < atoms_per_state; j++)
    {
      if (DRAND < truth_probability)
        operator[](i).positiveAtoms().set(j);
    }
  }

  nodes[number_of_states - 1]->connectTo(LRAND % number_of_states);
}

/* ========================================================================= */
StateSpace& StateSpace::operator=(const StateSpace& statespace)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class StateSpace. Assigns the
 *                contents of a reachability graph to another one.
 *
 * Argument:      statespace  --  A reference to the constant StateSpace whose
 *                                contents are to be copied.
 *
 * Returns:       A reference to the StateSpace assigned to.
 *
 * ------------------------------------------------------------------------- */
{
  if (&statespace != this) 
  {
    Graph<GraphEdgeContainer>::operator=(statespace);
    atoms_per_state = statespace.atoms_per_state;
    initial_state = statespace.initial_state;
  }

  return *this;
}

/* ========================================================================= */
StateSpace::size_type StateSpace::expand(size_type node_count = 1)
/* ----------------------------------------------------------------------------
 *
 * Description:   Inserts a given number of states to a StateSpace.
 *
 * Argument:      node_count  --  Number of states to be inserted.
 *
 * Returns:       The index of the last inserted state.
 *
 * ------------------------------------------------------------------------- */
{
  nodes.reserve(nodes.size() + node_count);

  while (node_count > 0) 
  {
    nodes.push_back(new State(atoms_per_state));
    node_count--;
  }

  return size() - 1;
}

/* ======================================================================== */
void StateSpace::print
  (ostream& stream = cout, int indent = 0, GraphOutputFormat fmt = NORMAL)
  const
/* ---------------------------------------------------------------------------
 *
 * Description:   Writes information about a StateSpace to a stream.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the output format.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------ */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (fmt == DOT)
    estream << string(indent, ' ') + "digraph G {\n";

  if (nodes.empty())
  {
    if (fmt == NORMAL)
      estream << string(indent, ' ') + "The state space is empty.\n";
  }
  else 
  {
    if (fmt == NORMAL)
    {
      pair<size_type, unsigned long int> statistics = stats();
      pair<size_type, unsigned long int> reachable_part_statistics =
        subgraphStats(initial_state);

      estream << string(indent, ' ') + "The state space consists of\n"
	         + string(indent + 4, ' ')
              << statistics.first
              << " states and\n" + string(indent + 4, ' ')
              << statistics.second
              << " transitions.\n" + string(indent, ' ')
                 + "The reachable part of the state space contains\n"
                 + string(indent + 4, ' ')
              << reachable_part_statistics.first
              << " states and\n" + string(indent + 4, ' ')
              << reachable_part_statistics.second 
              << " transitions.\n" + string(indent, ' ') + "Initial state: "
              << initial_state << '\n';
    }

    size_type s = nodes.size();
    for (size_type state = 0; state < s; ++state) 
    {
      estream << string(indent, ' ');
      if (fmt == NORMAL)
      {
        estream << "State " << state << ":\n";
        operator[](state).print(stream, indent + 4);
      }
      else if (fmt == DOT)
      {
        GraphEdgeContainer::const_iterator transition;

        estream << "  n" << state << " [";
        if (state == 0)
          estream << "style=filled,";
        estream << "shape=ellipse,label=\"" << state << ": ";
        operator[](state).print(stream, 0, fmt);
        estream << "\",fontsize=12];\n";

        for (transition = nodes[state]->edges().begin();
             transition != nodes[state]->edges().end();
             ++transition)
        {
          estream << string(indent + 2, ' ') + 'n' << state;
          (*transition)->print(stream, indent, fmt);
          estream << ";\n";
        }
      }
    }
  }

  if (fmt == DOT)
    estream << string(indent, ' ') + "}\n";

  estream.flush();
}



/*****************************************************************************
 *
 * Function definitions for class StateSpace::State.
 *
 ****************************************************************************/

/* ======================================================================== */
void StateSpace::State::print
  (ostream& stream = cout, int indent = 0, GraphOutputFormat fmt = NORMAL)
  const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about a state of a reachability graph.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *                fmt     --  Determines the format of output.
 * 
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------ */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (fmt == DOT)
    estream << '{';

  if (positive_atoms.count() == 0)
  {
    if (fmt == NORMAL)
      estream << string(indent, ' ') + "No true propositions in this state.\n";
    else if (fmt == DOT)
      estream << '}';
  }
  else 
  {
    bool first_printed = false;

    if (fmt == NORMAL)
      estream << string(indent, ' ')
	         + "Propositions that hold in this state: {";

    for (unsigned long int atom = 0; atom < positive_atoms.size(); ++atom) 
    {
      if (holds(atom)) 
      {
        if (first_printed)
          estream << ", ";
        else
          first_printed = true;
        estream << 'p' << atom;
      }
    }
    estream << '}';
    if (fmt == NORMAL)
      estream << '\n';
  }

  if (fmt == NORMAL)
  {
    if (edges().empty())
      estream << string(indent, ' ') + "No transitions to other states.\n";
    else 
    {
      bool first_printed = false;
      estream << string(indent, ' ') + "Transitions to states {";
    
      for (GraphEdgeContainer::const_iterator edge = edges().begin();
           edge != edges().end();
           ++edge) 
      {
        if (first_printed)
          estream << ", ";
        else
          first_printed = true;
        estream << (*edge)->targetNode();
      }
      estream << "}\n";
    }
  }

  estream.flush();
}

}
