/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "TestOperations.h"

/******************************************************************************
 *
 * Implementations for the operations used in the main test loop.
 *
 *****************************************************************************/

/* ========================================================================= */
void openFile
  (const char* filename, ifstream& stream, ios::openmode mode, int indent = 0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for opening a file for input.
 *
 * Arguments:     filename  --  A pointer to a constant C-style string
 *                              containing the name of the file to be opened.
 *                stream    --  A reference to the input stream which is to be
 *                              associated with the file.
 *                mode      --  An constant of type `ios::openmode' determining
 *                              the open mode.
 *                indent    --  Number of spaces to leave to the left of
 *                              messages given to the user.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity == 5)
  {
    cout << string(indent, ' ') + "<opening `" + filename + "'>";
    cout.flush();
  }

  stream.open(filename, mode);

  if (!stream.good())
  {
    if (configuration.global_options.verbosity == 5)
      cout << " error" << endl;

    throw FileOpenException(filename);
  }
  else if (configuration.global_options.verbosity == 5)
    cout << " ok" << endl;
}

/* ========================================================================= */
void openFile
  (const char* filename, ofstream& stream, ios::openmode mode, int indent = 0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for opening a file for output.
 *
 * Arguments:     filename  --  A pointer to a constant C-style string with
 *                              the name of the file to be opened.
 *                stream    --  A reference to the output stream which is to be
 *                              associated with the file.
 *                mode      --  An constant of type `ios::openmode' determining
 *                              the open mode.
 *                indent    --  Number of spaces to leave to the left of
 *                              messages given to the user.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity == 5)
  {
    cout << string(indent, ' ') + '<'
            + (mode & ios::trunc ? "creat" : "open")
            + "ing `" + filename + "'>";
    cout.flush();
  }

  stream.open(filename, mode);

  if (!stream.good())
  {
    if (configuration.global_options.verbosity == 5)
      cout << " error" << endl;

    if (mode & ios::trunc)
      throw FileCreationException(filename);
    else
      throw FileOpenException(filename);
  }
  else if (configuration.global_options.verbosity == 5)
    cout << " ok" << endl;
}

/* ========================================================================= */
void removeFile(const char* filename, int indent = 0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Removes a file.
 *
 * Arguments:     filename  --  A pointer to a constant C-style string with
 *                              the name of the file to be removed.
 *                indent    --  Number of spaces to leave to the left of
 *                              messages given to the user.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity == 5)
  {
    cout << string(indent, ' ') + "<removing `" + filename + "'>";
    cout.flush();
  }

  if (remove(filename) == 0)
  {
    if (configuration.global_options.verbosity == 5)
      cout << " ok" << endl;
  }
  else if (configuration.global_options.verbosity == 5)
    cout << " error" << endl;
}

/* ========================================================================= */
void printFileContents
  (ostream& stream, const char* message, const char* filename, int indent,
   const char* line_prefix = "")
/* ----------------------------------------------------------------------------
 *
 * Description:   Outputs the contents of a file into a stream.
 *
 * Arguments:     stream       --  A reference to the output stream into which
 *                                 the file contents are to be outputted.
 *                message      --  A pointer to a constant C-style string
 *                                 containing a message to be outputted to the
 *                                 stream before the contents of the file. (To
 *                                 display only the file contents, use a null
 *                                 pointer for the message; an empty string
 *                                 results in an empty line to be outputted
 *                                 before the file contents.)
 *                filename     --  A pointer to a constant C-style string
 *                                 containing the name of the input file.
 *                indent       --  Number of spaces to leave to the left of
 *                                 output.
 *                line_prefix  --  A pointer to a constant C-style string
 *                                 which will be prepended to each outputted
 *                                 line.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  ifstream file;
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  try
  {
    openFile(filename, file, ios::in, indent);

    bool first_line_printed = false;
    string message_line;

    if (configuration.global_options.verbosity == 5)
      cout << string(indent, ' ') + "<reading from `" + filename + "'>"
	   << endl;

    while (file.good())
    {
      message_line = "";

      getline(file, message_line, '\n');

      if (!file.eof())
      {
	if (!first_line_printed && message != 0)
	{
	  first_line_printed = true;
	  estream << string(indent, ' ') + message + '\n';
	}
                    
	estream << string(indent, ' ') + line_prefix + message_line + '\n';
      }
    }

    estream.flush();

    file.close();
  }
  catch (FileOpenException e)
  {
    estream << string(indent, ' ') + e.what() << endl;
  }
  catch (IOException e)
  {
  }
}

/* ========================================================================= */
StateSpace* generateStateSpace()
/* ----------------------------------------------------------------------------
 *
 * Description:   Generates a random state space. The global variable
 *                `configuration.statespace_options.generate_mode' determines
 *                whether a random graph or a random path is generated.
 *
 * Arguments:     None.
 *
 * Returns:       A pointer to a newly allocated StateSpace object.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity >= 2)
  {
    if (configuration.global_options.verbosity == 2)
      cout << "    Generating random state space...\n";
    else
    {
      cout << "    Random state space:\n";
      if (configuration.global_options.verbosity >= 4)
        cout << string(6, ' ') + "<generating>";
    }
    cout.flush();
  }

  /*
   *  Determine the number of states in the state space to be generated. If
   *  the size can be chosen from an interval instead of a single value, choose
   *  the size randomly in the interval.
   */

  StateSpace::size_type size = configuration.statespace_options.size;

  if (configuration.statespace_options.max_size > size)
    size += LRAND % (configuration.statespace_options.max_size + 1 - size);

  /*
   *  Determine the type of the state space to be generated according to the
   *  configuration options, then generate it.
   */

  StateSpace* statespace = new StateSpace();

  if (configuration.statespace_options.generate_mode
        == Configuration::RANDOMGRAPH)
    statespace->randomize(size,
			  configuration.statespace_options.atoms_per_state,
			  configuration.statespace_options.edge_probability,
			  configuration.statespace_options.truth_probability);
  else if (configuration.statespace_options.generate_mode
	     == Configuration::RANDOMPATH)
    statespace->randomizePath(configuration.statespace_options.size,
			      configuration.statespace_options.
			        atoms_per_state,
			      configuration.statespace_options.
			        truth_probability);

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;

  if (configuration.global_options.verbosity >= 3)
  {
    cout << string(6, ' ') + "number of states:  " << size
         << '\n' + string(6, ' ') + "variables/state:   "
	 << configuration.statespace_options.atoms_per_state
         << '\n' + string(6, ' ');

    if (configuration.statespace_options.generate_mode
	  == Configuration::RANDOMGRAPH)
      cout << "edge probability:  "
	   << configuration.statespace_options.edge_probability
	   << '\n' + string(6, ' ');

    cout << "truth probability: "
         << configuration.statespace_options.truth_probability
         << '\n' << endl;
  }

  return statespace;
}

/* ========================================================================= */
void generateFormulae
  (FormulaString::FormulaString formulae[4], const char* formula = 0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Generates a random LTL formula according to the current
 *                configuration and stores the formula, its negation and their
 *                negated normal forms into an array.
 *
 * Arguments:     formulae  --  An array of four FormulaStrings for storing
 *                              the results. The correspondence between the
 *                              different formulae and the array indices is
 *                              as follows:
 *                                formulae[0] will contain the random formula
 *                                            expanded into negation normal
 *                                            form.
 *                                formulae[1] will contain the negated random
 *                                            formula expanded into negation
 *                                            normal form.
 *                                formulae[2] will contain the random formula.
 *                                formulae[3] will contain the negated random
 *                                            formula.
 *                formula   --  A pointer to a constant C-style string. If the
 *                              pointer is non-null, no random formula will be
 *                              generated, but the formula in this string will
 *                              be used instead in initializing the result
 *                              array.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  using FormulaString::FormulaString;

  if (configuration.global_options.verbosity >= 2)
  {
    if (configuration.global_options.verbosity == 2)
      cout << "    Generating random LTL formula...\n";
    else
    {
      cout << "    Random LTL formula:\n";

      if (configuration.global_options.verbosity >= 4)
        cout << string(6, ' ') + "<generating>";
    }
    cout.flush();
  }

  if (formula == 0)
  {
    unsigned long int random_formula_length
      = formulae[0].randomize
                      (configuration.formula_options.generation_parameters);

    if (configuration.global_options.verbosity >= 3)
    {
      if (configuration.global_options.verbosity >= 4)
        cout << " ok\n";

      cout << string(6, ' ') + "variables:" + string(17, ' ')
           << configuration.formula_options.generation_parameters.
                number_of_available_variables
           << '\n' + string(6, ' ') + "length:" + string(20, ' ')
           << random_formula_length
           << '\n';
    }
  }
  else
  {
    istrstream formula_stream(formula);
    formulae[0].read(formula_stream);
  }

  formulae[1] = formulae[0];
  formulae[2] = formulae[0];

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(6, ' ') + "<converting to negation normal form>";
    cout.flush();
  }

  formulae[0].toNegationNormalForm();

  if (configuration.global_options.verbosity >= 4)
  {
    cout << " ok\n" + string(6, ' ') + "<negating formula>";
    cout.flush();
  }
  
  formulae[1].negate();
  formulae[3] = formulae[1];

  if (configuration.global_options.verbosity >= 4)
  {
    cout << " ok\n" + string(6, ' ')
            + "<converting negated formula to negation normal form>";
    cout.flush();
  }

  formulae[1].toNegationNormalForm();

  if (configuration.global_options.verbosity >= 3)
  {
    if (configuration.global_options.verbosity >= 4)
      cout << " ok\n";

    cout << string(6, ' ') + "formula:" + string(19, ' ');
    formulae[2].prettyPrint();

    if (configuration.formula_options.output_mode == Configuration::NNF)
    {
      cout << '\n' + string(8, ' ') + "in negation normal form: ";
      formulae[0].prettyPrint();
    }

    cout << '\n' + string(6, ' ') + "negated formula:" + string(11, ' ');    
    formulae[3].prettyPrint();

    if (configuration.formula_options.output_mode == Configuration::NNF)
    {
      cout << '\n' + string(8, ' ') + "in negation normal form: ";
      formulae[1].prettyPrint();
    }

    cout << endl << endl;
  }
}

/* ========================================================================= */
void writeFormulaToFile
  (const FormulaString::FormulaString& formula, const char* filename)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a FormulaString into a file.
 *
 * Arguments:     formula   --  A reference to a constant FormulaString.
 *                filename  --  A pointer to a constant C-style string with
 *                              the name of the file where the formula is to be
 *                              stored.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  bool file_opened = false;

  try
  {
    ofstream formula_file;
    Exceptional_ostream eformula_file(&formula_file,
                                      ios::failbit | ios::badbit);

    openFile(filename, formula_file, ios::out | ios::trunc, 6);
    file_opened = true;

    if (configuration.global_options.verbosity == 5)
    {
      cout << string(6, ' ') + "<writing formula to `" + filename + "'>";
      cout.flush();
    }

    eformula_file << formula << endl;
    formula_file.close();
  }
  catch (IOException e)
  {
    if (configuration.global_options.verbosity >= 1)
    {
      if (file_opened && configuration.global_options.verbosity == 5)
	cout << " error\n";

      if (configuration.global_options.verbosity >= 2)
	cout << string(6, ' ');

      cout << string("Error: ") + e.what() << endl;
    }
    
    throw;
  }

  if (configuration.global_options.verbosity == 5)
    cout << " ok\n";
}

/* ========================================================================= */
BuchiAutomaton* generateBuchiAutomaton
  (const char* formula_file_name,
   const Configuration::AlgorithmInformation& algorithm,
   AutomatonStats& automaton_stats, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructs a newly allocated BuchiAutomaton by invoking an
 *                external program which will perform the conversion of a
 *                LTL formula (stored into a file) into a Bchi automaton.
 *
 * Arguments:     formula_file_name  --  Name of the file where the formula is
 *                                       stored.
 *                algorithm          --  Reference to a constant
 *                                       AlgorithmInformation object specifying
 *                                       the external program to use.
 *                automaton_stats    --  A reference to an AutomatonStats
 *                                       object for storing statistics on the
 *                                       generated automaton.
 *                indent             --  Number of spaces to leave on the left
 *                                       of all output messages.
 *
 * Returns:       A pointer to a newly allocated BuchiAutomaton object.
 *
 * ------------------------------------------------------------------------- */
{
  BuchiAutomaton* automaton = new BuchiAutomaton;

  char automaton_file_name[L_tmpnam + 1];
  char cout_capture_file[L_tmpnam + 1];
  char cerr_capture_file[L_tmpnam + 1];
  automaton_file_name[0] = '\0';
  cout_capture_file[0] = '\0';
  cerr_capture_file[0] = '\0';

  try
  {
    if (tmpnam(automaton_file_name) == 0)
      throw FileCreationException("a temporary file");

    tmpnam(cout_capture_file);
    tmpnam(cerr_capture_file);

    /*
     *  Generate the command line to be used for executing the program that is
     *  supposed to generate the automaton from a formula. The name of the
     *  executable and any extra command line parameters are stored in the
     *  program configuration data structures. The command line will be of the
     *  form
     *    [path to executable] [additional parameters] [input file]
     *                         [output file] 1>[stdout capture file]
     *                         2>[stderr capture file],
     *  so the program used for generating the automaton is assumed to follow
     *  this format for passing the necessary input and output filenames.
     *  Note: This kind of output redirection requires that the call to
     *        `system' invokes a POSIX compatible shell!
     *
     *  The output of stdout and stderr will be redirected to two temporary
     *  files if possible.
     */

    string command_line = *(algorithm.path_to_program) + ' ';

    if (algorithm.extra_parameters != 0)
      command_line += *(algorithm.extra_parameters) + ' ';

    command_line += string(formula_file_name) + ' ' + automaton_file_name;

    if (cout_capture_file[0] != '\0')
      command_line += string(" 1>") + cout_capture_file;

    if (cerr_capture_file[0] != '\0')
      command_line += string(" 2>") + cerr_capture_file;

    if (configuration.global_options.verbosity >= 4)
    {
      cout << string(indent, ' ')
              + (configuration.global_options.verbosity == 4
                 ? "<computing Bchi automaton>"
                 : "<executing `" + command_line + "'>");
      cout.flush();
    }

    ifstream automaton_file;
    struct tms timing_information_begin, timing_information_end;

    times(&timing_information_begin);

    if (system(command_line.c_str()) != 0)
      throw ExecFailedException(*(algorithm.path_to_program));

    if (configuration.global_options.verbosity >= 4)
      cout << " ok" << endl;

    openFile(automaton_file_name, automaton_file, ios::in, indent);

    if (configuration.global_options.verbosity == 5)
    {
      cout << string(indent, ' ') + "<reading and regularizing automaton>";
      cout.flush();
    }

    automaton_file >> *automaton;

    times(&timing_information_end);

    if (configuration.global_options.verbosity == 5)
      cout << " ok" << endl;

    if (timing_information_begin.tms_utime == static_cast<clock_t>(-1)
        || timing_information_begin.tms_cutime == static_cast<clock_t>(-1)
        || timing_information_end.tms_utime == static_cast<clock_t>(-1)
        || timing_information_end.tms_cutime == static_cast<clock_t>(-1))
      automaton_stats.buchi_generation_time = -1.0;
    else
      automaton_stats.buchi_generation_time =
        static_cast<double>(timing_information_end.tms_utime
                            + timing_information_end.tms_cutime
                            - timing_information_begin.tms_utime
                            - timing_information_begin.tms_cutime) / CLK_TCK;

    automaton_file.close();
  }
  catch (IOException e)
  {
    delete automaton;

    if (automaton_file_name[0] != '\0' &&
        configuration.global_options.verbosity >= 4)
      cout << " error\n";

    if (configuration.global_options.verbosity >= 2)
      cout << string(indent, ' ');

    if (configuration.global_options.verbosity >= 1)
      cout << string("Error: ") + e.what() << endl;

    if (automaton_file_name[0] != '\0')
      removeFile(automaton_file_name, indent);

    if (cout_capture_file[0] != '\0')
    {
      if (configuration.global_options.verbosity >= 3)
        printFileContents(cout, "Contents of stdout:",
                          cout_capture_file, indent + 2, "> ");

      removeFile(cout_capture_file, indent);
    }

    if (cerr_capture_file[0] != '\0')
    {
      if (configuration.global_options.verbosity >= 3)
        printFileContents(cout, "Contents of stderr:", cerr_capture_file,
                          indent + 2, "> ");

      removeFile(cerr_capture_file, indent);
    }

    throw BuchiAutomatonGenerationException();
  }

  removeFile(automaton_file_name, indent);

  if (cout_capture_file[0] != '\0')
  {
    printFileContents(cout, "Contents of stdout:", cout_capture_file,
                      indent + 2, "> ");

    removeFile(cout_capture_file, indent);
  }

  if (cerr_capture_file[0] != '\0')
  {
    printFileContents(cout, "Contents of stderr:", cerr_capture_file, 
                      indent + 2, "> ");

    removeFile(cerr_capture_file, indent);
  }

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(indent, ' ') + "<computing statistics>";
    cout.flush();
  }

  pair<BuchiAutomaton::size_type, unsigned long int> buchi_stats =
    automaton->stats();

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;

  automaton_stats.number_of_buchi_states = buchi_stats.first;
  automaton_stats.number_of_buchi_transitions = buchi_stats.second;

  return automaton;
}

/* ========================================================================= */
ProductAutomaton* generateProductAutomaton
  (const BuchiAutomaton& automaton, const StateSpace& statespace,
   AutomatonStats& automaton_stats, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes the product of a Bchi automaton with a reachability
 *                graph.
 *
 * Arguments:     automaton        --  A reference to a constant
 *                                     BuchiAutomaton.
 *                statespace       --  A reference to a constant StateSpace.
 *                automaton_stats  --  Data structure for storing information
 *                                     about the product automaton.
 *                indent           --  Number of spaces to leave on the left of
 *                                     the messages given to the user.
 *
 * Returns:       A pointer to a newly allocated ProductAutomaton.
 *
 * ------------------------------------------------------------------------- */
{
  ProductAutomaton* product_automaton = new ProductAutomaton();

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(indent, ' ') + "<computing product automaton>";
    cout.flush();
  }

  /*
   *  Initialize the product automaton. The variable
   *  `configuration.global_options.product_mode' is used to determine whether
   *  the product is to be computed with respect to all the states of the
   *  reachability graph or only the initial state of the graph.
   */

  try
  {
    product_automaton->computeProduct
      (automaton, statespace,
       configuration.global_options.product_mode == Configuration::GLOBAL);
  }
  catch (...)
  {
    delete product_automaton;

    try
    {
      throw;
    }
    catch (UserBreakException)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " user break" << endl;
	else
	  cout << string(10, ' ') + "[User break]" << endl;
      }

      throw ProductAutomatonGenerationException();
    }
    catch (bad_alloc)
    {
      if (configuration.global_options.verbosity >= 4)
	cout << " error (out of memory)" << endl;
      
      throw ProductAutomatonGenerationException();
    }
    catch (...)
    {
      throw;
    }
  }

  /*
   *  Determine the number of states and transitions in the product.
   */

  if (configuration.global_options.verbosity >= 4)
  {
    cout << " ok\n" + string(indent, ' ') + "<computing statistics>";
    cout.flush();
  }

  pair<ProductAutomaton::size_type, unsigned long int> product_stats =
    product_automaton->stats();

  automaton_stats.number_of_product_states = product_stats.first;
  automaton_stats.number_of_product_transitions = product_stats.second;

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;

  return product_automaton;
}

/* ========================================================================= */
void performEmptinessCheck
  (const ProductAutomaton& product_automaton, AutomatonStats& automaton_stats,
   StateSpace::size_type result_size)
/* ----------------------------------------------------------------------------
 *
 * Description:   Performs the emptiness check on a ProductAutomaton, i.e.
 *                finds the states of the original reachability graph from
 *                which an accepting cycle of the Bchi automaton can be
 *                reached.
 *
 * Arguments:     product_automaton  --  A reference to a constant
 *                                       ProductAutomaton.
 *                automaton_stats    --  Data structure for storing the result
 *                                       of the emptiness check.
 *                result_size        --  Size of the result vector (dependent
 *                                       on whether the product automaton was
 *                                       computed globally or locally).
 *
 * Returns:       Nothing. The result of the emptiness check can be found
 *                in the Bitset object pointed to by
 *                `automaton_stats.emptiness_check_result'.
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(10, ' ') + "<checking product automaton for emptiness>";
    cout.flush();
  }

  automaton_stats.emptiness_check_result = new Bitset(result_size);

  try
  {
    product_automaton.emptinessCheck(*automaton_stats.emptiness_check_result);
  }
  catch (...)
  {
    delete automaton_stats.emptiness_check_result;
    automaton_stats.emptiness_check_result = 0;

    try
    {
      throw;
    }
    catch (UserBreakException)
    {
      if (configuration.global_options.verbosity >= 3)
      {
	if (configuration.global_options.verbosity >= 4)
	  cout << " user break" << endl;
	else
	  cout << string(10, ' ') + "[User break]" << endl;
      }

      throw EmptinessCheckFailedException();
    }
    catch (bad_alloc)
    {
      if (configuration.global_options.verbosity >= 4)
	cout << " error (out of memory)" << endl;

      throw EmptinessCheckFailedException();
    }
    catch (...)
    {
      throw;
    }
  }

  if (configuration.global_options.verbosity >= 4)
    cout << " ok" << endl;
}

/* ========================================================================= */
bool performConsistencyCheck(AlgorithmTestResults& test_results)
/* ----------------------------------------------------------------------------
 *
 * Description:   Checks the model checking results for consistency for a
 *                particular LTL-to-Bchi conversion algorithm implementation,
 *                i.e. verifies that the model checking results for a formula
 *                and its negation are not contradictory.
 *
 * Arguments:     test_results  --  Data structure for storing information
 *                                  about the consistency check.
 *
 * Returns:       A truth value according to whether the check succeeded or
 *                not.
 *
 * ------------------------------------------------------------------------- */
{
  StateSpace::size_type state;
  const Bitset& acceptance_vector_for_formula
    = *test_results.automaton_stats[0].emptiness_check_result;
  const Bitset& acceptance_vector_for_negation
    = *test_results.automaton_stats[1].emptiness_check_result;

  if (configuration.global_options.verbosity >= 4)
  {
    cout << string(8, ' ') + "<comparing results>";
    cout.flush();
  }

  /*
   *  The consistency check will succeed if
   *  `test_results.consistency_check_result' has the value 1 at the end of the
   *  following loop.
   *
   *  The consistency check fails if there is a state in which both the
   *  formula and its negation are claimed to be false.
   */

  test_results.consistency_check_result = 1;

  for (state = 0; state < acceptance_vector_for_formula.size(); ++state)
  {
    if (acceptance_vector_for_formula[state] !=
        acceptance_vector_for_negation[state])
      ++test_results.consistency_check_comparisons;
    else if (!acceptance_vector_for_formula[state])
    {
      ++test_results.consistency_check_comparisons;
      ++test_results.failed_consistency_check_comparisons;
      test_results.consistency_check_result = 0;
    }
  }

  if (configuration.global_options.verbosity >= 4)
    cout << (test_results.consistency_check_result == 1 ? " ok" : " failed")
         << endl;

  return (test_results.consistency_check_result == 1);
}      

/* ========================================================================= */
bool compareResults
  (const TestRoundInfo& round_info,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >& test_results,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults)>::size_type
     current_algorithm,
   vector<TestStatistics, ALLOC(TestStatistics) >& stats)
/* ----------------------------------------------------------------------------
 *
 * Description:   Compares the model checking results obtained using different
 *                LTL->Bchi conversion algorithm implementations against each
 *                other.
 *
 * Arguments:     round_info         --  A reference to a constant data
 *                                       structure with information about the
 *                                       current test round.
 *                test_results       --  A reference to a vector storing the
 *                                       model checking results obtained using
 *                                       different implementations in the
 *                                       current round.
 *                current_algorithm  --  Index to the vector
 *                                       `configuration.algorihms' indicating
 *                                       the last executed algorithm.
 *                stats              --  Data structure storing e.g. the total
 *                                       number of cross-comparisons performed
 *                                       for each algorithm.
 *
 * Returns:       A truth value telling whether the current algorithm passed
 *                the result cross-comparison test with the other algorithms.
 *
 * ------------------------------------------------------------------------- */
{
  if (current_algorithm == 0)
    return true;

  bool comparison_ok = true;

  for (int counter = 0; counter < 2; counter++)
  {
    for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type 
	   algorithm_id = 0;
         algorithm_id < current_algorithm; ++algorithm_id)
    {
      if (configuration.algorithms[algorithm_id].enabled
	  && test_results[current_algorithm].automaton_stats[counter].
               emptinessCheckPerformed()
          && test_results[algorithm_id].automaton_stats[counter].
               emptinessCheckPerformed())
      {
        test_results[current_algorithm].automaton_stats[counter].
          cross_comparison_stats[algorithm_id].first = true;
        test_results[algorithm_id].automaton_stats[counter].
          cross_comparison_stats[current_algorithm].first = true;

        (stats[current_algorithm].cross_comparisons_performed[algorithm_id])++;
        (stats[algorithm_id].cross_comparisons_performed[current_algorithm])++;

        unsigned long int dist =
          *test_results[current_algorithm].automaton_stats[counter].
            emptiness_check_result -
          *test_results[algorithm_id].automaton_stats[counter].
            emptiness_check_result;

        if (dist > 0)
        {
          test_results[current_algorithm].automaton_stats[counter].
            cross_comparison_stats[algorithm_id].second = dist;
          test_results[algorithm_id].automaton_stats[counter].
            cross_comparison_stats[current_algorithm].second = dist;
          
          (stats[current_algorithm].
           cross_comparison_mismatches[algorithm_id])++;
          (stats[algorithm_id].
           cross_comparison_mismatches[current_algorithm])++;

          comparison_ok = false;
        }
      }
    }
  }

  return comparison_ok;
}

/* ========================================================================= */
bool performPathCheck
  (const Bitset* path_truth_values,
   const AlgorithmTestResults& test_result)
/* ----------------------------------------------------------------------------
 *
 * Description:
 *
 * Arguments:
 *
 * Returns:
 *
 * ------------------------------------------------------------------------- */
{
  if (configuration.global_options.verbosity >= 2)
  {
    cout << string(6, ' ') + (configuration.global_options.verbosity == 2
			      ? "Performing path check..."
			      : "Path check:")
         << endl;
  }

  unsigned long int max_state
    = (configuration.global_options.product_mode == Configuration::GLOBAL
       ? path_truth_values->size()
       : 1);

  const Bitset& pos_result
    = *(test_result.automaton_stats[0].emptiness_check_result);
  const Bitset& neg_result
    = *(test_result.automaton_stats[1].emptiness_check_result);

  unsigned long int state;

  for (state = 0; state < max_state; state++)
  {
    if ((pos_result[state] != (*path_truth_values)[state])
	|| (neg_result[state] != !(*path_truth_values)[state]))
      break;
  }

  if (configuration.global_options.verbosity >= 3)
  {
    cout << string(8, ' ') + "result:" + string(18, ' ')
            + (state == max_state ? "pass" : "fail") + "ed"
	 << endl;
  }

  return (state == max_state);
}
