/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _TestOperations_h_
#define _TestOperations_h_

#include <cstdio>
#include <ctime>
#include <sys/times.h>
#include <fstream>
#include <iostream>
#include <strstream>
#include <vector>
#include "Bitset.h"
#include "BuchiAutomaton.h"
#include "Configuration.h"
#include "Exception.h"
#include "FormulaString.h"
#include "ProductAutomaton.h"
#include "StateSpace.h"
#include "StlConfig.h"
#include "TestRoundInfo.h"
#include "TestStatistics.h"

using namespace std;

extern Configuration configuration;

using Graph::BuchiAutomaton;
using Graph::StateSpace;
using Graph::ProductAutomaton;

void openFile
  (const char* filename, ifstream& stream, ios::openmode mode, int indent = 0);

void openFile
  (const char* filename, ofstream& stream, ios::openmode mode, int indent = 0);

void removeFile(const char* filename, int indent = 0);

void printFileContents
  (ostream& stream, const char* message, const char* filename, int indent,
   const char* line_prefix = "");

StateSpace* generateStateSpace();

void generateFormulae
  (FormulaString::FormulaString formulae[4], const char* formula = 0);

void writeFormulaToFile
  (const FormulaString::FormulaString& formula,
   const char* filename);

BuchiAutomaton* generateBuchiAutomaton
  (const char* formula_file_name,
   const Configuration::AlgorithmInformation& algorithm, 
   AutomatonStats& automaton_stats, int indent);

ProductAutomaton* 
  generateProductAutomaton
    (const BuchiAutomaton& automaton, const StateSpace& statespace,
     AutomatonStats& automaton_stats, int indent);

void performEmptinessCheck
  (const ProductAutomaton& product_automaton, AutomatonStats& automaton_stats,
   StateSpace::size_type result_size);

bool performConsistencyCheck(AlgorithmTestResults& test_results);

bool performTruthValueCheck
  (const AlgorithmTestResults& test_results,
   const Bitset& positive_formula_truth_values,
   const Bitset& negative_formula_truth_values);

bool compareResults
  (const TestRoundInfo& round_info,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >& test_results,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     current_algorithm,
   vector<TestStatistics, ALLOC(TestStatistics) >& stats);

bool performPathCheck
  (const Bitset* path_truth_values, const AlgorithmTestResults& test_result);

class StateSpaceGenerationException : public Exception
{
public:
  StateSpaceGenerationException();                  /* Constructor. */

  /* default copy constructor */

  ~StateSpaceGenerationException();                 /* Destructor. */

  StateSpaceGenerationException& operator=          /* Assignment operator. */
    (const StateSpaceGenerationException& e);
};

inline StateSpaceGenerationException::StateSpaceGenerationException() :
  Exception("State space generation failed.")
{
}

inline StateSpaceGenerationException::~StateSpaceGenerationException()
{
}

inline StateSpaceGenerationException&
StateSpaceGenerationException::operator=
  (const StateSpaceGenerationException& e)
{
  Exception::operator=(e);
  return *this;
}

class BuchiAutomatonGenerationException : public Exception
{
public:
  BuchiAutomatonGenerationException();              /* Constructor. */

  /* default copy constructor */

  ~BuchiAutomatonGenerationException();             /* Destructor. */

  BuchiAutomatonGenerationException& operator=      /* Assignment operator. */
    (const BuchiAutomatonGenerationException& e);
};

inline BuchiAutomatonGenerationException::BuchiAutomatonGenerationException() :
  Exception("Bchi automaton generation failed.")
{
}

inline BuchiAutomatonGenerationException::~BuchiAutomatonGenerationException()
{
}

inline BuchiAutomatonGenerationException&
BuchiAutomatonGenerationException::operator=
  (const BuchiAutomatonGenerationException& e)
{
  Exception::operator=(e);
  return *this;
}

class ProductAutomatonGenerationException : public Exception
{
public:
  ProductAutomatonGenerationException();            /* Constructor. */

  /* default copy constructor */

  ~ProductAutomatonGenerationException();           /* Destructor. */

  ProductAutomatonGenerationException& operator=    /* Assignment operator. */
    (const ProductAutomatonGenerationException& e);
};

inline
ProductAutomatonGenerationException::ProductAutomatonGenerationException() :
  Exception("Product automaton generation failed.")
{
}

inline
ProductAutomatonGenerationException::~ProductAutomatonGenerationException()
{
}

inline ProductAutomatonGenerationException&
ProductAutomatonGenerationException::operator=
  (const ProductAutomatonGenerationException& e)
{
  Exception::operator=(e);
  return *this;
}

class EmptinessCheckFailedException : public Exception
{
public:
  EmptinessCheckFailedException();                  /* Constructor. */

  /* default copy constructor */

  ~EmptinessCheckFailedException();                 /* Destructor. */

  EmptinessCheckFailedException& operator=          /* Assignment operator. */
    (const EmptinessCheckFailedException& e);
};

inline
EmptinessCheckFailedException::EmptinessCheckFailedException() :
  Exception("Emptiness check failed.")
{
}

inline EmptinessCheckFailedException::~EmptinessCheckFailedException()
{
}

inline EmptinessCheckFailedException&
EmptinessCheckFailedException::operator=
  (const EmptinessCheckFailedException& e)
{
  Exception::operator=(e);
  return *this;
}

#endif /* _TestOperations_h_ */
