/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "UserInterface.h"

/******************************************************************************
 *
 * Implementations for the user commands.
 *
 *****************************************************************************/

/* ========================================================================= */
void UserInterface::computeBuchiAutomaton
  (BuchiAutomaton*& buchi_automaton,
   pair<unsigned long int, bool>& last_automaton,
   const pair<unsigned long int, bool>& new_automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes a Bchi automaton.
 *
 * Arguments:     buchi_automaton  --  A reference to a pointer giving the
 *                                     storage location of the generated
 *                                     automaton.
 *                last_automaton   --  A pair telling the algorithm and the
 *                                     formula last used for computing an
 *                                     automaton (for testing whether the
 *                                     result is already available).
 *                new_automaton    --  A pair telling the algorithm and the
 *                                     formula which are to be used for
 *                                     computing the automaton.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (buchi_automaton == 0 || last_automaton != new_automaton)
  {
    AutomatonStats dummy_stats(0);

    int old_verbosity = configuration.global_options.verbosity;
    configuration.global_options.verbosity = 4;

    BuchiAutomaton* automaton;

    try
    {
      automaton =
        generateBuchiAutomaton((new_automaton.second
                                ? round_info.formula_file_name
                                : round_info.negated_formula_file_name),
                               configuration.algorithms[new_automaton.first],
                               dummy_stats,
                               2);
    }
    catch (...)
    {
      configuration.global_options.verbosity = old_verbosity;
      throw;
    }

    configuration.global_options.verbosity = old_verbosity;

    if (buchi_automaton != 0)
    {
      if (configuration.global_options.verbosity >= 4)
        cout << "  <cleaning up memory allocated for Bchi automaton>" << endl;

      delete buchi_automaton;
    }

    buchi_automaton = automaton;
    last_automaton = new_automaton;
  }
}

/* ========================================================================= */
void UserInterface::computeProductAutomaton
  (ProductAutomaton*& product_automaton,
   const BuchiAutomaton& buchi_automaton,
   pair<unsigned long int, bool>& last_automaton,
   const pair<unsigned long int, bool>& new_automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes a product automaton.
 *
 * Arguments:     product_automaton  --  A reference to a pointer giving the
 *                                       storage location of the generated
 *                                       automaton.
 *                buchi_automaton    --  The Bchi automaton to be used for
 *                                       computing the product automaton.
 *                last_automaton     --  A pair telling the algorithm and the
 *                                       formula last used for computing an
 *                                       automaton (for testing whether the
 *                                       result is already available).
 *                new_automaton      --  A pair telling the algorithm and the
 *                                       formula which are to be used for
 *                                       computing the automaton.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (product_automaton == 0 || last_automaton != new_automaton)
  {
    AutomatonStats dummy_stats(0);

    if (product_automaton != 0)
    {
      if (configuration.global_options.verbosity >= 4)
        cout << "  <cleaning up memory allocated for product automaton>"
             << endl;

      delete product_automaton;
    }

    int old_verbosity = configuration.global_options.verbosity;
    configuration.global_options.verbosity = 4;

    try
    {
      product_automaton
	= generateProductAutomaton(buchi_automaton,
				   *(round_info.statespace),
				   dummy_stats,
				   2);
    }
    catch (...)
    {
      configuration.global_options.verbosity = old_verbosity;
      delete product_automaton;
      product_automaton = 0;
      throw;
    }

    configuration.global_options.verbosity = old_verbosity;

    last_automaton = new_automaton;
  }
}

/* ========================================================================= */
void UserInterface::printAlgorithmList(ostream& stream, int indent)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `algorithms', i.e. writes a list
 *                of algorithms used in the tests to a stream.
 *
 * Arguments:     stream             --  A reference to an output stream.
 *                indent             --  Number of spaces to leave on the left
 *                                       of the output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  vector<Configuration::AlgorithmInformation,
         ALLOC(Configuration::AlgorithmInformation) >::size_type algorithm_id;

  estream << string(indent, ' ') + "List of algorithms:\n";
    
  algorithm_id = 0;
  for (vector<Configuration::AlgorithmInformation,
              ALLOC(Configuration::AlgorithmInformation) >::const_iterator
         algorithm = configuration.algorithms.begin();
       algorithm != configuration.algorithms.end();
       ++algorithm)
  {
    estream << string(indent + 2, ' ') << algorithm_id
            << ": `" + *((*algorithm).name) + "' ("
               + (algorithm->enabled ? "en" : "dis") + "abled)\n";
    ++algorithm_id;
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::printAnalysisResults
  (ostream& stream, int indent, bool formula_type,
   unsigned long int algorithm1,
   unsigned long int algorithm2,
   BuchiAutomaton*& buchi_automaton,
   pair<unsigned long int, bool>& last_buchi_automaton,
   ProductAutomaton*& product_automaton,
   pair<unsigned long int, bool>& last_product_automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `analyze', i.e. analyzes a
 *                discrepancy between the results of two algorithms by
 *                searching for a system execution producting contradicting
 *                results and then displaying the execution.
 *
 * Arguments:     stream                  --  A reference to an output stream.
 *                indent                  --  Number of spaces to leave on the
 *                                            left of the output.
 *                formula_type            --  Tells the LTL formula for which
 *                                            the analysis is to be performed.
 *                algorithm1,             --  Identifiers of the algorithms for
 *                algorithm2                  which the analysis is to be
 *                                            performed.
 *                buchi_automaton         --  A reference to a pointer telling
 *                                            the storage location of the
 *                                            generated Bchi automaton.
 *                last_buchi_automaton    --  A pair telling the algorithm and
 *                                            the formula last used for
 *                                            computing a Bchi automaton.
 *                product_automaton       --  A reference to a pointer telling
 *                                            the storage location of the
 *                                            generated product automaton.
 *                last_product_automaton  --  A pair telling the algorithm and
 *                                            the formula last used for
 *                                            computing a product automaton
 *                                            (for testing whether the result
 *                                            is already available).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  if (algorithm1 >= configuration.algorithms.size()
      || algorithm2 >= configuration.algorithms.size())
  {
    ostrstream error_msg;
    error_msg << "No such algorithm ("
              << (algorithm1 >= configuration.algorithms.size()
                  ? algorithm1
                  : algorithm2)
              << ")." << ends;
    error_msg.freeze();
    throw CommandErrorException(error_msg.str());
  }
  else if (algorithm1 == algorithm2)
    throw CommandErrorException("Algorithm identifiers must be different.");
  else
  {
    int formula = (formula_type ? 0 : 1);

    estream << string(indent, ' ') + "Result cross-comparison analysis:\n"
               + string(indent + 2, ' ') + "Formula: ";

    round_info.formulae[configuration.formula_options.output_mode
		          == Configuration::NNF
		        ? formula
		        : 2 + formula].prettyPrint(stream);

    estream << '\n';

    if (!test_results[algorithm1].automaton_stats[formula].
          crossComparisonPerformed(algorithm2))
    {
      estream << string(indent + 2, ' ')
                 + "Result cross-comparison was not performed between\n"
                 + string(indent + 2, ' ')
                 + '`' + *(configuration.algorithms[algorithm1].name)
                 + "' and `"
                 + *(configuration.algorithms[algorithm2].name) + "'.\n\n";
      return;
    }
    else if (test_results[algorithm1].automaton_stats[formula].
               cross_comparison_stats[algorithm2].second == 0)
    {
      estream << string(indent + 2, ' ')
                 + "No discrepancy between the results of\n"
                 + string(indent + 2, ' ')
                 + '`' + *(configuration.algorithms[algorithm1].name)
	         + "' and `"
                 + *(configuration.algorithms[algorithm2].name) + "'.\n\n";
      return;
    }

    /*
     *  Search for a system state in which the results of the two algorithms
     *  differ.
     */

    StateSpace::size_type state;
    for (state = 0;
         test_results[algorithm1].automaton_stats[formula].
           emptiness_check_result->test(state)
           == test_results[algorithm2].automaton_stats[formula].
                emptiness_check_result->test(state);
         ++state)
      ;

    /*
     *  Translate the formula to be analyzed into a Bchi automaton using the
     *  algorithm which claims the existence of an accepting cycle beginning in
     *  the state.
     */

    unsigned long int accepting_algorithm = 
      (test_results[algorithm1].automaton_stats[formula].
         emptiness_check_result->test(state)
       ? algorithm1
       : algorithm2);

    unsigned long int rejecting_algorithm = (accepting_algorithm == algorithm1
					     ? algorithm2
					     : algorithm1);

    computeBuchiAutomaton(buchi_automaton,
                          last_buchi_automaton,
                          make_pair(accepting_algorithm, formula_type));

    /*
     *  Compute the synchronous product of the automaton and the reachability
     *  graph.
     */

    computeProductAutomaton(product_automaton,
                            *buchi_automaton,
                            last_product_automaton,
                            make_pair(accepting_algorithm, formula_type));

    /*
     *  Search the product automaton for an accepting cycle.
     */

    cout << "  <searching for a system execution producing contradicting "
            "results>"
         << endl;

    pair<vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >,
         vector<StateSpace::size_type, ALLOC(StateSpace::size_type) > >
      execution;
    product_automaton->findAcceptingExecution(state, execution);

    /*
     *  Write information about the execution to the output stream. That is,
     *  display the states in the infinite execution (a prefix and a cycle)
     *  and tell which one of the algorithms would accept and which one
     *  would reject the execution.
     */

    estream << string(indent + 2, ' ') + "Execution M:\n"
               + string(indent + 4, ' ') + "prefix:      <";

    vector<StateSpace::size_type, ALLOC(StateSpace::size_type) >
      ::const_iterator
      execution_state;

    execution_state = execution.first.begin();
    do
    {
      estream << *execution_state;
      ++execution_state;
      if (execution_state != execution.first.end())
        estream << ", ";
    }
    while (execution_state != execution.first.end());

    estream << ">\n" + string(indent + 4, ' ') + "cycle:       <";

    execution_state = execution.second.begin();
    do
    {
      estream << *execution_state;
      ++execution_state;
      if (execution_state != execution.second.end())
        estream << ", ";
    }
    while (execution_state != execution.second.end());

    estream << ">\n" + string(indent + 4, ' ') + "accepted by: `"
               + *(configuration.algorithms[accepting_algorithm].name) + "'\n"
               + string(indent + 4, ' ') + "rejected by: `"
               + *(configuration.algorithms[rejecting_algorithm].name) + "'\n";

    /*
     *  Verify the formula separately in the obtained execution to find out
     *  which one of the algorithms was in error.
     */

    state = execution.first.size();

    for (execution_state = execution.second.begin();
	 execution_state != execution.second.end();
	 ++execution_state)
      execution.first.push_back(*execution_state);

    execution.second.clear();

    LtlFormula::LtlFormula* f;

    f = LtlFormula::formulaStringToLtlFormula
          (round_info.formulae[configuration.formula_options.output_mode
		                 == Configuration::NNF
                               ? formula
	 	               : 2 + formula]);

    estream << string(indent + 2, ' ')
            << "Analysis of the formula in the execution:\n";

    bool result = f->displayPathProof(stream, indent + 4,
				      *(round_info.statespace),
				      execution.first, state);

    estream << '\n'
            << string(indent + 2, ' ') + "The formula is "
               + (result ? "" : "not ") + "satisfied in the execution, which "
                                          "should therefore be\n"
               + string(indent + 2, ' ') +
               + (result ? "accep" : "rejec") + "ted by all Bchi automata "
                 "obtained from the formula. This suggests\n"
               + string(indent + 2, ' ') + "an error in `"
               + *(configuration.algorithms[result ? rejecting_algorithm
					           : accepting_algorithm].name)
               + "' implementation.\n";

    delete f;

    estream << endl;
    estream.flush();
  }
}
  
/* ========================================================================= */
void UserInterface::printBuchiAutomaton
  (ostream& stream, int indent, bool formula_type,
   vector<string, ALLOC(string) >& input_tokens,
   BuchiAutomaton*& buchi_automaton,
   pair<unsigned long int, bool>& last_automaton,
   Graph::GraphOutputFormat fmt)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `buchi', i.e. writes information
 *                about a set of states of a Bchi automaton computed using a
 *                given algorithm to a stream.
 *
 * Arguments:     stream           --  A reference to an output stream.
 *                indent           --  Number of spaces to leave on the left of
 *                                     the output.
 *                formula_type     --  Determines the formula to be translated
 *                                     into a Bchi automaton.
 *                input_tokens     --  A reference to a vector containing the
 *                                     arguments of the user command (the
 *                                     algorithm and automaton state numbers).
 *                buchi_automaton  --  A reference to a pointer giving the
 *                                     storage location of the generated
 *                                     automaton.
 *                last_automaton   --  A pair telling the algorithm and the
 *                                     formula last used for computing an
 *                                     automaton (for testing whether the
 *                                     result is already available).
 *                fmt              --  Determines the format of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  unsigned long int algorithm = parseNumber(input_tokens[1]);

  if (algorithm >= configuration.algorithms.size())
    throw CommandErrorException("No such algorithm ("
                                + input_tokens[1] + ").");
  else
  {
    /*
     *  Compute a Bchi automaton from the formula using the selected
     *  algorithm.
     */

    computeBuchiAutomaton(buchi_automaton,
                          last_automaton,
                          make_pair(algorithm, formula_type));

    if (fmt == Graph::NORMAL)
    {

      /*
       *  Display information about the states of the automaton. If no state
       *  list was given, display the whole automaton.
       */

      estream << string(indent, ' ') + "Bchi automaton information:\n"
                 + string(indent + 2, ' ') + "Formula: ";

      round_info.formulae[(configuration.formula_options.output_mode
			     == Configuration::NNF
			   ? 0
			   : 2)
			  + (formula_type ? 0 : 1)].prettyPrint(stream);

      estream << '\n' + string(indent + 2, ' ') + "Algorithm: "
                 + *(configuration.algorithms[algorithm].name) + '\n';

      if (buchi_automaton->empty())
        estream << string(indent + 2, ' ') + "The automaton is empty.\n";
      else
      {
        set<unsigned long int, less<unsigned long int>,
            ALLOC(unsigned long int) > states;

        if (input_tokens.size() == 2)
          input_tokens.push_back("*");

        parseInterval(input_tokens[2], states, 0, buchi_automaton->size() - 1);

        for (set<unsigned long int, less<unsigned long int>,
                 ALLOC(unsigned long int) >::const_iterator
	       state = states.begin();
             state != states.end(); ++state)
        {
          if (*state >= buchi_automaton->size())
            {
              ostrstream error_msg;
              error_msg << "State identifier out of range (" << *state << ")."
                        << ends;
              error_msg.freeze();
              throw CommandErrorException(error_msg.str());
            }
          estream << string(indent + 2, ' ') + "State " << *state
                  << string(*state == buchi_automaton->initialState()
                            ? " (initial state)" : "") + ":\n";
          buchi_automaton->node(*state).print(stream, indent + 4);
        }
      }
    }
    else if (fmt == Graph::DOT)
      buchi_automaton->print(stream, indent, Graph::DOT);

    estream << endl;
  }
}

/* ========================================================================= */
void UserInterface::evaluateFormula
  (ostream& stream, int indent, bool formula_type, 
   vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `evaluate', i.e. tells whether
 *                there exists an accepting execution beginning at a given
 *                system state (or several system states) according to the
 *                product automaton constructed from a Bchi automaton computed
 *                using a given algorithm (or algorithms).
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                formula_type  --  Determines the LTL formula to be evaluated.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    algorithms;
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    system_states;
  string algorithm_name;
  int formula = (formula_type ? 0 : 1);

  /*
   *  If no list of algorithms was given as an argument, show the results of
   *  all algorithms.
   */

  if (input_tokens.size() < 2)
    input_tokens.push_back("*");

  /*
   *  If no list of states was given, show information about all states.
   *  (If only a local product was computed, show information only about the
   *  initial state.)
   */

  if (input_tokens.size() < 3)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], algorithms, 0,
                configuration.algorithms.size() - 1);
  parseInterval(input_tokens[2], system_states, 0,
                configuration.global_options.product_mode
                  == Configuration::GLOBAL ? round_info.statespace->size() - 1
                                           : 0);

  estream << string(indent, ' ') + "Acceptance information:\n"
             + string(indent + 2, ' ') + "CTL* formula: E ";

  round_info.formulae[configuration.formula_options.output_mode
		        == Configuration::NNF
		      ? formula
		      : 2 + formula].prettyPrint(stream);

  estream << '\n';

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 state = system_states.begin();
       state != system_states.end();
       ++state)
  {
    if (*state >= (configuration.global_options.product_mode
                     == Configuration::GLOBAL
		        ? round_info.statespace->size()
		        : 1))
    {
      ostrstream error_msg;
      error_msg << "State identifier out of range (" << *state << ")." << ends;
      error_msg.freeze();
      throw CommandErrorException(error_msg.str());
    }

    estream << string(indent + 2, ' ') + "State "
            << *state << ":\n";

    for (set<unsigned long int, less<unsigned long int>,
             ALLOC(unsigned long int) >::const_iterator
	   algorithm = algorithms.begin();
         algorithm != algorithms.end();
         ++algorithm)
    {
      estream << string(indent + 4, ' ');
      if (*algorithm >= configuration.algorithms.size())
      {
        ostrstream error_msg;
        error_msg << "No such algorithm (" << *algorithm << ")." << ends;
        error_msg.freeze();
        throw CommandErrorException(error_msg.str());
      }

      algorithm_name =
        configuration.algorithms[*algorithm].name->substr(0, 20);
      estream << algorithm_name + ':'
                 + string(21 - algorithm_name.length(), ' ');

      if (!test_results[*algorithm].automaton_stats[formula].
             emptinessCheckPerformed())
        estream << "acceptance cycles not computed\n";
      else
        estream << (test_results[*algorithm].automaton_stats[formula].
                      emptiness_check_result->test(*state)
                    ? "true"
                    : "false")
                << '\n';
    }
  }

  estream << endl;
}    

/* ========================================================================= */
void UserInterface::printFormula
  (ostream& stream, int indent, bool formula_type)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `formula', i.e. displays the
 *                random LTL formula (or its negation) used for generating the
 *                Bchi automata. The formula is shown in its original form and
 *                possibly also in negation normal form if the user has
 *                requested the conversion.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                formula_type  --  Identifies the formula to be displayed.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << string(indent, ' ') + "Formula:" + string(17, ' ');
  round_info.formulae[formula_type ? 2 : 3].prettyPrint(stream);

  if (configuration.formula_options.output_mode == Configuration::NNF)
  {
    estream << '\n' + string(indent, ' ') + "In negation normal form: ";
    round_info.formulae[formula_type ? 0 : 1].prettyPrint(stream);
  }

  estream << "\n\n";
  estream.flush();
}
  
/* ========================================================================= */
void UserInterface::printCommandHelp
  (ostream& stream, int indent,
   const vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `help', i.e. gives instructions
 *                on the use of different commands.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of 
 *                                  the output.
 *                input_tokens  --  A reference to a constant vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  TokenType command = _NO_INPUT;

  if (input_tokens.size() > 1)
    command = parseCommand(input_tokens[1]);

  if (command == UNKNOWN)
    estream << string(indent, ' ') + "Unknown command (`" + input_tokens[1]
                                 + "').\n\n";

  switch (command)
  {
    case ALGORITHMS :
      estream << string(indent, ' ') + "algorithms\n";

      printTextBlock(stream, indent + 4,
                     "Display a list of algorithm identifiers which are used "
                     "as parameters for other commands. Tell also whether the"
		     " algorithms are currently enabled for testing.",
                     78);
      break;

    case ANALYZE :
      estream << string(indent, ' ') + "analyze [\"+\"|\"-\"] <algorithm> "
                                       "<algorithm>\n";

      printTextBlock(stream, indent + 4,
                     "Analyze a contradiction in the results of two "
                     "algorithms for a formula (+ denotes the positive "
                     "formula, - the negated formula). If no formula "
                     "is specified, use the positive formula.",
                     78);
      break;

    case BUCHI :
      estream << string(indent, ' ') + "buchi [\"+\"|\"-\"] <algorithm> "
                                       "[states|\"dot\"]\n";

      printTextBlock(stream, indent + 4,
                     "Display a (set of) state(s) of a Bchi automaton "
                     "constructed from the positive or negated formula by a "
                     "given algorithm (+ denotes the positive formula, - the "
                     "negated formula). If no formula is specified, use the "
                     "positive formula. If no state list is given, display "
                     "all the states of the automaton. Instead of a state "
                     "list, the description of the entire automaton can be "
                     "obtained in a format processable by dot (a tool for "
                     "visualizing graphs) by replacing the state list with "
                     "the string \"dot\".",
                     78);
      break;

    case CONTINUE :
      estream << string(indent, ' ') + "continue [number of rounds]\n";

      printTextBlock(stream, indent + 4,
                     "Continue testing, pausing next time after a given "
                     "number of rounds (this defaults to 1 if no number is "
                     "specified). \n \n Note: The output of this command is "
                     "not redirectable to a file.",
                     78);
      break;

    case DISABLE :
      estream << string(indent, ' ') + "disable [algorithms]\n";

      printTextBlock(stream, indent + 4,
		     "Disables a given set of algorithms to use in the tests. "
		     "If no algorithms are specified, disables all algorithms."
		     " \n \n Note: The output of this command is not "
		     "redirectable to a file.",
		     78);
      break;

    case ENABLE :
      estream << string(indent, ' ') + "enable [algorithms]\n";

      printTextBlock(stream, indent + 4,
		     "Enables a given set of algorithms to use in the tests. "
		     "If no algorithms are specified, enables all algorithms."
		     " \n \n Note: The output of this command is not "
		     "redirectable to a file.",
		     78);
      break;

    case EVALUATE :
      estream << string(indent, ' ') + "evaluate [\"+\"|\"-\"] "
                                       "[algorithms] [states]\n";

      printTextBlock(stream, indent + 4,
                     "Determine whether the Bchi automaton constructed from "
                     "a formula (+ denotes the positive formula, - the "
                     "negated formula) using a given algorithm accpets some "
                     "system execution starting from a given system state. "
                     "If no formula is specified, use the positive formula. "
                     "Leaving the list of algorithms or the list of states "
                     "unspecified will display the results for all algorithms "
                     "or all system states, respectively.",
                     78);
      break;

    case FORMULA :
      estream << string(indent, ' ') + "formula [\"+\"|\"-\"]\n";

      printTextBlock(stream, indent + 4,
                     "Display the LTL formula used in the last round for "
                     "generating Bchi automata (+ denotes the positive "
                     "formula, - the negated formula). If no formula is "
                     "specified, use the positive formula.",
                     78);
      break;

    case HELP :
      estream << string(indent, ' ') + "help [command]\n";

      printTextBlock(stream, indent + 4,
                     "Display help about a specific command. If no command "
                     "name is given, give a list of all available commands."
                     " \n \n In command-specific help, arguments in angle "
                     "brackets < > are obligatory, while square bracketed [ ] "
                     "arguments are optional. A vertical bar | denotes "
                     "selection between several possibilities. Arguments "
                     "in quotes must be entered literally (without the "
                     "quotes itself).",
                     78);
      break;

    case INCONSISTENCIES :
      estream << string(indent, ' ') + "inconsistencies [algorithms]\n";

      printTextBlock(stream, indent + 4,
                     "List the system states where the consistency check "
                     "failed for a given (set of) algorithm(s).",
                     78);
      break;

    case QUIT :
      estream << string(indent, ' ') + "quit\n";

      printTextBlock(stream, indent + 4, "Abort testing.", 78);
      break;

    case RESULTS :
      estream << string(indent, ' ') + "results [algorithms]\n";

      printTextBlock(stream, indent + 4,
                     "Display this round's test results for a given (set of) "
                     "algorithm(s). If no algorithms are specified, show "
                     "the results for all algorithms.",
                     78);
      break;

    case SKIP :
      estream << string(indent, ' ') + "skip [number of rounds]\n";

      printTextBlock(stream, indent + 4,
                     "Skip a given number of rounds (defaults to 1 if no "
                     "number is specified). \n \n Note: The output of this "
                     "command is not redirectable to a file.",
                     78);
      break;

    case STATESPACE :
      estream << string(indent, ' ') + "statespace [states|\"dot\"]\n";

      printTextBlock(stream, indent + 4,
                     "Display information about a given (set of) system "
                     "state(s). Display the whole reachability graph if no "
                     "states are specified. The entire reachability graph "
                     "can be obtained in a format processable by dot (a "
                     "graph visualization tool) by replacing the state list "
                     "with the string \"dot\".",
                     78);
      break;

    case STATISTICS :
      estream << string(indent, ' ') + "statistics\n";

      printTextBlock(stream, indent + 4,
                     "Display test statistics over all tests performed.",
                     78);
      break;
    
    case VERBOSITY :
      estream << string(indent, ' ') + "verbosity [0-5]\n";

      printTextBlock(stream, indent + 4,
                     "Change the verbosity of the output produced by the "
                     "program. If no value is given, display the current "
                     "setting. \n \n Note: The output of this command is "
                     "not redirectable to a file.",
                     78);
      break;
    
    default :
      printTextBlock(stream, indent,
                     "List of available commands (use `help "
                     "[command]' for command specific help): \n",
                     78);

      printTextBlock(stream, indent + 2,
                     "algorithms \n "
                     "analyze \n "
                     "buchi \n "
                     "continue \n "
		     "disable \n "
		     "enable \n "
                     "evaluate \n "
                     "help \n "
                     "inconsistencies \n "
                     "quit \n "
                     "results \n "
                     "skip \n "
                     "statespace \n "
                     "statistics \n "
                     "verbosity \n",
                     78);

      printTextBlock(stream, indent,
                     "When using commands, only a command name prefix "
                     "sufficient for identifying the command unambiguously "
                     "needs to be entered. For example, the prefix `sk' can "
                     "be used for the `skip' command. \n \n "
                     "Lists of algorithm or state identifiers may be "
                     "specified as comma-separated intervals, e.g. the "
                     "command `statespace -5,8,14-18,22-' would give the "
                     "list of all states with an identifier less than or "
                     "equal to 5, state 8 and states 14--18, together with "
                     "the states whose identifier is greater than or equal to "
                     "22. \n \n "
                     "The output of most commands can be redirected or "
                     "appended to a file by ending the command line with "
                     "`>filename' or `>>filename', respectively.",
                     78);
      break;
  }

  estream << endl;
  estream.flush();
}

/* ========================================================================= */
void UserInterface::printInconsistencies
  (ostream& stream, int indent, unsigned long int round,
   vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `inconsistencies', i.e. lists
 *                the states where the consistency check failed for a set of
 *                algorithms.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                round         --  Number of test round.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the parameters of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    number_set;

  estream << string(indent, ' ') + "Consistency check results for round "
          << round << ":\n";

  indent += 2;

  if (input_tokens.size() == 1)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], number_set, 0,
                configuration.algorithms.size() - 1);

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 algorithm = number_set.begin();
       algorithm != number_set.end(); ++algorithm)
  {
    estream << '\n';

    if (*algorithm >= configuration.algorithms.size())
      throw CommandErrorException("No such algorithm (" + input_tokens[1]
                                  + ").");

    estream << string(indent, ' ') + "Algorithm: "
               + *(configuration.algorithms[*algorithm].name) + '\n'
               + string(indent, ' ');

    if (test_results[*algorithm].consistency_check_result == -1)
      estream << "Consistency check was not performed for this algorithm.\n";
    else
    {
      /*
      estream << "Consistency check coverage: ";

      changeStreamFormatting(stream, 0, 2, ios::fixed | ios::left);
      estream << static_cast<double>
                   (test_results[*algorithm].consistency_check_comparisons)
                 / (test_results[*algorithm].automaton_stats[0].
                      emptiness_check_result->size())
                 * 100.0;
      restoreStreamFormatting(stream);

      estream << "% ["
              << test_results[*algorithm].consistency_check_comparisons
              << " of "
              << test_results[*algorithm].automaton_stats[0].
                   emptiness_check_result->size()
              << " states tested]\n" + string(indent, ' ');
      */

      if (test_results[*algorithm].consistency_check_result > 0)
        estream << "No inconsistencies detected.\n";
      else
      {
        vector<unsigned long int, ALLOC(unsigned long int) > fail_vector;

        for (unsigned long int state = 0;
             state != test_results[*algorithm].automaton_stats[0].
                        emptiness_check_result->size();
             state++)
        {
          if (!test_results[*algorithm].automaton_stats[0].
                 emptiness_check_result->test(state)
              && !test_results[*algorithm].automaton_stats[1].
                    emptiness_check_result->test(state))
            fail_vector.push_back(state);
        }

        estream << "System states failing consistency check: {";

        vector<unsigned long int, ALLOC(unsigned long int) >::const_iterator
	  state = fail_vector.begin();

        do
        {
          if (state != fail_vector.end())
          {
            estream << *state;
            ++state;
            if (state != fail_vector.end())
              estream << ", ";
          }
        }
        while (state != fail_vector.end());

        estream << "}\n";
      }
    }
    estream.flush();
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::printTestResults
  (ostream& stream, int indent, unsigned long int round,
   vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `results', i.e. displays the
 *                current round's test results for a given algorithm.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                round         --  Number of test round.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    number_set;

  estream << string(indent, ' ') + "Test results for round " << round << ":\n";

  if (input_tokens.size() == 1)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], number_set, 0,
                configuration.algorithms.size() - 1);

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 algorithm = number_set.begin();
       algorithm != number_set.end(); ++algorithm)
  {
    if (*algorithm < configuration.algorithms.size())
      printAllStats(stream, indent, *algorithm);
    else
      throw CommandErrorException("No such algorithm (" + input_tokens[1]
                                  + ").");
  }
}

/* ========================================================================= */
void UserInterface::printStateSpace
  (ostream& stream, int indent, vector<string, ALLOC(string) >& input_tokens,
   Graph::GraphOutputFormat fmt)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `statespace', i.e. displays
 *                information about a given set of states of the reachability
 *                graph.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  the output.
 *                input_tokens  --  A reference to a vector of strings
 *                                  containing the arguments of the user
 *                                  command.
 *                fmt           --  Determines the format of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    number_set;

  if (fmt == Graph::NORMAL)
  {
    estream << string(indent, ' ') + "Reachability graph information:\n";

    if (input_tokens.size() == 1)
      input_tokens.push_back("*");

    parseInterval(input_tokens[1], number_set, 0,
		  round_info.statespace->size() - 1);

    for (set<unsigned long int, less<unsigned long int>,
             ALLOC(unsigned long int) >::const_iterator
	   state = number_set.begin();
         state != number_set.end(); ++state)
    {
      if (*state < round_info.statespace->size())
      {
        estream << string(indent, ' ') + "State " << *state << ":\n";
        (*(round_info.statespace))[*state].print(stream, indent + 2);
      }
      else
      {
        ostrstream error_msg;
        error_msg << "No such state (" << *state << ")." << ends;
        error_msg.freeze();
        throw CommandErrorException(error_msg.str());
      }
    }
  }
  else if (fmt == Graph::DOT)
    round_info.statespace->print(stream, indent, Graph::DOT);

  estream << endl;
}

/* ========================================================================= */
void UserInterface::changeVerbosity
  (const vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Implements the user command `verbosity', i.e. displays or
 *                changes the output verbosity level.
 *
 * Argument:      input_tokens  --  A reference to a constant vector of strings
 *                                  containing the argument of the command.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (input_tokens.size() == 1)
    cout << "  Output verbosity is currently set to "
         << configuration.global_options.verbosity
         << ".\n" << endl;
  else
  {
    unsigned long int new_verbosity = parseNumber(input_tokens[1]);
    if (new_verbosity < 0 || new_verbosity > 5)
      throw CommandErrorException("Verbosity level must be between 0 and 5 "
                                  "inclusive.");

    configuration.global_options.verbosity = new_verbosity;

    cout << "  Output verbosity level set to " << new_verbosity << ".\n"
	 << endl;
  }
}

/* ========================================================================= */
void UserInterface::changeAlgorithmState
  (vector<string, ALLOC(string) >& input_tokens, bool enable)
/* ----------------------------------------------------------------------------
 *
 * Description:   Changes the enabledness of a set of algorithms used in the
 *                tests.
 *
 * Argument:      input_tokens  --  A reference to a constant vector of strings
 *                                  containing the argument of the command.
 *                enable        --  Determines whether the algorithms are to be
 *                                  enabled or disabled.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >
    algorithms;

  if (input_tokens.size() < 2)
    input_tokens.push_back("*");

  parseInterval(input_tokens[1], algorithms, 0,
                configuration.algorithms.size() - 1);

  for (set<unsigned long int, less<unsigned long int>,
           ALLOC(unsigned long int) >::const_iterator
	 alg = algorithms.begin();
       alg != algorithms.end(); alg++)
  {
    if (*alg >= configuration.algorithms.size())
    {
      ostrstream error_msg;
      error_msg << "No such algorithm (" << *alg << ")." << ends;
      error_msg.freeze();
      throw CommandErrorException(error_msg.str());
    }

    cout << string("  ") + (enable ? "En" : "Dis") + "abling algorithm "
	 << *alg 
	 << " (`" + *(configuration.algorithms[*alg].name) + "').\n";
    
    configuration.algorithms[*alg].enabled = enable;
  }

  cout << endl;
}
