// This file is based on code originally written by Mauno Rnkk,
// Copyright (C) 1998 Mauno Rnkk <mronkko@abo.fi>.
// Minor modifications by Heikki Tauriainen <Heikki.Tauriainen@hut.fi>.
//
// The original version of this file can be obtained through
//   <URL: http://www.abo.fi/~mauno.ronkko/ASA/ltlalg.html>.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#ifndef __Ltl_h__

#define __Ltl_h__

#include "CString.h"
#include "Set.h"
#include <iomanip.h>


class Ltl; // defined later..



/**
 * A class for Ltl formula.
 *
 *
 * The class for pointer to a Ltl formula provides
 * a front end for the Set, so that elements in Set<LtlFormula>
 * are not maltreated. This class is primarily used by the LtlGraph
 * class.
 *
 *
 * This class provides:
 * constructors/ destructor, assignment operator,
 * safe pointer to the formula, comparison functions, and output to ostream.
 **/ 
class LtlFormula {

  /// ---------- LtlFormula private attributes ---------- ///
 private:
  /// Contents.
  Ltl *m_ltl;
  

  /// ---------- LtlFormula public methods ---------- ///
 public:
  /// Initializes with a null pointer; null expression.
  LtlFormula ();
  
  /// Initializes with a ltl formula.
  LtlFormula (const Ltl &formula);
  
  /// Copy constructor.
  LtlFormula (const LtlFormula &ltl);
  
  /// Destructs this LtlFormula.
  ~LtlFormula ();
  
  /// Assigns the LtlFormula to this.
  const LtlFormula& operator = (const LtlFormula &ltl);
  
  /// Returns a pointer to this ltl formula.
  const Ltl* ltl() const
    { return m_ltl; }
  
  /// Are the two Ltl formulae the same? See the operator== in Ltl.
  const bool operator == (const LtlFormula &ltl) const;
  
  /// Are the two Ltl formulae not the same? See the operator!= in Ltl.
  const bool operator != (const LtlFormula &ltl) const;
  
  /// Outputs a LtlFormula to ostream.
  friend ostream& operator << (ostream& str, const LtlFormula& ltl);
};
  





/**
 * A class for a graph representing a Ltl formula.
 *
 *
 * A LtlGraph is automatically constructed from a given
 * Ltl formula. All relevant information is fully stored to the
 * graph, so that the formula itself may well be destoyed right after
 * the conversion. The Ltl operators/operands provide themelves
 * the knowledge of how to be expanded into a LtlGraph.
 *
 *
 * This class provides:
 * PRIVATE: a subclass for nodes, constructor, function returning
 * a pointer to a node, which has the same 'm_old' and 'm_next' fields 
 * as the given node, and a function returning a unique name for a node.
 * PUBLIC: constructors/destructor, assignment operator, comparison
 * functions, and output to ostream.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlGraph {
  friend class Ltl;

  /// ---------- LtlGraph private sub-classes ---------- ///
 private:
  /**
   * A class for a node in a graph representing a Ltl formula.
   *
   *
   * The graph itself is a set of nodes of this kind.
   * The Node class is merely a front end for the set representation.
   *
   *
   * This class provides:
   * constructors/destructor, assignment operator, comparison
   * functions, and output to ostream, as well as a matching
   * function for detecting a node from the graph, for which
   * both 'm_old' and 'm_next' fields are the same.
   **/
  class Node {

    /// ---------- Node public attributes ---------- ///
  public:
    /// A constant name for the initial node in the graph.
    static const int INITNAME=-1;

    /// Name for the node.
    int             m_name;
    
    /// The name of the node that was splitted into this another node.
    int             m_father;
    
    /// From which nodes can one come to this node.
    Set<int>        m_incoming;
    
    /// A set of formulae to be processed.
    Set<LtlFormula> m_new;
    
    /// A set of formulae that must hold, when this node is entered.
    Set<LtlFormula> m_old;
    
    /**
     * A set of atomic propositions that must hold, when this node is entered.
     **/
    Set<LtlFormula> m_atomic;
    
    /// A set of formulae that must hold in immediate next nodes.
    Set<LtlFormula> m_next;
    

    /// ---------- Node public methods ---------- ///
  public:
    /// Constructs an empty node.
    Node ();
    
    /// Constructs a specific node.
    Node (const int name, 
	  const int father, 
	  const Set<int> &incoming,
	  const Set<LtlFormula> &neww,
	  const Set<LtlFormula> &old,
	  const Set<LtlFormula> &atomic,
	  const Set<LtlFormula> &next);
    
    /// Copy constructor.
    Node (const Node &node);
    
    /// Destructs this node.
    ~Node ()
      { }
    
    /// Assigns the value of another node into this.
    const Node& operator = (const Node &node);
    
    /// Are two nodes the same?
    const bool operator == (const Node& node) const;
    
    /// Is there a difference between two nodes?
    const bool operator != (const Node& node) const
      { return !(*this==node); }
    
    /// Outputs this node to ostream.
    friend ostream& operator << (ostream& str, const Node& node);
    
    /// Does the two nodes have equal 'm_old' and 'm_next' fields?
    friend const bool sameOldAndNext (const Node &a, const Node &c);
    
    /// Does the two nodes have equal 'm_name' fields?
    friend const bool sameName (const Node &a, const Node &c);
  };


  /// ---------- LtlGraph private attributes ---------- ///
 private:
  /// Internal representation of the graph.
  Set<Node> m_nodes;


  /// ---------- LtlGraph private methods ---------- ///
 private:
  /**
   * Constructs a graph from a set of nodes. The set may well be
   * a partially constructed graph.
   **/
  LtlGraph (const Set<Node> &gnset)
    { m_nodes=gnset; }

  /**
   * Returns a pointer to a node in the graph, which has the same
   * 'm_old' and 'm_next' fields as the given node. The node pointed
   * by the returned pointer is allowed to be altered!
   **/
  Node *whichSameOldAndNext (const Node &node)
    { return m_nodes.matching(sameOldAndNext, node); }

  /// Returns a unique name for a node.
  friend int newName();


  /// ---------- LtlGraph public methods ---------- ///
 public:
  /// Constructs an empty graph.
  LtlGraph()
    { } 

  /// Copy constructor.
  LtlGraph (const LtlGraph &set)
    { m_nodes=set.m_nodes; }

  /// Constructs a graph from a given ltl formula.
  LtlGraph (const Ltl &formula);

  /// Constructs a graph from a given LtlFormula.
  LtlGraph (const LtlFormula &formula);

  /// Destructs this graph.
  ~LtlGraph()
    { }

  /// Assigns the graph from another graph.
  const LtlGraph& operator = (const LtlGraph &set)
    { m_nodes=set.m_nodes; return *this; }

  /**
   * Are two graphs the same?
   * Note that the constructed graphs may well be different
   * for two formulae that are equal!
   **/
  const bool operator == (const LtlGraph& set) const
    { return m_nodes==set.m_nodes; }
  
  /**
   * Is there a difference between two graphs?
   * Note that the constructed graphs may well be different
   * for two formulae that are equal!
   * However, if the graphs are the same, the formulae must be the same, too.
   **/
  const bool operator != (const LtlGraph& set) const
    { return !(*this==set); }
  
  /// Outputs this graph to ostream.
  friend ostream& operator << (ostream& str, const LtlGraph& set);

  /// Returns set of state labels.
  const Set<int> states() const;

  /// Returns set of propositions that must hold at a certain state.
  const Set<LtlFormula> propositions(int state) const;
  
  /// Returns set of temporal properties that must hold at a certain state.
  const Set<LtlFormula> temporalProperties(int state) const;
  
  /// Returns set labels of reachable states form given state.
  const Set<int> nextStates (int state) const;
  
  /// Returns the label of the initial state.
  const int initialState() const
    { return Node::INITNAME; }

};






/**
 * A base class for Ltl formulae.
 *
 *
 * Other Ltl operators and atomic operands are inherited from this
 * class.
 *
 *
 * This class provides:
 * PROTECTED: both virtual and fixed expansion functions for LtlGraph, 
 * and virtual negation cloning function.
 * PUBLIC: virtual destructor and conversion function to CString, 
 * virtual cloning function, global comparison functions, 
 * and output to ostream.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class Ltl {
  friend class LtlGraph;
  friend class LtlNot;
  friend class LtlAnd;
  friend class LtlOr;
  friend class DerivedLtl;
  friend class LtlUntil;
  friend class LtlResist;
  friend class LtlNext;


  /// ---------- Ltl protected methods ---------- ///
 protected:
  /**
   * This is the main part of the algorithm that expands a ltl formula
   * into a graph.
   **/
  const LtlGraph expand (const LtlGraph::Node &node, 
			 const LtlGraph &nodes) const;
 
  /// Implements operator/operand specific details of the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node, 
				  const LtlGraph &nodes) const
    { return LtlGraph(); }

  /// Returns a clone of a negation of this Ltl formula.
  virtual Ltl *negClone() const
    { return new Ltl(); }
 

  /// ---------- Ltl public methods ---------- ///
 public:
  /// Destructs this Ltl formula.
  virtual ~Ltl()
    { }

  /// Returns a clone of this Ltl formula.
  virtual Ltl *clone() const
    { return new Ltl(); }

  /// Converts this to CString.
  virtual const CString expr() const 
    { return ""; }
  
  /**
   * Are two Ltls structurally equal?
   * Note that two Ltls may well be structurally different
   * even though they are te same!
   **/
  const bool operator == (const Ltl& formula) const
    { return expr()==formula.expr(); }
  
  /// Is this Ltl structurally 'less' than another?
  const bool operator < (const Ltl& formula) const
    { return expr()<formula.expr(); }
  
  /**
   * Is there a difference between the structures of two Ltls?
   * Note that two Ltls may well be structurally different
   * even though they are te same!
   * However, if the ltls are structurally the same, the formulae 
   * must be the same, too.
   **/
  const bool operator != (const Ltl& formula) const
    { return expr()!=formula.expr(); }
  
  /// Outputs a Ltl to ostream.
  friend ostream& operator << (ostream& str, const Ltl& formula);
};






/**
 * A class for a negative Ltl formula.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, and
 * virtual negation cloning function.
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlNot : public Ltl {

  /// ---------- LtlNot private attributes ---------- ///
 private:
  /// Contents.
  Ltl *m_formula;

  /// Same formula, where negation has been propagated.
  Ltl *m_propagated;


  /// ---------- LtlNot protected methods ---------- ///
 protected:
  /// Uses the propagated formula for expansion.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const
    { return m_propagated->expand1(node, set); }

  /// Returns a clone of the non-negated Ltl formula.
  virtual Ltl *negClone() const
    { return m_formula->clone(); }


  /// ---------- LtlNot public methods ---------- ///
 public:
  /// Constructs a negative Ltl formula.
  LtlNot(const Ltl &formula);

  /// Destructs this negative Ltl formula.
  virtual ~LtlNot();
  
  /// Returns a clone of a negative Ltl formula.
  virtual Ltl *clone() const
    { return new LtlNot(*m_formula); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "!"+m_formula->expr(); }
};






/**
 * A class for Ltl 'false'.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, and 
 * virtual negation cloning function.
 * PUBLIC: virtual conversion function to CString, 
 * as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlFalse : public Ltl {

  /// ---------- LtlFalse protected methdos ---------- ///
 protected:
  /// False is never expanded; contradiction.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const
    { return set; }
  
  /// Returns a clone of LtlTrue.
  virtual Ltl *negClone() const;


  /// ---------- LtlFalse public methods ---------- ///
 public:
  /// Returns a clone of false.
  virtual Ltl *clone() const
    { return new LtlFalse(); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "false"; }
};






/**
 * A class for Ltl 'true'.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, and 
 * virtual negation cloning function
 * PUBLIC: virtual conversion function to CString, 
 * as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlTrue : public Ltl {

  /// ---------- LtlTrue protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;
  
  /// Returns a clone of LtlFalse.
  virtual Ltl *negClone() const
    { return new LtlFalse(); }


  /// ---------- LtlTrue public methods ---------- ///
 public:
  /// Returns a clone of true.
  virtual Ltl *clone() const
    { return new LtlTrue(); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "true"; }
};






/**
 * A class for a Ltl atom.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, and 
 * virtual negation cloning function.
 * PUBLIC: constructor, virtual conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlAtom : public Ltl {

  /// ---------- LtlAtom private attributes ---------- ///
 private:
  /// Contents.
  CString m_value;


  /// ---------- LtlAtom protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;
  
  /// Returns a clone of NegAtom.
  virtual Ltl *negClone() const;


  /// ---------- LtlAtom public methods ---------- ///
 public:
  /// Constructs a Ltl atom.
  LtlAtom(const CString &atom)
    { m_value=atom; }

  /// Returns a clone of an atom.
  virtual Ltl *clone() const
    { return new LtlAtom(m_value); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return m_value; }
};






/**
 * A class for a negative Ltl atom.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, and 
 * virtual negation cloning function.
 * PUBLIC: constructor, virtual conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlNegAtom : public Ltl {

  /// ---------- LtlNegAtom private attributes ---------- ///
 private:
  /// Contents.
  CString m_value;

 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;

  /// Return a clone of an non-negated LtlAtom.
  virtual Ltl *negClone() const
    { return new LtlAtom(m_value); }
  
 public:
  /// Constructs a negative Ltl atom.
  LtlNegAtom(const CString &atom)
    { m_value=atom; }

  /// Returns a clone of a negative Ltl atom.
  virtual Ltl *clone() const
    { return new LtlNegAtom(m_value); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "!"+m_value; }
};







/**
 * A class for a Ltl conjunction.
 *
 *
 * This class provides:
 * PROTECTED virtual expansion function for LtlGraph, twin constructor,
 * and virtual negation cloning function.
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlAnd : public Ltl {
  friend class LtlOr;

  /// ---------- LtlAnd private attributes ---------- ///
 private:
  /// The left and right hand side sub-formulae.
  Ltl *m_left, *m_right;


  /// ---------- LtlAnd protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;
  
  /// Constructs a twin.
  LtlAnd(Ltl *lf, Ltl *rf);

  /// Returns a clone of disjunction of negated sub-formulae.
  virtual Ltl *negClone() const;


  /// ---------- LtlAnd public methods ---------- ///
 public:
  /// Constructs a Ltl conjunction.
  LtlAnd(const Ltl &lf, const Ltl &rf);

  /// Destructs this conjucntion.
  virtual ~LtlAnd();
  
  /// Return a clone of this conjunction.
  virtual Ltl *clone() const
    { return new LtlAnd(*m_left, *m_right); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "("+m_left->expr()+"/\\"+m_right->expr()+")"; }
};






/**
 * A class for a Ltl disjunction.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, twin constructor,
 * and virtual negation cloning function.
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlOr : public Ltl {
  friend class LtlAnd;

  /// ---------- LtlOr private attributes ---------- ///
 private:
  /// The left and right hand side sub-formulae.
  Ltl *m_left, *m_right;


  /// ---------- LtlOr protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;
  
  /// Constructs a twin.
  LtlOr(Ltl *lf, Ltl *rf);

  /// Returns a clone of conjunction of negated sub-formulae.
  virtual Ltl *negClone() const;


  /// ---------- LtlOr public methods ---------- ///
 public:
  /// Constructs a Ltl disjunction.
  LtlOr(const Ltl &lf, const Ltl &rf);

  /// Destruct this disjunction.
  virtual ~LtlOr();
  
  /// Return a clone of this Ltl disjunction.
  virtual Ltl *clone() const
    { return new LtlOr(*m_left, *m_right); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "("+m_left->expr()+"\\/"+m_right->expr()+")"; }
};






/**
 * A class for a Ltl 'until'.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, twin constructor,
 * and virtual negation cloning function.
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlUntil : public Ltl {
  friend class LtlResist;

  /// ---------- LtlUntil private attributes ---------- ///
 public:
  /// Contents.
  Ltl *m_left, *m_right;


  /// ---------- LtlUntil protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;
  
  /// Constructs a twin.
  LtlUntil(Ltl *lf, Ltl *rf);

  /// Returns a clone of a Ltl resist with negated sub-formulae.
  virtual Ltl *negClone() const;


  /// ---------- LtlUntil public methods ---------- ///
 public:
  /// Constructs a Ltl until.
  LtlUntil(const Ltl &lf, const Ltl &rf);

  /// Destructs this Ltl until.
  virtual ~LtlUntil();
  
  /// Returns a clone of this Ltl until.
  virtual Ltl *clone() const
    { return new LtlUntil(*m_left, *m_right); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "("+m_left->expr()+"U"+m_right->expr()+")"; }
};






/**
 * A class for a Ltl 'resist'.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, twin constructor,
 * and virtual negation cloning function.
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlResist : public Ltl {
  friend class LtlUntil;

  /// ---------- LtlResist private attributes ---------- ///
 private:
  /// Contents.
  Ltl *m_left, *m_right;


  /// ---------- LtlResist protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;
  
  /// Constructs a twin.
  LtlResist(Ltl *lf, Ltl *rf);

  /// Returns a clone of a Ltl until with negated sub-formulae.
  virtual Ltl *negClone() const;


  /// ---------- LtlResist ---------- ///
 public:
  /// Constructs a Ltl resist.
  LtlResist(const Ltl &lf, const Ltl &rf);

  /// Destructs this Ltl resist.
  virtual ~LtlResist();
  
  /// Returns a clone of this Ltl resist.
  virtual Ltl *clone() const
    { return new LtlResist(*m_left, *m_right); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "("+m_left->expr()+"V"+m_right->expr()+")"; }
};






/**
 * A class for a Ltl 'next'.
 *
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph, and 
 * virtual negation cloning function.
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlNext : public Ltl {

  /// ---------- LtlNext private attributes ---------- ///
 private:
  /// Contents.
  Ltl *m_formula;


  /// ---------- LtlNext protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const;
  
  /// Constructs a twin.
  LtlNext(Ltl *formula)
    { m_formula=formula; }

  /// Returns a clone of a next with negated sub-formula.
  virtual Ltl *negClone() const
    {
      /**
       * Note that the twin constructor is used instead of the public
       * constructor. This binds the created negClones directly to the
       * LtlNext formula. No copies are left hanging.
       **/
      return new LtlNext(m_formula->negClone());
    }


  /// ---------- LtlNext public methods ---------- ///
 public:
  /// Constructs a Ltl next.
  LtlNext(const Ltl &formula)
    { m_formula=formula.clone(); }

  /// Destructs this Ltl next.
  virtual ~LtlNext()
    { delete m_formula; }
  
  /// Returns a clone of this Ltl next.
  virtual Ltl *clone() const
    { return new LtlNext(*m_formula); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "X"+m_formula->expr(); }
};






/**
 * A base class for a derived Ltl operators.
 *
 *
 * Any derivable Ltl operators should be inherited from this class.
 * In a nutshell, DerivedLtl class hides the expansion algorithm
 * for LtlGraph (as well as the cloning of a negation) 
 * completely behind the unfolding.
 *
 * This class provides:
 * PROTECTED: virtual expansion function for LtlGraph,
 * and virtual negation cloning function.
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: See LtlImply class.
 **/
class DerivedLtl : public Ltl {

  /// ---------- DerivedLtl protected attributes ---------- ///
 protected:
  /// The unfolded form of the derivable Ltl operator.
  Ltl *m_unfolded;


  /// ---------- DerivedLtl protected methods ---------- ///
 protected:
  /// Implements the expansion algorithm through the unfolded form.
  virtual const LtlGraph expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const
    { return m_unfolded->expand1(node, set); }
  
  /// Return a clone of a negation of the unfolded form.
  virtual Ltl *negClone() const
    { return m_unfolded->negClone(); }


  /// ---------- DerivedLtl public methods ---------- ///
 public:
  /// Constructs a derived Ltl expression.
  DerivedLtl(const Ltl &unfolded)
    { m_unfolded = unfolded.clone(); }

  /// Destruct this derived expression.
  virtual ~DerivedLtl()
    { delete m_unfolded; }
  
  /// Return a clone of this derived expression.
  virtual Ltl *clone() const
    { return new DerivedLtl(*m_unfolded); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "("+m_unfolded->expr()+")"; }
};






/**
 * A class for a Ltl implication.
 *
 *
 * This class provides:
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlImply : public DerivedLtl {

  /// ---------- LtlImply private attributes ---------- ///
 private:
  /// The contents of the implication.
  Ltl *m_assume, *m_conclude;


  /// ---------- LtlImply public methods ---------- ///
 public:
  /// Constructs an implication.
  LtlImply(const Ltl &assume, const Ltl &conclude) :
    DerivedLtl(LtlOr(LtlNot(assume),conclude))
    { m_assume=assume.clone(); m_conclude=conclude.clone(); }

  /// Destruct this implication.
  virtual ~LtlImply()
    { delete m_assume; delete m_conclude; }
  
  /// Return a clone of this implication.
  virtual Ltl *clone() const
    { return new LtlImply(*m_assume, *m_conclude); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "("+m_assume->expr()+"=>"+m_conclude->expr()+")"; }
};





/**
 * A class for a Ltl equivalence.
 *
 *
 * This class provides:
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlIff : public DerivedLtl {

  /// ---------- LtlIff private attributes ---------- ///
 private:
  /// The contents of the equivalence.
  Ltl *m_left, *m_right;


  /// ---------- LtlIff public methods ---------- ///
 public:
  /// Constructs an equivalence.
  LtlIff(const Ltl &lf, const Ltl &rf) :
    DerivedLtl(LtlAnd(LtlImply(lf,rf),LtlImply(rf,lf)))
    { m_left=lf.clone(); m_right=rf.clone(); }

  /// Destruct this equivalence.
  virtual ~LtlIff()
    { delete m_left; delete m_right; }
  
  /// Return a clone of this equivalence.
  virtual Ltl *clone() const
    { return new LtlIff(*m_left, *m_right); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "("+m_left->expr()+"<=>"+m_right->expr()+")"; }
};






/**
 * A class for a Ltl eventually.
 *
 *
 * This class provides:
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlEventually : public DerivedLtl {

  /// ---------- LtlEventually private attributes ---------- ///
 private:
  /// The contents of the eventuality.
  Ltl *m_formula;


  /// ---------- LtlEventually public methods ---------- ///
 public:
  /// Constructs an eventuality.
  LtlEventually(const Ltl &formula) :
    DerivedLtl(LtlUntil(LtlTrue(),formula))
    { m_formula=formula.clone(); }

  /// Destruct this eventuality.
  virtual ~LtlEventually()
    { delete m_formula; }
  
  /// Return a clone of this eventuality.
  virtual Ltl *clone() const
    { return new LtlEventually(*m_formula); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "<>"+m_formula->expr(); }
};






/**
 * A class for a Ltl always.
 *
 *
 * This class provides:
 * PUBLIC: constructor, virtual destructor and conversion function to 
 * CString, as well as a virtual cloning function.
 *
 * 
 * An example: <A HREF="LtlExample.cpp">LtlExample.cpp</A>
 **/
class LtlAlways : public DerivedLtl {

  /// ---------- LtlAlways private attributes ---------- ///
 private:
  /// The contents of the foreverness.
  Ltl *m_formula;


  /// ---------- LtlAlways public methods ---------- ///
 public:
  /// Constructs an foreverness.
  LtlAlways(const Ltl &formula) :
    DerivedLtl(LtlNot(LtlEventually(LtlNot(formula))))
    { m_formula=formula.clone(); }

  /// Destruct this foreverness.
  virtual ~LtlAlways()
    { delete m_formula; }
  
  /// Return a clone of this foreverness.
  virtual Ltl *clone() const
    { return new LtlAlways(*m_formula); }
  
  /// Converts this to CString.
  virtual const CString expr() const
    { return "[]"+m_formula->expr(); }
};



#endif

