/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "aasa_lbt.h"

namespace abo_buchi
{

/******************************************************************************
 *
 * Function definitions for class AcceptanceSetInfo.
 *
 *****************************************************************************/

/* ========================================================================= */
AcceptanceSetInfo& AcceptanceSetInfo::operator=(const AcceptanceSetInfo& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class AcceptanceSetInfo. Copies the
 *                contents of an AcceptanceSetInfo object to another one.
 *
 * Argument:      info  --  A reference to a constant AcceptanceSetInfo
 *                          object.
 *
 * Returns:       A reference to the AcceptanceSetInfo object whose value was
 *                (possibly) changed.
 *
 * ------------------------------------------------------------------------- */
{
  if (&info != this)
  {
    delete right_subformula;
    right_subformula = info.right_subformula->clone();
    accepting_states = info.accepting_states;
  }

  return *this;
}



/******************************************************************************
 *
 * Function definitions for class AboAutomaton.
 *
 *****************************************************************************/

/* ========================================================================= */
AboAutomaton::AboAutomaton(FormulaString& formula)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class AboAutomaton. Translates a
 *                FormulaString object to a generalized Bchi automaton using
 *                the bo implementation, computes the acceptance sets of the
 *                automaton and stores the result to the constructed object.
 *
 * Argument:      formula  --  A reference to a FormulaString object to be
 *                             translated to a Bchi automaton.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  FormulaString::TraversalInfo<AboFormulaCreationInfo> info;

  /*
   *  Translate a FormulaString to a Ltl object. This is done by `traversing'
   *  the FormulaString, using the various CreateAbo___ objects to maintain
   *  a stack of partial results, until the result of the translation is ready
   *  on the top of the stack.
   *
   *  As a side effect, the map `acceptance_sets' will finally contain an 
   *  (empty) entry for each Until subformula of the original formula, that is,
   *  an entry for each acceptance set of the Bchi automaton to be constructed
   *  later.
   */  

  formula.template
    traverse<CreateAboAtom, CreateAboConstant<LtlTrue>,
             CreateAboConstant<LtlFalse>, CreateAboUnary<LtlNot>,
             CreateAboUnary<LtlNext>, CreateAboUnary<LtlEventually>,
             CreateAboUnary<LtlAlways>, CreateAboBinary<LtlAnd>,
             CreateAboBinary<LtlOr>, CreateAboBinary<LtlImply>,
             CreateAboBinary<LtlIff>, CreateAboBinary<LtlUntil>,
             CreateAboBinary<LtlResist> >
    (FormulaString::POSTORDER, info);

  Ltl* result = info.data.top();
  info.data.pop();

  /*
   *  Construct a Bchi automaton from the Ltl object.
   */

  gba = LtlGraph(*result);

  delete result;

  /*
   *  Finally, determine the contents of the sets of accepting states.
   */   

  computeAcceptanceSets();
}

/* ========================================================================= */
AboAutomaton::AboAutomaton(const AboAutomaton& automaton) : 
  gba(automaton.gba)
/* ----------------------------------------------------------------------------
 *
 * Description:   Copy constructor for class AboAutomaton. Creates an identical
 *                copy of an AboAutomaton.
 *
 * Argument:      automaton  --  A reference to the constant AboAutomaton to be
 *                               copied.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  for (map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
           ALLOC(AcceptanceSetInfo*) >::const_iterator
         acceptance_set = automaton.acceptance_sets.begin();
       acceptance_set != automaton.acceptance_sets.end();
       ++acceptance_set)
  {
    acceptance_sets.insert
      (make_pair((*acceptance_set).first->clone(),
                 new AcceptanceSetInfo(*((*acceptance_set).second))));
  }
}

/* ========================================================================= */
AboAutomaton::~AboAutomaton()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class AboAutomaton. Deallocates the memory
 *                reserved by the AboAutomaton object.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  for (map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
           ALLOC(AcceptanceSetInfo*) >::iterator
         acceptance_set = acceptance_sets.begin();
       acceptance_set != acceptance_sets.end();
       ++acceptance_set)
  {
    delete (*acceptance_set).first;
    delete (*acceptance_set).second;
  }
}

/* ========================================================================= */
AboAutomaton& AboAutomaton::operator=(const AboAutomaton& automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class AboAutomaton. Copies the
 *                contents of an AboAutomaton object to another one.
 *
 * Argument:      automaton  --  A reference to the constant automaton to be
 *                               assigned.
 *
 * Returns:       A reference to the object whose value was (possibly) changed.
 *
 * ------------------------------------------------------------------------- */
{
  if (&automaton != this)
  {
    gba = automaton.gba;
    for (map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
             ALLOC(AcceptanceSetInfo*) >::iterator
           acceptance_set = acceptance_sets.begin();
         acceptance_set != acceptance_sets.end();
         ++acceptance_set)
    {
      delete (*acceptance_set).first;
      delete (*acceptance_set).second;
    }
    acceptance_sets.clear();
    for (map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
             ALLOC(AcceptanceSetInfo*) >::const_iterator
           acceptance_set = automaton.acceptance_sets.begin();
         acceptance_set != automaton.acceptance_sets.end();
         ++acceptance_set)
    {
      acceptance_sets.insert
        (make_pair((*acceptance_set).first->clone(),
                   new AcceptanceSetInfo(*((*acceptance_set).second))));
    }
  }

  return *this;
}

/* ========================================================================= */
unsigned long int AboAutomaton::size() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Determines the size (number of states) in an AboAutomaton.
 *
 * Arguments:     None.
 *
 * Returns:       Size of the automaton.
 *
 * ------------------------------------------------------------------------- */
{
  Set<int> states = gba.states();
  int state;
  unsigned long int result;

  for (result = 0; !states.isEmpty(); ++result)
  {
    state = *states.some();
    states = states - state;
  }

  return result;
}

/* ========================================================================= */
void AboAutomaton::computeAcceptanceSets()
/* ----------------------------------------------------------------------------
 *
 * Description:   Computes the acceptance sets of a generalized Bchi
 *                automaton constructed by the bo implementation. (This
 *                function is called only when constructing an AboAutomaton
 *                object.)
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  int state;
  Set<int> states;
  const LtlFormula*temporal_property;
  const LtlUntil* until_formula;
  Set<LtlFormula> temporal_properties;
  bool accepting;

  /*
   *  Scan through all automaton states and all their temporal properties to
   *  find all the Until formulae occurring as temporal properties.
   */

  acceptance_sets.clear();

  states = gba.states();
  while (!states.isEmpty())
  {
    state = *states.some();
    states = states - state;

    temporal_properties = gba.temporalProperties(state);
    while (!temporal_properties.isEmpty())
    {
      temporal_property = temporal_properties.some();
      until_formula = dynamic_cast<const LtlUntil*>(temporal_property->ltl());

      /*
       *  If the dynamic cast of a Ltl* to a const LtlUntil* succeeded, the
       *  formula is a LtlUntil formula, so a new acceptance set is
       *  created for it (unless there already exists one).
       */

      if (until_formula != 0)
      {
        if (acceptance_sets.find(until_formula) == acceptance_sets.end())
          acceptance_sets.insert
            (make_pair(until_formula->clone(),
                       new AcceptanceSetInfo(until_formula->m_right->clone())));
      }

      temporal_properties = temporal_properties - *temporal_property;
    }
  }

  /*
   *  Scan again through the states and temporal properties of the automaton,
   *  this time "filling the acceptance sets" with appropriate states.
   */

  states = gba.states();
  while (!states.isEmpty())
  {
    state = *states.some();
    states = states - state;

    /*
     *  Given an acceptance set associated with some Until formula, the state
     *  belongs to the acceptance set if either (or both) of the following
     *  conditions holds:
     *     1. The Until formula itself is not true in the state.
     *     2. The right subformula of the Until formula is true in the state.
     *
     *  The following code tests these conditions in each state and adds the
     *  state to the appropriate acceptance sets.
     */

    for (map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
             ALLOC(AcceptanceSetInfo*) >::iterator
           acceptance_set = acceptance_sets.begin();
         acceptance_set != acceptance_sets.end();
         ++acceptance_set)
    {
      temporal_properties = gba.temporalProperties(state);
      accepting = true;

      while (!temporal_properties.isEmpty())
      {
        temporal_property = temporal_properties.some();

        if (*temporal_property ==
            *((*acceptance_set).second->right_subformula))
        {
          accepting = true;
          break;
        }
        else if (*temporal_property == *((*acceptance_set).first))
          accepting = false;

        temporal_properties = temporal_properties - *temporal_property;
      }

      if (accepting)
        (*acceptance_set).second->accepting_states.insert(state);
    }
  }
}

/* ========================================================================= */
void AboAutomaton::print(ostream& stream = cout) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Prints information about the automaton.
 *
 * Argument:      stream  --  Output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << "Number of states:          " << size()
          << "\nNumber of acceptance sets: " << acceptance_sets.size()
          << "\nInitial state:             " << gba.initialState()
          << "\n\nAcceptance sets:\n";

  for (map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
           ALLOC(AcceptanceSetInfo*) >::const_iterator
         acceptance_set = acceptance_sets.begin();
       acceptance_set != acceptance_sets.end(); ++acceptance_set)
  {
    estream << "    " << *((*acceptance_set).first) << " : {";

    set<int, less<int>, ALLOC(int) >::const_iterator
      state = (*acceptance_set).second->accepting_states.begin();

    do
    {
      if (state != (*acceptance_set).second->accepting_states.end())
      {
        estream << *state;
        ++state;
        if (state != (*acceptance_set).second->accepting_states.end())
          estream << ", ";
      }
    }
    while (state != (*acceptance_set).second->accepting_states.end());
           
    estream << "}\n";
  }

  Set<int> states = gba.states();
  int* state;
  
  while (!states.isEmpty())
  {
    state = states.some();
    estream << "\nState " << *state
            << "\n    Propositions:        " << gba.propositions(*state)
            << "\n    Temporal properties: " << gba.temporalProperties(*state)
            << "\n    Next states:         " << gba.nextStates(*state) << '\n';
    states = states - *state;
  }
  
  estream.flush();
}

/* ========================================================================= */
void AboAutomaton::write(ostream& output_stream) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes the automaton to a stream using the testbench format
 *                for automaton representation.
 *
 * Argument:      output_stream  --  A reference to the output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&output_stream, ios::failbit | ios::badbit);

  /*
   *  Write the size of the automaton and the number of its acceptance sets
   *  to the stream. Note: Since the acceptance sets are defined in a special
   *  way (one acceptance set for each Until subformula of the formula from
   *  which the automaton was constructed), the nonexistence of any Until
   *  subformulae (and therefore the nonexistence of any acceptance sets)
   *  in fact means that every state of the automaton is an accepting state.
   *  This is taken care of when reading in the automaton.
   */

  estream << size() << '\n' << acceptance_sets.size() << '\n';

  Set<int> states = gba.states();
  Set<int> next_states;
  int state;

  Set<LtlFormula> propositions;

  /*
   *  Iterate through the states of the automaton. For each state, write the
   *  following information to the output stream:
   *
   *       - The number of the state.
   *
   *       - Information about whether the state is the initial state of
   *         the automaton.
   *
   *       - A list of the acceptance sets to which the state belongs
   *         (terminated with `-1').
   *
   *       - Transitions from the state to other states (in the form
   *         [state number] [guard formula]; end the transition list with
   *         `-1'.
   */

  while (!states.isEmpty())
  {
    state = *states.some();
    states = states - state;

    estream << state << ' ' << (state == gba.initialState() ? '1' : '0');

    map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
        ALLOC(AcceptanceSetInfo*) >::size_type 
      acceptance_set_number = 0;

    /*
     *  Write the list of the identifiers of the acceptance sets to which
     *  `state' belongs.
     */

    for (map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less,
             ALLOC(AcceptanceSetInfo*) >::const_iterator
           acceptance_set = acceptance_sets.begin();
         acceptance_set != acceptance_sets.end();
         ++acceptance_set)
    {
      if ((*acceptance_set).second->accepting_states.find(state)
          != (*acceptance_set).second->accepting_states.end())
        estream << ' ' << acceptance_set_number;

      ++acceptance_set_number;
    }

    estream << " -1\n";

    /*
     *  Iterate through the transitions from `state' to other states and write
     *  each transition to the stream (terminating the transition list with
     *  `-1').
     */

    next_states = gba.nextStates(state);

    while (!next_states.isEmpty())
    {
      state = *next_states.some();
      next_states = next_states - state;

      estream << state;

      propositions = gba.propositions(state);

      if (propositions.isEmpty())
        estream << ' ' << FormulaString::LTL_TRUE;
      else
      {
        LtlFormula* proposition;
        ostrstream* proposition_strstream;
        char* output_proposition;

        do
        {
          proposition_strstream = new ostrstream;

          proposition = propositions.some();
          *proposition_strstream << *proposition << ends;
          propositions = propositions - *proposition;

          estream << ' ';
          if (!propositions.isEmpty())
            estream << FormulaString::LTL_AND << ' ';

          output_proposition = proposition_strstream->str();

          switch (output_proposition[0])
          {
          case FormulaString::LTL_TRUE :
            estream << FormulaString::LTL_TRUE;
            break;
          case FormulaString::LTL_FALSE :
            estream << FormulaString::LTL_FALSE;
            break;
          case FormulaString::LTL_NOT :
            estream << FormulaString::LTL_NOT << ' ' << output_proposition + 1;
            break;
          default :
            estream << output_proposition;
            break;
          }
          
          delete proposition_strstream;
        }
        while (!propositions.isEmpty());
      }
      estream << '\n';
    }
    estream << "-1\n";
  }

  estream.flush();
}



/******************************************************************************
 *
 * Function definitions for class AboAutomaton::CreateAboAtom.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void AboAutomaton::CreateAboAtom::
eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info,
     const string& label)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a LtlAtom from a FormulaString token.
 *                Used by AboAutomaton::AboAutomaton(FormulaString& formula)
 *                when traversing the FormulaString to produce an Ltl object.
 *
 * Arguments:     info   --  A reference to a TraversalInfo structure
 *                           containing the stack where partial results of the
 *                           traversal (in this case, the LtlAtom constructed
 *                           from the FormulaString token) are placed.
 *                label  --  The propositional atom in a string representation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  CString id(label.c_str());
  info.data.push(new LtlAtom(id));
}



/******************************************************************************
 *
 * Function definitions for class AboAutomaton::CreateAboConstant.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void AboAutomaton::CreateAboConstant::
eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a Ltl{True, False} object from a
 *                FormulaString token. Used by
 *                AboAutomaton::AboAutomaton(FormulaString& formula) when
 *                traversing the FormulaString to produce an Ltl object.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure
 *                          containing the stack for partial results (in this
 *                          case, the Ltl object representing the constant).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  info.data.push(new T);
}



/******************************************************************************
 *
 * Function definitions for class AboAutomaton::CreateAboUnary.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void AboAutomaton::CreateAboUnary::
eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a Ltl{Not, Next, Finally, Globally}
 *                object from a FormulaString token. Used by
 *                AboAutomaton::AboAutomaton(FormulaString& formula) when
 *                traversing the FormulaString to produce an Ltl object.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure
 *                          containing the stack for partial results (in this
 *                          case, the unary Ltl object).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Ltl* subformula = info.data.top();
  info.data.top() = new T(*subformula);
  delete subformula;
}



/******************************************************************************
 *
 * Function definitions for class AboAutomaton::CreateAboBinary.
 *
 *****************************************************************************/

/* ========================================================================= */
template<class T>
inline void AboAutomaton::CreateAboBinary::
eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for creating a Ltl{Imply, Iff, Resist} object from a
 *                FormulaString token. Used by
 *                AboAutomaton::AboAutomaton(FormulaString& formula) when
 *                traversing the FormulaString to produce an Ltl object.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure
 *                          containing the stack for partial results (in this
 *                          case, the binary Ltl object).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Ltl* subformula2 = info.data.top();
  info.data.pop();
  Ltl* subformula1 = info.data.top();
  info.data.top() = new T(*subformula1, *subformula2);
  delete subformula1;
  delete subformula2;
}

}



/******************************************************************************
 *
 * Main function.
 *
 *****************************************************************************/

int main(int argc, char* argv[])
{
  using FormulaString::FormulaString;
  using abo_buchi::AboAutomaton;

  if (argc == 2 && string(argv[1]) == "-h")
  {
    cerr << "usage: " << argv[0]
         << " [-h] | [{-a} [-f formula | infile] outfile]\n"
            "    -h            show this help\n"
            "    -a            (optional) write automaton information to "
                               "standard output\n"
            "    -f formula    use input formula `formula'\n"
            "    infile        name of input file\n"
            "    outfile       name of output file\n";
    exit(0);
  }

  bool print_automaton = (argc >= 2 && string(argv[1]) == "-a");

  bool use_command_line_formula =
    ((argc >= 2 + (print_automaton ? 1 : 0))
     && string(argv[1 + (print_automaton ? 1 : 0)]) == "-f");

  if ((argc < 3 + (print_automaton ? 1 : 0)
              + (use_command_line_formula ? 1 : 0))
      || (argc > 4 + (print_automaton ? 1 : 0)
                 + (use_command_line_formula ? 1 : 0)))
  {
    cerr << argv[0] << ": Error in command line. Try running `"
         << argv[0] << " -h' for help." << endl;
    exit(-1);
  }

  bool output_file_created = false;
  int exitcode = 0;

  try
  {
    /*
     *  Read the formula from input file or parse it from a command line
     *  argument.
     */

    FormulaString formula;

    if (use_command_line_formula)
    {
      istrstream formula_stream(argv[argc - 2]);
      formula.read(formula_stream);
    }
    else
      formula.read(argv[argc - 2]);

    /*
     *  Construct a generalized Bchi automaton from the formula.
     */

    AboAutomaton automaton(formula);

    /*
     *  Write the constructed automaton to file.
     */

    ofstream automaton_file;
    automaton_file.open(argv[argc - 1], ios::out | ios::trunc);
    if (!automaton_file.good())
      throw FileCreationException(argv[argc - 1]);

    output_file_created = true;
    automaton_file << automaton;
    automaton_file.close();

    if (print_automaton)
      automaton.print(cout);
  }
  catch (IOException e)
  {
    cerr << argv[0] << ": Error: " << e.what() << endl;
    exitcode = -1;
  }
  catch (FormulaString::FormulaString::ParseErrorException e)
  {
    cerr << argv[0] << ": Error: " << e.what() << endl;
    exitcode = -1;
  }
  catch (...)
  {
    cerr << argv[0] << ": Fatal error, aborting." << endl;
    exitcode = -1;
  }

  if (exitcode != 0 && output_file_created)
    remove(argv[argc - 1]);

  return exitcode;
}
