/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _aasa_lbt_h_
#define _aasa_lbt_h_

#include <cstdio>
#include <deque>
#include <fstream>
#include <map>
#include <set>
#include <stack>
#include <string>
#include <strstream>
#include <utility>
#include "aasa/CString.h"
#include "aasa/Ltl.h"
#include "Exception.h"
#include "FormulaString.h"
#include "StlConfig.h"

using namespace std;

namespace abo_buchi
{

using FormulaString::FormulaString;

/******************************************************************************
 *
 * A class for storing information about an acceptance set of a Bchi
 * automaton. Each acceptance set is associated with an Until subformula;
 * this class stores the right operand of the subformula (actually only a
 * pointer to an Ltl object), in addition with the states belonging to the
 * acceptance set associated with the subformula.
 *
 *****************************************************************************/

class AcceptanceSetInfo
{
public:
  AcceptanceSetInfo(const Ltl* subformula);         /* Creates a new acceptance
                                                     * set information object.
                                                     */

  AcceptanceSetInfo(const AcceptanceSetInfo&);      /* Copy constructor. */

  ~AcceptanceSetInfo();                             /* Destructor. */

  AcceptanceSetInfo& operator=                      /* Assignment operator. */
    (const AcceptanceSetInfo&);

private:
  friend class AboAutomaton;

  const Ltl* right_subformula;                      /* Pointer to the right
                                                     * operand of the Until
                                                     * formula associated
                                                     * with the
                                                     * AcceptanceSetInfo
                                                     * object.
						     */

  set<int, less<int>, ALLOC(int) >                  /* A set of accepting */
    accepting_states;                               /* states.            */
};



/******************************************************************************
 *
 * A class defining a `less than'-relation for two pointers to Ltl objects
 * (to be used when storing such pointers in some standard container).
 *
 *****************************************************************************/

class ltl_ptr_less
{
public:
  bool operator()(const Ltl* f1, const Ltl* f2);
};



/******************************************************************************
 *
 * A class for storing a Bchi automaton constructed using the bo
 * implementation of the algorithm presented in [GPVW95].
 *
 * [GPVW95]  R. Gerth, D. Peled, M. Y. Vardi, P. Wolper. Simple on-the-fly
 *           automatic verification of linear temporal logic. In Protocol
 *           Specification, Testing and Verification, pages 3-18. Chapman &
 *           Hall, August 1995
 *
 *****************************************************************************/

class AboAutomaton
{
public:
  AboAutomaton(FormulaString& formula);             /* Creates an automaton
                                                     * from a FormulaString
                                                     * object.
						     */

  AboAutomaton(const AboAutomaton& automaton);      /* Copy constructor. */

  ~AboAutomaton();                                  /* Destructor. */

  AboAutomaton& operator=                           /* Assignment operator. */
    (const AboAutomaton&);

  unsigned long int size() const;                   /* Returns the number of
                                                     * states in the automaton.
                                                     */

  void print(ostream& stream = cout) const;         /* Prints information about
                                                     * the automaton.
						     */

  void write(ostream& output_stream) const;         /* Writes the automaton
                                                     * to a stream using a
                                                     * more compact
                                                     * representation.
						     */

private:
  LtlGraph gba;                                     /* Bchi automaton. */

  map<const Ltl*, AcceptanceSetInfo*, ltl_ptr_less, /* Acceptance sets of */
      ALLOC(AcceptanceSetInfo*) >                   /* the automaton.     */
    acceptance_sets;

  void computeAcceptanceSets();                     /* Computes the acceptance
                                                     * sets of the automaton.
                                                     */

  typedef stack<Ltl*, deque<Ltl*, ALLOC(Ltl*) > >   /* Data type used when  */
    AboFormulaCreationInfo;                         /* `traversing' a       */
                                                    /* FormulaString to     */
                                                    /* create a Ltl object  */
                                                    /* corresponding to the */
                                                    /* formula.             */

  /* Class for translating an atomic proposition into an LtlAtom. */

  class CreateAboAtom
  {
  public:
    static void
      eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info,
           const string& label);
  };

  /* Class for translating a propositional constant into a corresponding
   * Ltl{True, False} object. */

  template<class T>
  class CreateAboConstant
  {
  public:
    static void
      eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info);
  };

  /* Class for translating a unary formula to a corresponding Ltl object. */

  template<class T>
  class CreateAboUnary
  {
  public:
    static void
      eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info);
  };

  /* Class for translating a binary formula to a corresponding Ltl object. */

  template<class T>
  class CreateAboBinary
  {
  public:
    static void
      eval(FormulaString::TraversalInfo<AboFormulaCreationInfo>& info);
  };

};



/******************************************************************************
 *
 * Inline function definitions for class AcceptanceSetInfo.
 *
 *****************************************************************************/

/* ========================================================================= */
inline AcceptanceSetInfo::AcceptanceSetInfo(const Ltl* subformula) :
  right_subformula(subformula)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class AcceptanceSetInfo. Creates a new
 *                AcceptanceSetInfo object, storing a pointer to an Ltl object
 *                corresponding to the right operand subformula of an Until
 *                formula with which the AcceptanceSetInfo object is to be
 *                associated.
 *
 * Argument:      subformula  --  A pointer to a constant Ltl object.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline AcceptanceSetInfo::AcceptanceSetInfo(const AcceptanceSetInfo& info) :
  right_subformula(info.right_subformula->clone()),
  accepting_states(info.accepting_states)
/* ----------------------------------------------------------------------------
 *
 * Description:   Copy constructor for class AcceptanceSetInfo. Creates a copy
 *                of an AcceptanceSetInfo object.
 *
 * Argument:      info  --  A reference to a constant AcceptanceSetInfo object
 *                          of which the new object is made a copy.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
} 

/* ========================================================================= */
inline AcceptanceSetInfo::~AcceptanceSetInfo()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class AcceptanceSetInfo. Deallocates the
 *                memory reserved by an instance of the class.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  delete right_subformula;
}



/******************************************************************************
 *
 * Inline function definitions for class ltl_ptr_less.
 *
 *****************************************************************************/

/* ========================================================================= */
inline bool ltl_ptr_less::operator()(const Ltl* f1, const Ltl* f2)
/* ----------------------------------------------------------------------------
 *
 * Description:   Maps the `less than' relation between two pointers to Ltl
 *                objects to the corresponding relation between the objects
 *                itself.
 *
 * Arguments:     f1, f2  --  Two pointers to constant Ltl objects.
 *
 * Returns:       A truth value.
 *
 * ------------------------------------------------------------------------- */
{
  return (*f1 < *f2);
}

}

/* ========================================================================= */
inline ostream& operator<<
  (ostream& stream, const abo_buchi::AboAutomaton& automaton)
/* ----------------------------------------------------------------------------
 *
 * Description:   Defines an alternate method for writing an AboAutomaton to a
 *                stream, i.e. by using the << operator.
 *
 * Arguments:     stream     --  A reference to the output stream.
 *
 *                automaton  --  The Bchi automaton.
 *
 * -------------------------------------------------------------------------- */ 
{
  automaton.write(stream);
  return stream;
}

#endif /* _aasa_lbt_h_ */
