/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "main.h"

void breakHandler(int)
{
  user_break = true;
}

void testLoop(const char* command_line_formula)
{
  /*
   *  Initialize the data structure with information about the current test
   *  round.
   */

  TestRoundInfo round_info;

  round_info.abort = false;
  round_info.next_round_to_run = 1;

  round_info.next_round_to_stop
    = (configuration.global_options.interactive == Configuration::ALWAYS
       ? 1
       : configuration.global_options.number_of_rounds + 1);

  round_info.statespace = 0;
  round_info.number_of_rounds_to_change_statespace = 0;
  round_info.number_of_rounds_to_change_formula = 0;

  if (tmpnam(round_info.formula_file_name) == 0
      || tmpnam(round_info.negated_formula_file_name) == 0)
    throw Exception("Unable to allocate names for temporary files.");

  /*
   *  If the state spaces are to be enumerated using the given parameters,
   *  initialize a Graph::SinglePathIterator object.
   */

  Graph::SinglePathIterator* path_iterator = 0;

  if (configuration.statespace_options.generate_mode 
        == Configuration::ENUMERATEDPATH)
  {
    path_iterator = new Graph::SinglePathIterator
                          (configuration.statespace_options.atoms_per_state,
			   configuration.statespace_options.size);
    round_info.statespace = &(**path_iterator);
  }

  unsigned long int generated_statespaces = 0, generated_formulae = 0;

  Bitset* path_truth_values = 0;

  /*
   *  Intialize a vector for storing the test results for each algorithm.
   */

  vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) > test_results
    (configuration.algorithms.size(),
     AlgorithmTestResults(configuration.algorithms.size()));

  /*
   *  Initialize a vector for collecting the test statistics for each
   *  algorithm.
   */

  vector<TestStatistics, ALLOC(TestStatistics) >
    final_statistics(configuration.algorithms.size(),
                     TestStatistics(configuration.algorithms.size()));

  UserInterface user_interface
                  (round_info, test_results, final_statistics);

  for (round_info.current_round = 1;
       !round_info.abort
       && round_info.current_round
	 <= configuration.global_options.number_of_rounds;
       ++round_info.current_round)
  {
    user_break = false;
    round_info.error = false;
    round_info.skip = false;

    if (configuration.global_options.verbosity >= 1)
      cout << "Round "
	   << round_info.current_round
	   << '/'
	   << configuration.global_options.number_of_rounds
	   << endl
	   << endl;

    /*
     *  Test whether it is necessary to generate a new state space.
     */

    if (round_info.number_of_rounds_to_change_statespace == 0)
    {
      if (configuration.statespace_options.generate_mode
	    != Configuration::ENUMERATEDPATH)
      {
	if (round_info.statespace != 0)
	  delete round_info.statespace;

	round_info.statespace = generateStateSpace();
	++generated_statespaces;

	round_info.number_of_rounds_to_change_statespace
	  = configuration.global_options.statespace_change_interval;
      }
    }

    /*
     *  Test whether it is necessary to generate a new LTL formula. If a
     *  formula was given as a command line parameter, use that formula.
     */

    if (command_line_formula != 0)
      generateFormulae(round_info.formulae, command_line_formula);
    else if (round_info.number_of_rounds_to_change_formula == 0)
    {
      ++generated_formulae;
      generateFormulae(round_info.formulae);
      
      round_info.number_of_rounds_to_change_formula
	= configuration.global_options.formula_change_interval;
    }

    if (round_info.current_round < round_info.next_round_to_run)
    {
      if (configuration.global_options.verbosity >= 1)
	cout << "    Skipped round " << round_info.current_round << ".\n\n";
      round_info.skip = true;
	
      if (user_break)
	round_info.next_round_to_stop = round_info.current_round;
    }
    else
    {
      if (configuration.statespace_options.generate_mode
	    != Configuration::RANDOMGRAPH)
      {
	if (configuration.global_options.verbosity >= 2)
	{
	  user_interface.printHeading
	    (cout, 4, "Model checking formula in the path...");
	  cout << endl;
	}

	LtlFormula::LtlFormula* formula;
	path_truth_values = new Bitset(0);

	formula
	  = LtlFormula::formulaStringToLtlFormula(round_info.formulae[2]);
	formula->evaluatePath(*(round_info.statespace), path_truth_values);
	delete formula;
      }

      for (vector<AlgorithmTestResults,
	          ALLOC(AlgorithmTestResults) >::iterator
	     result = test_results.begin();
	   result != test_results.end(); ++result)
	result->reset();

      try
      {
	writeFormulaToFile
	  (round_info.formulae[configuration.formula_options.output_mode
		                 == Configuration::NNF ? 0 : 2],
	   round_info.formula_file_name);

	writeFormulaToFile
	  (round_info.formulae[configuration.formula_options.output_mode
			         == Configuration::NNF ? 1 : 3],
	   round_info.negated_formula_file_name);
      }
      catch (IOException e)
      {
	round_info.error = true;
      }

      if (!round_info.error)
      {
	for (vector<Configuration::AlgorithmInformation,
	            ALLOC(Configuration::AlgorithmInformation) >::size_type
	       algorithm_id = 0;
	     algorithm_id < configuration.algorithms.size();
	     ++algorithm_id)
	{
	  if (user_break)
	  {
	    round_info.next_round_to_stop = round_info.current_round;
	    user_break = false;
	    break;
	  }

	  if (configuration.global_options.verbosity >= 2)
	    cout << "    `"
		    + *(configuration.algorithms[algorithm_id].name) + '\''
		 << endl;

	  if (!configuration.algorithms[algorithm_id].enabled)
	  {
	    if (configuration.global_options.verbosity >= 2)
	    {
	      user_interface.printHeading(cout, 6, "[Disabled]");
	      cout << endl;
	    }
	  }
	  else
	  {
	    for (int counter = 0; counter < 2; counter++)
	    {
	      if (user_break)
	      {
		round_info.next_round_to_stop = round_info.current_round;
		break;
	      }

	      if (configuration.global_options.verbosity >= 2)
		user_interface.printHeading
		  (cout, 6, string(counter == 1 ? "Negated" : "Positive")
		            + " formula:");

	      AutomatonStats& automaton_stats = 
		test_results[algorithm_id].automaton_stats[counter];

	      try
	      {
		/*
		 *  Generate a Bchi automaton using the current algorithm.
		 *  `counter' determines the formula which is to be
		 *  translated into an automaton; 0 denotes the positive and
		 *  1 the negated formula.
		 */

		final_statistics[algorithm_id].automaton_count[counter]++;

		if (configuration.global_options.verbosity >= 2)
		  user_interface.printHeading
		    (cout, 8, (configuration.global_options.verbosity == 2
			       ? "Computing Bchi automaton..."
			       : "Bchi automaton:"));
            
		round_info.buchi_automaton = generateBuchiAutomaton
		  ((counter == 1 ? round_info.negated_formula_file_name
		                 : round_info.formula_file_name),
		   configuration.algorithms[algorithm_id],
		   automaton_stats,
		   10);
            
		if (configuration.global_options.verbosity >= 3)
		  user_interface.printBuchiAutomatonStats
		    (cout, 10, algorithm_id, counter);
		
		/*
		 *  Update Bchi automaton statistics for the current
		 *  algorithm.
		 */

		final_statistics[algorithm_id].
		  total_number_of_buchi_states[counter] +=
		  automaton_stats.number_of_buchi_states;
		final_statistics[algorithm_id].
		  total_number_of_buchi_transitions[counter] +=
		  automaton_stats.number_of_buchi_transitions;
		  
		if (final_statistics[algorithm_id].
		      total_buchi_generation_time[counter] < 0.0
		    || automaton_stats.buchi_generation_time < 0.0)
		  final_statistics[algorithm_id].
		    total_buchi_generation_time[counter] = -1.0;
		else
		  final_statistics[algorithm_id].
		    total_buchi_generation_time[counter] +=
		      automaton_stats.buchi_generation_time;
	      
		/*
		 *  Compute the product of the Bchi automaton with the
		 *  reachability graph.
		 */

		if (configuration.global_options.verbosity >= 2)
		  user_interface.printHeading
		    (cout, 8, (configuration.global_options.verbosity == 2
			       ? "Computing product automaton..."
			       : "Product automaton:"));

		round_info.product_automaton = generateProductAutomaton
		  (*(round_info.buchi_automaton), *(round_info.statespace),
		   automaton_stats, 10);
        
		if (configuration.global_options.verbosity >= 3)
		  user_interface.printProductAutomatonStats
		    (cout, 10, algorithm_id, counter);

		/*
		 *  Update product automaton statistics for the current
		 *  algorithm.
		 */

		final_statistics[algorithm_id].
		  total_number_of_product_states[counter] +=
		  automaton_stats.number_of_product_states;
		final_statistics[algorithm_id].
		  total_number_of_product_transitions[counter] +=
		  automaton_stats.number_of_product_transitions;

		/*
		 *  Release the memory allocated for the Bchi automaton.
		 */

		if (configuration.global_options.verbosity >= 4)
		{
		  cout << string(8, ' ')
		          + "<cleaning up memory allocated for Bchi "
		            "automaton>";
		  cout.flush();
		}

		delete round_info.buchi_automaton;
		  
		if (configuration.global_options.verbosity >= 4)
		  cout << " ok" << endl;

		/*
		 *  Find the system states from which an accepting execution
		 *  cycle can be reached by checking the product automaton
		 *  for emptiness.
		 */
          
		if (configuration.global_options.verbosity >= 2)
		  user_interface.printHeading
		    (cout, 8, (configuration.global_options.verbosity == 2
			       ? "Performing emptiness check..."
			       : "Accepting cycles:"));
	      
		performEmptinessCheck
		  (*(round_info.product_automaton), automaton_stats,
		   (configuration.global_options.product_mode
		      == Configuration::GLOBAL
		    ? round_info.statespace->size()
		    : 1));
                                
		if (configuration.global_options.verbosity >= 3)
		  user_interface.printAcceptanceCycleStats
		    (cout, 10, algorithm_id, counter);
		
		/*
		 *  Release the memory allocated for the product automaton.
		 */

		if (configuration.global_options.verbosity >= 4)
		{
		  cout << string(8, ' ')
		          + "<cleaning up memory allocated for product "
		            "automaton>";
		  cout.flush();
		}

		delete round_info.product_automaton;
	      
		if (configuration.global_options.verbosity >= 4)
		  cout << " ok" << endl;
	      }
	      catch (...)
	      {
		round_info.error = true;

		try
		{
		  throw;
		}
		catch (BuchiAutomatonGenerationException)
		{
		  final_statistics[algorithm_id].
		    failures_to_compute_automaton[counter]++;
		}
		catch (ProductAutomatonGenerationException)
		{
		  delete round_info.buchi_automaton;
		}
		catch (EmptinessCheckFailedException)
		{
		  delete round_info.product_automaton;
		}
		catch (...)
		{
		  throw;
		}
	      }
	    }

	    /*
	     *  If the emptiness check was performed successfully for the
	     *  product automata constructed from both the positive and
	     *  negated formulae, test whether the emptiness check results
	     *  are consistent with each other. (It should not be possible
	     *  for both the formula and its negation to be true in any
	     *  state.)
	     */

	    if (test_results[algorithm_id].automaton_stats[0].
		  emptinessCheckPerformed()
		&& test_results[algorithm_id].automaton_stats[1].
		     emptinessCheckPerformed())
	    {
	      final_statistics[algorithm_id].consistency_checks_performed++;

	      if (configuration.global_options.verbosity >= 2)
		user_interface.printHeading
		  (cout, 6, (configuration.global_options.verbosity == 2
			     ? "Performing consistency check..."
			     : "Consistency check:"));

	      if (!performConsistencyCheck(test_results[algorithm_id]))
	      {
		round_info.error = true;
		final_statistics[algorithm_id].consistency_check_failures++;
	      }

	      if (configuration.global_options.verbosity >= 3)
		user_interface.printConsistencyCheckStats
		  (cout, 8, algorithm_id);

	      if (configuration.statespace_options.generate_mode
                    != Configuration::RANDOMGRAPH)
	      {
		if (!performPathCheck(path_truth_values,
				      test_results[algorithm_id]))
		  round_info.error = true;
	      }
	    }
	    
	    /*
	     *  Compare the emptiness check results with the previously
	     *  computed results obtained using different algorithms.
	     */

	    if (configuration.global_options.verbosity >= 2)
	    {
	      if (algorithm_id == 0)
		cout << endl;
	      else
	      {
		user_interface.printHeading
		  (cout, 6, (configuration.global_options.verbosity == 2
			     ? "Comparing results with other algorithms..."
			     : "Result comparison:"));
	      }
	    }

	    if (!compareResults(round_info, test_results, algorithm_id,
				final_statistics))
	      round_info.error = true;
	      
	    if (configuration.global_options.verbosity >= 3)
	      user_interface.printCrossComparisonStats
		(cout, 8, algorithm_id, algorithm_id);
	  }
	}
      }
    }

    if (round_info.error &&
        configuration.global_options.interactive == Configuration::ONERROR)
      round_info.next_round_to_stop = round_info.current_round;

    if (round_info.next_round_to_stop == round_info.current_round)
      user_interface.executeUserCommands
	(generated_statespaces, generated_formulae);

    if (round_info.next_round_to_run == round_info.current_round)
      round_info.next_round_to_run++;

    if (!round_info.skip)
    {
      for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::iterator
	     result = test_results.begin();
           result != test_results.end();
           ++result)
      {
	for (int counter = 0; counter < 2; counter++)
	  if (result->automaton_stats[counter].emptiness_check_result != 0)
	    delete result->automaton_stats[counter].emptiness_check_result;
      }

      if (path_truth_values != 0)
      {
	delete path_truth_values;
	path_truth_values = 0;
      }

      removeFile(round_info.formula_file_name);
      removeFile(round_info.negated_formula_file_name);
    }

    if (configuration.statespace_options.generate_mode
	  != Configuration::ENUMERATEDPATH)
    {
      if (configuration.global_options.statespace_change_interval == 0)
	round_info.number_of_rounds_to_change_statespace = 1;
      else
	round_info.number_of_rounds_to_change_statespace--;
    }
    else
    {
      ++(*path_iterator);
      if (path_iterator->atEnd())
      {
	delete path_iterator;
	path_iterator
	  = new Graph::SinglePathIterator
	          (configuration.statespace_options.atoms_per_state,
		   configuration.statespace_options.size);
	round_info.statespace = &(**path_iterator);
      }
    }

    if (configuration.global_options.formula_change_interval == 0)
      round_info.number_of_rounds_to_change_formula = 1;
    else
      round_info.number_of_rounds_to_change_formula--;
  }

  if (path_iterator != 0)
    delete path_iterator;
  else
    delete round_info.statespace;

  if (configuration.global_options.verbosity >= 1)
    user_interface.printCollectiveStats
      (cout, 0, round_info.current_round - 1, generated_statespaces,
       generated_formulae);
}









int main(int argc, char* argv[])
{
  try 
  {
    configuration.read();
  }
  catch (Configuration::ConfigurationException e) 
  {
    cerr << argv[0] << ": Error reading configuration. ";
    if (e.line_number != -1)
      cerr << "Line " << e.line_number << ": ";
    cerr << e.what() << endl;
    exit(-1);
  }

  if (configuration.global_options.verbosity >= 3)
    configuration.print(cout);

  user_break = false;
  signal(SIGINT, breakHandler);

  try 
  {
    testLoop(0);
  }
  catch (Exception e) 
  {
    cerr << endl << argv[0] << ": Error: " << e.what() << endl;
    exit(-1);
  }
  catch (bad_alloc)
  {
    cerr << endl << argv[0] << ": out of memory" << endl;
    exit(-1);
  }

  return 0;
}
