/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "prod_lbt.h"

namespace prod_buchi
{

char* program_name;

inline void TranslateAtom::eval
  (FormulaString::TraversalInfo<string*>& info, const string& atom)
{
  atoms.insert(atom);
  *(info.data) += string("(") + atom + " != empty)";
}

template<const char* C>
inline void TranslateConstant<C>::
eval(FormulaString::TraversalInfo<string*>& info)
{
  *(info.data) += '(';
  *(info.data) += C;
  *(info.data) += ')';
}

template<const char* C>
inline void TranslateUnary<C>::
eval(FormulaString::TraversalInfo<string*>& info)
{
  if (info.operand == 0)
  {
    *(info.data) += C;
    *(info.data) += ' ';
  }
}

template<const char* C>
inline void TranslateBinary<C>::
eval(FormulaString::TraversalInfo<string*>& info)
{
  switch (info.operand)
  {
  case 0 : *(info.data) += '('; break;
  case 1 : *(info.data) += ' '; *(info.data) += C; *(info.data) += ' ';
    break;
  default : *(info.data) += ')'; break;
  }
}

void TranslateV::eval(FormulaString::TraversalInfo<string*>& info)
{
  switch (info.operand)
  {
    case 0 : 
      if (!warned_about_operator_expansion)
      {
        cerr << program_name << ": Warning: Formula length changed due to "
                                "operator expansion(s)." << endl;
        warned_about_operator_expansion = true;
      }
      *(info.data) += "not (not ";
      break;
    case 1 :
      *(info.data) += " until not ";
      break;
    default :
      *(info.data) += ')';
      break;
  }
}

void TranslateEquivalence::eval(FormulaString::TraversalInfo<string*>& info)
{
  switch (info.operand)
  {
    case 0 :
      if (!TranslateV::warned_about_operator_expansion)
      {
        cerr << program_name << ": Warning: Formula length changed due to "
                                "operator expansion(s)." << endl;
        TranslateV::warned_about_operator_expansion = true;
      }
      operand_stack.push(info.data);
      info.data = new string;
      break;
    case 1 :
      operand_stack.push(info.data);
      info.data = new string;
      break;
    default :
      {
        string* right_subformula = info.data;
        string* left_subformula = operand_stack.top();
        operand_stack.pop();
        info.data = operand_stack.top();
        operand_stack.pop();
        *(info.data) += "(((" + *left_subformula + ") implies ("
                        + *right_subformula + ")) and (("
                        + *right_subformula + ") implies ("
                        + *left_subformula + ")))";
        delete right_subformula;
        delete left_subformula;
      }
      break;
  }
}

void TranslateError::eval(FormulaString::TraversalInfo<string*>& info)
{
  throw FormulaString::FormulaString::ParseErrorException
    ("Illegal operators in formula.");
}

}

int main(int argc, char* argv[])
{
  using namespace prod_buchi;
  using prod_buchi::FormulaString;

  if (argc != 4)
  {
    cerr << argv[0] << ": Invalid number of command line arguments." << endl;
    exit(-1);
  }

  program_name = argv[0];

  char prod_input_filename[L_tmpnam + 1];
  char prod_output_filename[L_tmpnam + 1];
  bool prod_input_file_created = false, prod_output_file_created = false;

  int exitstatus = 0;

  try
  {
    if (tmpnam(prod_input_filename) == 0 || tmpnam(prod_output_filename) == 0)
      throw FileCreationException("a temporary file");

    /*  
     *  Read the formula string from input file.
     */

    FormulaString formula;

    formula.read(argv[2]);

    /*
     *  Translate the formula into PROD's input file format. This is done by
     *  `traversing' the FormulaString object, i.e. scanning through the
     *  formula and applying suitable translations to the symbols in the
     *  formula. The result of the translation will be found in the string
     *  `prod_formula'.
     */

    string prod_formula;
    FormulaString::TraversalInfo<string*> info;
    info.data = &prod_formula;

    TranslateAtom::atoms.clear();

    formula.template traverse<TranslateAtom,
                              TranslateConstant<PROD_TRUE>,
                              TranslateConstant<PROD_FALSE>,
                              TranslateUnary<PROD_NOT>,
                              TranslateError,
                              TranslateUnary<PROD_FINALLY>,
                              TranslateUnary<PROD_GLOBALLY>,
                              TranslateBinary<PROD_AND>,
                              TranslateBinary<PROD_OR>,
                              TranslateBinary<PROD_IMPLICATION>,
                              TranslateEquivalence,
                              TranslateBinary<PROD_UNTIL>,
                              TranslateV>
      (FormulaString::PREORDER | FormulaString::INORDER |
       FormulaString::POSTORDER, info);

    /*
     *  Create a file containing the input for PROD.
     */

    ofstream prod_input_file;
    prod_input_file.open((string(prod_input_filename) + ".net").c_str(),
                         ios::out | ios::trunc);
    if (!prod_input_file.good())
      throw FileCreationException(prod_input_filename);

    prod_input_file_created = true;

    Exceptional_ostream eprod_input_file(&prod_input_file,
                                         ios::failbit | ios::badbit);

    /*
     *  Write the propositional atoms used in the formula as #place's to
     *  PROD's input file.
     */

    for (set<string, less<string>, ALLOC(string) >::const_iterator
	   atom = TranslateAtom::atoms.begin();
         atom != TranslateAtom::atoms.end();
         ++atom)
      eprod_input_file << "#place " << *atom << '\n';

    /*
     *  Write the formula itself to the file.
     */

    eprod_input_file << "#verify not " << prod_formula << ";\n";
    eprod_input_file.flush();
    prod_input_file.close();

    /*
     *  Create a directory for the extra files that get created when PROD is
     *  run with the net description file containing the input formula
     *  (PROD requires that such a directory exists).
     */

    system((string("mkdir ") + prod_input_filename + ".src").c_str());

    /*
     *  Execute PROD (prpp to be exact).
     */

    string prod_command_line;

    prod_command_line = string(argv[1]) + " -N " + prod_input_filename
                        + " >" + prod_output_filename;

    int exitcode = system(prod_command_line.c_str());

    /*
     *  Check the output message file for cases where the input formula was
     *  valid or unsatisfiable; in these cases, PROD will not generate an
     *  automaton description for the formula, so it must be created here.
     */

    ifstream prod_message_file;
    string message_line;
    int message_type = 0;

    prod_message_file.open(prod_output_filename);
    while (prod_message_file.good())
    {
      message_line = "";

      getline(prod_message_file, message_line, '\n');

      if (!prod_message_file.eof())
      {
        cout << message_line << endl;

        if (message_line.find("initial marking") != message_line.npos)
          exitcode = 0;
        else if (message_line.find("equivalent to 1") != message_line.npos)
          message_type = -1;
        else if (message_line.find("equivalent to 0") != message_line.npos)
          message_type = 1;
      }
    }
    prod_message_file.close();

    /*
     *  If PROD returned with a nonzero exit code and no message indicating
     *  that the input formula was either valid or unsatisfiable, assume
     *  that an error occurred.
     */

    if (message_type == 0 && exitcode != 0)
      throw ExecFailedException(argv[1]);

    /*
     *  Remove the extra files created by PROD.
     */

    prod_output_file_created = true;

    remove(prod_output_filename);
    rename((string(prod_input_filename) + ".nev").c_str(),
           prod_output_filename);

    NeverClaimParser::NeverClaimAutomaton automaton;

    switch (message_type)
    {
      case -1 :
        /*
         *  In this case, the automaton is empty and will reject all
         *  executions.
         */

        break;

      case 0 :
        /*
         *  Parse the automaton in the never claim generated by PROD.
         */

        automaton.read(prod_output_filename);
        break;

    default :
        /*
         *  Create an automaton with a single accepting state, so that the
         *  automaton will accept all executions.
         */

      automaton.addNewState();
      automaton.addNewLabel("0");
      automaton.currentState()->initial() = true;
      automaton.currentState()->accepting() = true;
      automaton.currentState()->addTransition("0", new string("t"));
      break;
    }

    /*
     *  Write the automaton to the final output file.
     */

    automaton.write(argv[3]);
  }
  catch (...)
  {
    cerr << argv[0] << ": ";
    exitstatus = -1;

    try
    {
      throw;
    }
    catch (IOException e)
    {
      cerr << e.what();
    }
    catch (NeverClaimParser::ParseErrorException e)
    {
      cerr << e.what();
    }
    catch (FormulaString::FormulaString::ParseErrorException e)
    {
      cerr << e.what();
    }
    catch (ExecFailedException e)
    {
      cerr << e.what();
    }
    catch (...)
    {
      cerr << "Fatal error, aborting.";
    }

    cerr << endl;
  }

  if (prod_input_file_created)
    remove((string(prod_input_filename) + ".net").c_str());

  if (prod_output_file_created)
  {
    remove((string(prod_input_filename) + ".def").c_str());
    remove((string(prod_input_filename) + ".dat").c_str());
    system((string("rm -rf ") + prod_input_filename + ".src").c_str());
    remove(prod_output_filename);
  }

  return exitstatus;
}
